#line 1 "fs.brg"


/**
 * @file
 *
 * Compile parse tree into XQuery Core. This file implements
 * (more or less) the W3C XQuery Formal Semantics Working Draft.
 *
 * Copyright Notice:
 * -----------------
 *
 * The contents of this file are subject to the Pathfinder Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License.  You may obtain a copy of the License at
 * http://monetdb.cwi.nl/Legal/PathfinderLicense-1.1.html
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is the Pathfinder system.
 *
 * The Original Code has initially been developed by the Database &
 * Information Systems Group at the University of Konstanz, Germany and
 * the Database Group at the Technische Universitaet Muenchen, Germany.
 * It is now maintained by the Database Systems Group at the Eberhard
 * Karls Universitaet Tuebingen, Germany.  Portions created by the
 * University of Konstanz, the Technische Universitaet Muenchen, and the
 * Universitaet Tuebingen are Copyright (C) 2000-2005 University of
 * Konstanz, (C) 2005-2008 Technische Universitaet Muenchen, and (C)
 * 2008-2010 Eberhard Karls Universitaet Tuebingen, respectively.  All
 * Rights Reserved.
 *
 *
 * $Id$
 */ 


#include "pf_config.h"
#include "pathfinder.h"
#include <assert.h>
#include <stdio.h>

#include "oops.h"
#include "abssyn.h"
#include "qname.h"
#include "mem.h"
#include "options.h"      /* extract option declarations from parse tree */
#include "string_utils.h" /* trimming for pragmas */

/* PFcnode_t */
#include "core.h"

/*
 * We need subtyping stuff for semantics of function application.
 */
#include "subtyping.h"

/* PFvar_t */
#include "variable.h"

/* Easily access subtree-parts */
#include "child_mnemonic.h"

/*
 * Accessors for the burg matcher
 */
typedef struct PFpnode_t *NODEPTR_TYPE;

/* accessor to the node kind */
#define OP_LABEL(p)    ((p)->kind) 

/* accessors to left and right child node */
#define LEFT_CHILD(p)  ((p)->child[0])
#define RIGHT_CHILD(p) ((p)->child[1])

/* the state determined during bottom-up labeling is stored here */
#define STATE_LABEL(p) ((p)->state_label)

/* If something goes wrong, call PFoops */
#define PANIC(...) PFoops (OOPS_BURG, __VA_ARGS__)

#ifndef PFfs_PANIC
#define PFfs_PANIC	PANIC
#endif /* PFfs_PANIC */
#include <stdlib.h> /* for abort() */
#ifdef NDEBUG
#define PFfs_assert(x,y)	;
#else
#define PFfs_assert(x,y)	if(!(x)) {y; abort();}
#endif
static short PFfs_r1_nts[] ={ 2, 0 };
static short PFfs_r2_nts[] ={ 4, 5, 0 };
static short PFfs_r3_nts[] ={ 5, 26, 0 };
static short PFfs_r5_nts[] ={ 0 };
static short PFfs_r6_nts[] ={ 6, 0 };
static short PFfs_r7_nts[] ={ 7, 6, 0 };
static short PFfs_r9_nts[] ={ 21, 39, 28, 6, 0 };
static short PFfs_r10_nts[] ={ 21, 39, 6, 0 };
static short PFfs_r11_nts[] ={ 8, 0 };
static short PFfs_r12_nts[] ={ 9, 0 };
static short PFfs_r13_nts[] ={ 10, 0 };
static short PFfs_r14_nts[] ={ 11, 0 };
static short PFfs_r15_nts[] ={ 12, 0 };
static short PFfs_r16_nts[] ={ 14, 0 };
static short PFfs_r17_nts[] ={ 15, 0 };
static short PFfs_r18_nts[] ={ 16, 0 };
static short PFfs_r19_nts[] ={ 17, 0 };
static short PFfs_r20_nts[] ={ 13, 0 };
static short PFfs_r21_nts[] ={ 18, 0 };
static short PFfs_r22_nts[] ={ 19, 0 };
static short PFfs_r23_nts[] ={ 25, 0 };
static short PFfs_r24_nts[] ={ 3, 0 };
static short PFfs_r36_nts[] ={ 20, 0 };
static short PFfs_r41_nts[] ={ 22, 0 };
static short PFfs_r43_nts[] ={ 23, 22, 0 };
static short PFfs_r44_nts[] ={ 39, 72, 0 };
static short PFfs_r46_nts[] ={ 39, 0 };
static short PFfs_r47_nts[] ={ 72, 27, 0 };
static short PFfs_r48_nts[] ={ 72, 39, 27, 0 };
static short PFfs_r49_nts[] ={ 72, 24, 0 };
static short PFfs_r50_nts[] ={ 27, 0 };
static short PFfs_r52_nts[] ={ 28, 0 };
static short PFfs_r53_nts[] ={ 27, 27, 0 };
static short PFfs_r54_nts[] ={ 29, 0 };
static short PFfs_r55_nts[] ={ 30, 0 };
static short PFfs_r56_nts[] ={ 36, 0 };
static short PFfs_r57_nts[] ={ 37, 0 };
static short PFfs_r58_nts[] ={ 46, 0 };
static short PFfs_r59_nts[] ={ 47, 0 };
static short PFfs_r60_nts[] ={ 31, 34, 27, 0 };
static short PFfs_r61_nts[] ={ 31, 34, 35, 27, 0 };
static short PFfs_r63_nts[] ={ 72, 27, 31, 0 };
static short PFfs_r64_nts[] ={ 72, 33, 27, 31, 0 };
static short PFfs_r65_nts[] ={ 72, 32, 27, 31, 0 };
static short PFfs_r66_nts[] ={ 72, 33, 32, 27, 31, 0 };
static short PFfs_r69_nts[] ={ 72, 0 };
static short PFfs_r74_nts[] ={ 27, 35, 0 };
static short PFfs_r75_nts[] ={ 72, 27, 27, 0 };
static short PFfs_r76_nts[] ={ 72, 33, 27, 27, 0 };
static short PFfs_r79_nts[] ={ 27, 38, 39, 27, 38, 27, 0 };
static short PFfs_r82_nts[] ={ 41, 0 };
static short PFfs_r83_nts[] ={ 42, 0 };
static short PFfs_r85_nts[] ={ 43, 0 };
static short PFfs_r92_nts[] ={ 44, 0 };
static short PFfs_r97_nts[] ={ 45, 0 };
static short PFfs_r101_nts[] ={ 27, 27, 27, 0 };
static short PFfs_r102_nts[] ={ 48, 0 };
static short PFfs_r103_nts[] ={ 48, 47, 0 };
static short PFfs_r104_nts[] ={ 49, 0 };
static short PFfs_r105_nts[] ={ 49, 48, 0 };
static short PFfs_r106_nts[] ={ 50, 0 };
static short PFfs_r107_nts[] ={ 50, 50, 0 };
static short PFfs_r123_nts[] ={ 51, 0 };
static short PFfs_r125_nts[] ={ 52, 0 };
static short PFfs_r126_nts[] ={ 51, 52, 0 };
static short PFfs_r128_nts[] ={ 53, 0 };
static short PFfs_r129_nts[] ={ 53, 52, 0 };
static short PFfs_r133_nts[] ={ 54, 0 };
static short PFfs_r134_nts[] ={ 54, 53, 0 };
static short PFfs_r135_nts[] ={ 55, 0 };
static short PFfs_r136_nts[] ={ 55, 54, 0 };
static short PFfs_r138_nts[] ={ 56, 0 };
static short PFfs_r139_nts[] ={ 56, 39, 0 };
static short PFfs_r140_nts[] ={ 57, 0 };
static short PFfs_r141_nts[] ={ 57, 39, 0 };
static short PFfs_r142_nts[] ={ 58, 0 };
static short PFfs_r143_nts[] ={ 58, 40, 0 };
static short PFfs_r144_nts[] ={ 59, 0 };
static short PFfs_r145_nts[] ={ 59, 40, 0 };
static short PFfs_r146_nts[] ={ 60, 0 };
static short PFfs_r149_nts[] ={ 61, 0 };
static short PFfs_r150_nts[] ={ 62, 0 };
static short PFfs_r151_nts[] ={ 89, 0 };
static short PFfs_r153_nts[] ={ 64, 0 };
static short PFfs_r154_nts[] ={ 63, 0 };
static short PFfs_r156_nts[] ={ 62, 62, 0 };
static short PFfs_r157_nts[] ={ 43, 62, 0 };
static short PFfs_r158_nts[] ={ 63, 27, 0 };
static short PFfs_r160_nts[] ={ 65, 0 };
static short PFfs_r161_nts[] ={ 62, 27, 0 };
static short PFfs_r163_nts[] ={ 66, 0 };
static short PFfs_r165_nts[] ={ 73, 0 };
static short PFfs_r166_nts[] ={ 74, 0 };
static short PFfs_r167_nts[] ={ 75, 0 };
static short PFfs_r168_nts[] ={ 79, 0 };
static short PFfs_r169_nts[] ={ 77, 0 };
static short PFfs_r170_nts[] ={ 78, 0 };
static short PFfs_r171_nts[] ={ 104, 0 };
static short PFfs_r172_nts[] ={ 67, 0 };
static short PFfs_r173_nts[] ={ 71, 0 };
static short PFfs_r174_nts[] ={ 68, 0 };
static short PFfs_r175_nts[] ={ 69, 0 };
static short PFfs_r176_nts[] ={ 70, 0 };
static short PFfs_r185_nts[] ={ 76, 0 };
static short PFfs_r187_nts[] ={ 27, 76, 0 };
static short PFfs_r190_nts[] ={ 80, 0 };
static short PFfs_r191_nts[] ={ 81, 0 };
static short PFfs_r192_nts[] ={ 82, 0 };
static short PFfs_r193_nts[] ={ 83, 0 };
static short PFfs_r194_nts[] ={ 84, 0 };
static short PFfs_r195_nts[] ={ 85, 0 };
static short PFfs_r196_nts[] ={ 86, 87, 0 };
static short PFfs_r197_nts[] ={ 86, 88, 0 };
static short PFfs_r200_nts[] ={ 27, 87, 0 };
static short PFfs_r206_nts[] ={ 87, 87, 0 };
static short PFfs_r209_nts[] ={ 88, 88, 0 };
static short PFfs_r210_nts[] ={ 90, 0 };
static short PFfs_r211_nts[] ={ 90, 28, 0 };
static short PFfs_r214_nts[] ={ 91, 0 };
static short PFfs_r216_nts[] ={ 92, 0 };
static short PFfs_r217_nts[] ={ 93, 0 };
static short PFfs_r218_nts[] ={ 95, 0 };
static short PFfs_r219_nts[] ={ 94, 0 };
static short PFfs_r220_nts[] ={ 96, 0 };
static short PFfs_r221_nts[] ={ 99, 100, 0 };
static short PFfs_r222_nts[] ={ 99, 0 };
static short PFfs_r223_nts[] ={ 100, 29, 0 };
static short PFfs_r224_nts[] ={ 99, 101, 0 };
static short PFfs_r225_nts[] ={ 97, 29, 29, 0 };
static short PFfs_r227_nts[] ={ 98, 97, 0 };
static short PFfs_r233_nts[] ={ 102, 0 };
static short PFfs_r234_nts[] ={ 72, 103, 29, 29, 0 };
static short PFfs_r237_nts[] ={ 28, 75, 0 };
short *PFfs_nts[] = {
	0,
	PFfs_r1_nts,
	PFfs_r2_nts,
	PFfs_r3_nts,
	PFfs_r2_nts,
	PFfs_r5_nts,
	PFfs_r6_nts,
	PFfs_r7_nts,
	PFfs_r5_nts,
	PFfs_r9_nts,
	PFfs_r10_nts,
	PFfs_r11_nts,
	PFfs_r12_nts,
	PFfs_r13_nts,
	PFfs_r14_nts,
	PFfs_r15_nts,
	PFfs_r16_nts,
	PFfs_r17_nts,
	PFfs_r18_nts,
	PFfs_r19_nts,
	PFfs_r20_nts,
	PFfs_r21_nts,
	PFfs_r22_nts,
	PFfs_r23_nts,
	PFfs_r24_nts,
	PFfs_r5_nts,
	PFfs_r5_nts,
	PFfs_r5_nts,
	PFfs_r5_nts,
	PFfs_r5_nts,
	PFfs_r5_nts,
	PFfs_r5_nts,
	PFfs_r5_nts,
	PFfs_r5_nts,
	PFfs_r5_nts,
	PFfs_r5_nts,
	PFfs_r36_nts,
	PFfs_r36_nts,
	PFfs_r5_nts,
	PFfs_r36_nts,
	PFfs_r5_nts,
	PFfs_r41_nts,
	PFfs_r5_nts,
	PFfs_r43_nts,
	PFfs_r44_nts,
	PFfs_r5_nts,
	PFfs_r46_nts,
	PFfs_r47_nts,
	PFfs_r48_nts,
	PFfs_r49_nts,
	PFfs_r50_nts,
	PFfs_r5_nts,
	PFfs_r52_nts,
	PFfs_r53_nts,
	PFfs_r54_nts,
	PFfs_r55_nts,
	PFfs_r56_nts,
	PFfs_r57_nts,
	PFfs_r58_nts,
	PFfs_r59_nts,
	PFfs_r60_nts,
	PFfs_r61_nts,
	PFfs_r5_nts,
	PFfs_r63_nts,
	PFfs_r64_nts,
	PFfs_r65_nts,
	PFfs_r66_nts,
	0,
	PFfs_r5_nts,
	PFfs_r69_nts,
	PFfs_r46_nts,
	PFfs_r5_nts,
	PFfs_r50_nts,
	PFfs_r5_nts,
	PFfs_r74_nts,
	PFfs_r75_nts,
	PFfs_r76_nts,
	PFfs_r75_nts,
	PFfs_r76_nts,
	PFfs_r79_nts,
	PFfs_r5_nts,
	PFfs_r69_nts,
	PFfs_r82_nts,
	PFfs_r83_nts,
	PFfs_r5_nts,
	PFfs_r85_nts,
	PFfs_r5_nts,
	PFfs_r5_nts,
	PFfs_r5_nts,
	PFfs_r5_nts,
	PFfs_r5_nts,
	PFfs_r5_nts,
	PFfs_r92_nts,
	PFfs_r5_nts,
	PFfs_r92_nts,
	PFfs_r5_nts,
	PFfs_r5_nts,
	PFfs_r97_nts,
	PFfs_r5_nts,
	PFfs_r97_nts,
	PFfs_r5_nts,
	PFfs_r101_nts,
	PFfs_r102_nts,
	PFfs_r103_nts,
	PFfs_r104_nts,
	PFfs_r105_nts,
	PFfs_r106_nts,
	PFfs_r107_nts,
	PFfs_r107_nts,
	PFfs_r107_nts,
	PFfs_r107_nts,
	PFfs_r107_nts,
	PFfs_r107_nts,
	PFfs_r107_nts,
	PFfs_r107_nts,
	PFfs_r107_nts,
	PFfs_r107_nts,
	PFfs_r107_nts,
	PFfs_r107_nts,
	PFfs_r107_nts,
	0,
	PFfs_r107_nts,
	PFfs_r107_nts,
	PFfs_r123_nts,
	PFfs_r107_nts,
	PFfs_r125_nts,
	PFfs_r126_nts,
	PFfs_r126_nts,
	PFfs_r128_nts,
	PFfs_r129_nts,
	PFfs_r129_nts,
	PFfs_r129_nts,
	PFfs_r129_nts,
	PFfs_r133_nts,
	PFfs_r134_nts,
	PFfs_r135_nts,
	PFfs_r136_nts,
	PFfs_r136_nts,
	PFfs_r138_nts,
	PFfs_r139_nts,
	PFfs_r140_nts,
	PFfs_r141_nts,
	PFfs_r142_nts,
	PFfs_r143_nts,
	PFfs_r144_nts,
	PFfs_r145_nts,
	PFfs_r146_nts,
	PFfs_r146_nts,
	PFfs_r146_nts,
	PFfs_r149_nts,
	PFfs_r150_nts,
	PFfs_r151_nts,
	PFfs_r50_nts,
	PFfs_r153_nts,
	PFfs_r154_nts,
	PFfs_r5_nts,
	PFfs_r156_nts,
	PFfs_r157_nts,
	PFfs_r158_nts,
	0,
	PFfs_r160_nts,
	PFfs_r161_nts,
	0,
	PFfs_r163_nts,
	PFfs_r69_nts,
	PFfs_r165_nts,
	PFfs_r166_nts,
	PFfs_r167_nts,
	PFfs_r168_nts,
	PFfs_r169_nts,
	PFfs_r170_nts,
	PFfs_r171_nts,
	PFfs_r172_nts,
	PFfs_r173_nts,
	PFfs_r174_nts,
	PFfs_r175_nts,
	PFfs_r176_nts,
	PFfs_r5_nts,
	PFfs_r5_nts,
	PFfs_r5_nts,
	PFfs_r5_nts,
	PFfs_r5_nts,
	PFfs_r50_nts,
	PFfs_r5_nts,
	PFfs_r5_nts,
	PFfs_r185_nts,
	PFfs_r5_nts,
	PFfs_r187_nts,
	PFfs_r50_nts,
	PFfs_r50_nts,
	PFfs_r190_nts,
	PFfs_r191_nts,
	PFfs_r192_nts,
	PFfs_r193_nts,
	PFfs_r194_nts,
	PFfs_r195_nts,
	PFfs_r196_nts,
	PFfs_r197_nts,
	PFfs_r50_nts,
	PFfs_r50_nts,
	PFfs_r200_nts,
	PFfs_r50_nts,
	PFfs_r5_nts,
	PFfs_r50_nts,
	PFfs_r52_nts,
	PFfs_r5_nts,
	PFfs_r206_nts,
	PFfs_r52_nts,
	PFfs_r5_nts,
	PFfs_r209_nts,
	PFfs_r210_nts,
	PFfs_r211_nts,
	PFfs_r5_nts,
	PFfs_r210_nts,
	PFfs_r214_nts,
	PFfs_r5_nts,
	PFfs_r216_nts,
	PFfs_r217_nts,
	PFfs_r218_nts,
	PFfs_r219_nts,
	PFfs_r220_nts,
	PFfs_r221_nts,
	PFfs_r222_nts,
	PFfs_r223_nts,
	PFfs_r224_nts,
	PFfs_r225_nts,
	PFfs_r5_nts,
	PFfs_r227_nts,
	PFfs_r47_nts,
	PFfs_r54_nts,
	PFfs_r54_nts,
	PFfs_r54_nts,
	PFfs_r5_nts,
	PFfs_r233_nts,
	PFfs_r234_nts,
	PFfs_r5_nts,
	PFfs_r46_nts,
	PFfs_r237_nts,
	PFfs_r5_nts,
};
static unsigned char PFfs_and_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_args_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_attr_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_bind_transition[5][2] = {
{    0,    0}	/* row 0 */,
{    0,    3}	/* row 1 */,
{    0,    4}	/* row 2 */,
{    0,    2}	/* row 3 */,
{    0,    1}	/* row 4 */
};
static unsigned char PFfs_binds_transition[7][3] = {
{    0,    0,    0}	/* row 0 */,
{    0,    6,    6}	/* row 1 */,
{    0,    1,    1}	/* row 2 */,
{    0,    5,    5}	/* row 3 */,
{    0,    4,    5}	/* row 4 */,
{    0,    3,    3}	/* row 5 */,
{    0,    2,    3}	/* row 6 */
};
static unsigned char PFfs_case__transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_cases_transition[3][3] = {
{    0,    0,    0}	/* row 0 */,
{    0,    2,    0}	/* row 1 */,
{    0,    0,    1}	/* row 2 */
};
static unsigned char PFfs_cast_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_castable_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_contseq_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_decl_imps_transition[4][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */,
{    0,    3}	/* row 2 */,
{    0,    2}	/* row 3 */
};
static unsigned char PFfs_default__transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_div_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_elem_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_eq_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_every_transition[3][2] = {
{    0,    0}	/* row 0 */,
{    0,    2}	/* row 1 */,
{    0,    1}	/* row 2 */
};
static unsigned char PFfs_except_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_exprseq_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_ext_expr_transition[2][3] = {
{    0,    0,    0}	/* row 0 */,
{    0,    2,    1}	/* row 1 */
};
static unsigned char PFfs_flwr_transition[2][3] = {
{    0,    0,    0}	/* row 0 */,
{    0,    2,    1}	/* row 1 */
};
static unsigned char PFfs_fun_transition[2][3] = {
{    0,    0,    0}	/* row 0 */,
{    0,    1,    2}	/* row 1 */
};
static unsigned char PFfs_fun_sig_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_ge_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_gt_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_gtgt_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_idiv_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_if__transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_instof_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_intersect_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_is_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_le_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_let_transition[3][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */,
{    0,    2}	/* row 2 */
};
static unsigned char PFfs_lib_mod_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_locpath_transition[3][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */,
{    0,    2}	/* row 2 */
};
static unsigned char PFfs_lt_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_ltlt_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_main_mod_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_minus_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_mod_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_mod_imp_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_mult_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_ne_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_or_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_ord_ret_transition[3][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */,
{    0,    2}	/* row 2 */
};
static unsigned char PFfs_orderspecs_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_param_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_params_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_pi_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_plus_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_pragmas_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_pred_transition[3][2] = {
{    0,    0}	/* row 0 */,
{    0,    2}	/* row 1 */,
{    0,    1}	/* row 2 */
};
static unsigned char PFfs_range_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_req_ty_transition[3][3] = {
{    0,    0,    0}	/* row 0 */,
{    0,    1,    4}	/* row 1 */,
{    0,    2,    3}	/* row 2 */
};
static unsigned char PFfs_schm_ats_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_schm_imp_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_some_transition[3][2] = {
{    0,    0}	/* row 0 */,
{    0,    2}	/* row 1 */,
{    0,    1}	/* row 2 */
};
static unsigned char PFfs_then_else_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_treat_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_typeswitch_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_union__transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_val_eq_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_val_ge_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_val_gt_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_val_le_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_val_lt_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_val_ne_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_var_decl_transition[3][3] = {
{    0,    0,    0}	/* row 0 */,
{    0,    1,    3}	/* row 1 */,
{    0,    2,    3}	/* row 2 */
};
static unsigned char PFfs_var_type_transition[3][3] = {
{    0,    0,    0}	/* row 0 */,
{    0,    2,    3}	/* row 1 */,
{    0,    0,    1}	/* row 2 */
};
static unsigned char PFfs_vars_transition[3][3] = {
{    0,    0,    0}	/* row 0 */,
{    0,    3,    4}	/* row 1 */,
{    0,    1,    2}	/* row 2 */
};
static unsigned char PFfs_where_transition[2][3] = {
{    0,    0,    0}	/* row 0 */,
{    0,    2,    1}	/* row 1 */
};
static unsigned char PFfs_insert_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_replace_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_rename_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_transform_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_modify_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_transbinds_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_recursion_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_seed_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static unsigned char PFfs_xrpc_transition[2][2] = {
{    0,    0}	/* row 0 */,
{    0,    1}	/* row 1 */
};
static struct {
	unsigned int f124:2;
	unsigned int f125:2;
	unsigned int f126:2;
	unsigned int f127:2;
	unsigned int f128:2;
	unsigned int f129:2;
	unsigned int f130:2;
	unsigned int f131:2;
	unsigned int f132:2;
	unsigned int f133:5;
	unsigned int f134:3;
	unsigned int f135:2;
	unsigned int f136:2;
	unsigned int f137:2;
} PFfs_plank_0[] = {
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 0 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 1 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 2 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 3 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 4 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 5 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 6 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   5,   0,   0,   0,   0,},	/* row 7 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 8 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 9 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 10 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 11 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 12 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 13 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 14 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 15 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 16 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 17 */
	{   0,   0,   0,   0,   0,   0,   0,   1,   0,   3,   0,   0,   0,   0,},	/* row 18 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 19 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 20 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 21 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 22 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 23 */
	{   0,   0,   0,   0,   0,   0,   1,   0,   0,   4,   0,   0,   0,   0,},	/* row 24 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 25 */
	{   0,   0,   0,   0,   1,   0,   0,   0,   0,   6,   0,   0,   0,   0,},	/* row 26 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 27 */
	{   1,   0,   0,   0,   0,   0,   0,   0,   0,   9,   0,   0,   0,   0,},	/* row 28 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,},	/* row 29 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,},	/* row 30 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   0,   0,   0,},	/* row 31 */
	{   0,   1,   0,   0,   0,   0,   0,   0,   0,   8,   0,   0,   0,   0,},	/* row 32 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 33 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 34 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 35 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 36 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 37 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 38 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 39 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 40 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 41 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   1,   2,   0,   0,   0,   0,},	/* row 42 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 43 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 44 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 45 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 46 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 47 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 48 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 49 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 50 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 51 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 52 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   2,   2,   0,   0,   0,   0,},	/* row 53 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 54 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 55 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 56 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 57 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 58 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 59 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 60 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 61 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 62 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 63 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 64 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 65 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 66 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 67 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 68 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 69 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,  15,   0,   1,   0,   2,},	/* row 70 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 71 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 72 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 73 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 74 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 75 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 76 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 77 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 78 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   1,},	/* row 79 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 80 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 81 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,  14,   0,   0,   0,   0,},	/* row 82 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 83 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 84 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 85 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 86 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 87 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   4,   0,   0,   0,},	/* row 88 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 89 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 90 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 91 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 92 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 93 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,  10,   0,   0,   0,   0,},	/* row 94 */
	{   0,   0,   0,   1,   0,   0,   0,   0,   0,  11,   0,   0,   0,   0,},	/* row 95 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 96 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 97 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 98 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 99 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 100 */
	{   0,   0,   1,   0,   0,   0,   0,   0,   0,   7,   0,   0,   0,   0,},	/* row 101 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 102 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 103 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 104 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 105 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 106 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 107 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 108 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 109 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 110 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 111 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 112 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 113 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 114 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 115 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 116 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 117 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 118 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 119 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,  13,   0,   0,   0,   0,},	/* row 120 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 121 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 122 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 123 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 124 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,  12,   0,   0,   0,   0,},	/* row 125 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 126 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 127 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 128 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 129 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 130 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 131 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 132 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 133 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 134 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 135 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 136 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 137 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 138 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 139 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 140 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 141 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 142 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 143 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 144 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 145 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 146 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 147 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 148 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 149 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 150 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 151 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 152 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 153 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,},	/* row 154 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,},	/* row 155 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,},	/* row 156 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 157 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 158 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 159 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 160 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 161 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 162 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 163 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 164 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 165 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 166 */
};
static struct {
	unsigned int f109:2;
	unsigned int f111:2;
	unsigned int f112:2;
	unsigned int f113:3;
	unsigned int f114:2;
	unsigned int f115:4;
	unsigned int f116:2;
	unsigned int f117:2;
	unsigned int f118:3;
	unsigned int f119:2;
	unsigned int f120:2;
	unsigned int f121:2;
	unsigned int f122:2;
	unsigned int f123:2;
} PFfs_plank_1[] = {
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 0 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 1 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 2 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 3 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 4 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 5 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 6 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 7 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 8 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 9 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 10 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 11 */
	{   0,   0,   0,   4,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 12 */
	{   0,   0,   0,   5,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 13 */
	{   0,   0,   0,   5,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 14 */
	{   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 15 */
	{   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 16 */
	{   0,   0,   0,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 17 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 18 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 19 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 20 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 21 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 22 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 23 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 24 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 25 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 26 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 27 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 28 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 29 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 30 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 31 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 32 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 33 */
	{   0,   0,   2,   0,   0,   7,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 34 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 35 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 36 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 37 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 38 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 39 */
	{   0,   0,   2,   0,   0,   5,   1,   2,   0,   0,   0,   0,   0,   0,},	/* row 40 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 41 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 42 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 43 */
	{   0,   0,   2,   0,   0,   2,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 44 */
	{   0,   0,   2,   0,   0,   2,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 45 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 46 */
	{   0,   0,   2,   0,   0,   5,   2,   1,   0,   0,   0,   0,   0,   0,},	/* row 47 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 48 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 49 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 50 */
	{   0,   0,   2,   0,   1,   3,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 51 */
	{   0,   0,   2,   0,   2,   3,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 52 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 53 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 54 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 55 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 56 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 57 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 58 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 59 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 60 */
	{   0,   0,   2,   0,   0,   4,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 61 */
	{   0,   0,   2,   0,   0,   6,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 62 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 63 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 64 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 65 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 66 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 67 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 68 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 69 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 70 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 71 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 72 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 73 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 74 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 75 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 76 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 77 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 78 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 79 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 80 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 81 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 82 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 83 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 84 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 85 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 86 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 87 */
	{   0,   1,   1,   2,   0,   0,   0,   0,   0,   1,   2,   0,   0,   0,},	/* row 88 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 89 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 90 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 91 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 92 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 93 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,},	/* row 94 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 95 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 96 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 97 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 98 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 99 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 100 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 101 */
	{   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 102 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 103 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,},	/* row 104 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 105 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 106 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 107 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 108 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 109 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 110 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 111 */
	{   0,   0,   2,   0,   0,  11,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 112 */
	{   0,   0,   2,   0,   0,   8,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 113 */
	{   0,   0,   2,   0,   0,   9,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 114 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 115 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 116 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 117 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 118 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 119 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 120 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 121 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,},	/* row 122 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 123 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 124 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,},	/* row 125 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 126 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 127 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 128 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 129 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 130 */
	{   0,   0,   2,   0,   0,   2,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 131 */
	{   0,   0,   2,   0,   0,   2,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 132 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 133 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 134 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 135 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 136 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 137 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 138 */
	{   0,   0,   2,   0,   0,  10,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 139 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 140 */
	{   1,   0,   2,   0,   0,   1,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 141 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 142 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 143 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 144 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 145 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 146 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 147 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 148 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 149 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 150 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 151 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 152 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 153 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0,},	/* row 154 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,},	/* row 155 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,},	/* row 156 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 157 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 158 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 159 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 160 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 161 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 162 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 163 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 164 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 165 */
	{   0,   0,   2,   0,   0,   5,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 166 */
};
static struct {
	unsigned int f98:2;
	unsigned int f99:5;
	unsigned int f100:2;
	unsigned int f101:2;
	unsigned int f102:2;
	unsigned int f103:2;
	unsigned int f104:3;
	unsigned int f105:3;
	unsigned int f106:2;
	unsigned int f107:3;
	unsigned int f108:3;
	unsigned int f110:3;
} PFfs_plank_2[] = {
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 0 */
	{   1,   4,   2,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 1 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 2 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 3 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 4 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   1,   3,   0,   0,},	/* row 5 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 6 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 7 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 8 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 9 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 10 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 11 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 12 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 13 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 14 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 15 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 16 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 17 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 18 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 19 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 20 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 21 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 22 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 23 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 24 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 25 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 26 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 27 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 28 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 29 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 30 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 31 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 32 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 33 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 34 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 35 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 36 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 37 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 38 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 39 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 40 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   6,   0,   0,},	/* row 41 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 42 */
	{   1,   3,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 43 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   4,},	/* row 44 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   1,},	/* row 45 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 46 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 47 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 48 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 49 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 50 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 51 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 52 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 53 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 54 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 55 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 56 */
	{   1,  14,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 57 */
	{   1,  15,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 58 */
	{   1,   9,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 59 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 60 */
	{   1,   4,   1,   2,   1,   0,   0,   0,   0,   0,   0,   0,},	/* row 61 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 62 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 63 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 64 */
	{   1,   5,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 65 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   4,   0,   0,},	/* row 66 */
	{   1,  16,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 67 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 68 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 69 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 70 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 71 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 72 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 73 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 74 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 75 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 76 */
	{   1,   1,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 77 */
	{   1,  12,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 78 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 79 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 80 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 81 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 82 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 83 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 84 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 85 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,},	/* row 86 */
	{   1,   2,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 87 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 88 */
	{   0,   0,   0,   0,   0,   0,   0,   5,   0,   5,   0,   0,},	/* row 89 */
	{   0,   0,   0,   0,   0,   0,   0,   2,   0,   5,   0,   0,},	/* row 90 */
	{   0,   0,   0,   0,   0,   0,   0,   3,   0,   5,   0,   0,},	/* row 91 */
	{   0,   0,   0,   0,   0,   0,   0,   4,   0,   5,   0,   0,},	/* row 92 */
	{   0,   0,   0,   0,   0,   0,   0,   1,   0,   5,   0,   0,},	/* row 93 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 94 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 95 */
	{   1,   4,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 96 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 97 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 98 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 99 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 100 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 101 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 102 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 103 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 104 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 105 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 106 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 107 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 108 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 109 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 110 */
	{   2,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 111 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 112 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 113 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 114 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 115 */
	{   0,   0,   0,   0,   0,   0,   4,   0,   0,   0,   0,   0,},	/* row 116 */
	{   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,},	/* row 117 */
	{   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,},	/* row 118 */
	{   0,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0,},	/* row 119 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 120 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 121 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 122 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 123 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,},	/* row 124 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 125 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 126 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2,   0,},	/* row 127 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,},	/* row 128 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   0,},	/* row 129 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2,   0,},	/* row 130 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   2,},	/* row 131 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   3,},	/* row 132 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 133 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 134 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 135 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 136 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 137 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 138 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 139 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 140 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 141 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 142 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 143 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 144 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 145 */
	{   1,  13,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 146 */
	{   1,   6,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 147 */
	{   1,   7,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 148 */
	{   1,   8,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 149 */
	{   1,  10,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 150 */
	{   1,  11,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 151 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 152 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 153 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 154 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 155 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 156 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 157 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 158 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 159 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 160 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 161 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 162 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 163 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 164 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 165 */
	{   1,   4,   1,   2,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 166 */
};
static struct {
	unsigned int f85:2;
	unsigned int f86:3;
	unsigned int f87:2;
	unsigned int f88:3;
	unsigned int f89:3;
	unsigned int f90:2;
	unsigned int f91:2;
	unsigned int f92:2;
	unsigned int f93:2;
	unsigned int f94:3;
	unsigned int f95:2;
	unsigned int f96:3;
	unsigned int f97:3;
} PFfs_plank_3[] = {
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 0 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 1 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 2 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 3 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 4 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 5 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 6 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 7 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 8 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 9 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 10 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 11 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 12 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 13 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 14 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 15 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 16 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 17 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 18 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 19 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 20 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 21 */
	{   2,   0,   0,   2,   3,   2,   2,   2,   1,   2,   1,   2,   3,},	/* row 22 */
	{   2,   0,   0,   2,   3,   1,   1,   2,   1,   2,   1,   2,   3,},	/* row 23 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 24 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 25 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 26 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 27 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 28 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 29 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 30 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 31 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 32 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 33 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 34 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   5,   3,},	/* row 35 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 36 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 37 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 38 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 39 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 40 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 41 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 42 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 43 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 44 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 45 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   1,   1,   2,   3,},	/* row 46 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 47 */
	{   2,   0,   0,   3,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 48 */
	{   2,   0,   0,   3,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 49 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 50 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 51 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 52 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 53 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 54 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 55 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 56 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 57 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 58 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 59 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   3,   3,},	/* row 60 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 61 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 62 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   2,   2,   1,   2,   3,},	/* row 63 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   3,   1,   2,   3,},	/* row 64 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 65 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 66 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 67 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 68 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 69 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 70 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 71 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 72 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 73 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 74 */
	{   2,   4,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 75 */
	{   2,   1,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 76 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 77 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 78 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 79 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   1,},	/* row 80 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   4,   3,},	/* row 81 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 82 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 83 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 84 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   1,   3,},	/* row 85 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 86 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 87 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 88 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 89 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 90 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 91 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 92 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 93 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 94 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 95 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 96 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 97 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 98 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 99 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 100 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 101 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 102 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 103 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 104 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 105 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   2,},	/* row 106 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 107 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 108 */
	{   1,   2,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 109 */
	{   1,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 110 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 111 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 112 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 113 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 114 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 115 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 116 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 117 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 118 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 119 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 120 */
	{   2,   3,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 121 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 122 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 123 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 124 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 125 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 126 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 127 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 128 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 129 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 130 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 131 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 132 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 133 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 134 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 135 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 136 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 137 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 138 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 139 */
	{   2,   0,   0,   2,   3,   1,   2,   1,   1,   2,   1,   2,   3,},	/* row 140 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 141 */
	{   2,   0,   0,   2,   1,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 142 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   2,   2,   3,},	/* row 143 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 144 */
	{   2,   0,   0,   2,   2,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 145 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 146 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 147 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 148 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 149 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 150 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 151 */
	{   2,   0,   1,   1,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 152 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 153 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 154 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 155 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 156 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 157 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 158 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 159 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 160 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 161 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 162 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 163 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 164 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 165 */
	{   2,   0,   0,   2,   3,   1,   2,   2,   1,   2,   1,   2,   3,},	/* row 166 */
};
static struct {
	unsigned int f71:2;
	unsigned int f72:2;
	unsigned int f73:2;
	unsigned int f74:3;
	unsigned int f75:2;
	unsigned int f76:2;
	unsigned int f77:2;
	unsigned int f78:2;
	unsigned int f79:2;
	unsigned int f80:2;
	unsigned int f81:2;
	unsigned int f82:3;
	unsigned int f83:2;
	unsigned int f84:4;
} PFfs_plank_4[] = {
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 0 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 1 */
	{   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   6,},	/* row 2 */
	{   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   6,},	/* row 3 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 4 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 5 */
	{   0,   1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   5,},	/* row 6 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 7 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 8 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 9 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 10 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 11 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 12 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 13 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 14 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 15 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 16 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 17 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 18 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 19 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 20 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 21 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 22 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 23 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 24 */
	{   0,   0,   0,   5,   0,   0,   0,   0,   0,   0,   0,   0,   0,   5,},	/* row 25 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 26 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 27 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 28 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 29 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 30 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 31 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 32 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 33 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 34 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 35 */
	{   0,   0,   0,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   5,},	/* row 36 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 37 */
	{   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   7,},	/* row 38 */
	{   0,   0,   1,   4,   0,   0,   0,   0,   0,   0,   0,   0,   0,   5,},	/* row 39 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 40 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 41 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 42 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 43 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 44 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 45 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 46 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 47 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 48 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 49 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 50 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 51 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 52 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 53 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 54 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 55 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 56 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 57 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 58 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 59 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 60 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 61 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 62 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 63 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 64 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 65 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 66 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 67 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 68 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 69 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 70 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   3,   1,   9,},	/* row 71 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   1,   1,   9,},	/* row 72 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   2,   1,   9,},	/* row 73 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2,   9,},	/* row 74 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 75 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 76 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 77 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 78 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 79 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 80 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 81 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 82 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 83 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 84 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 85 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 86 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 87 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 88 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 89 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 90 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 91 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 92 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 93 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 94 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 95 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 96 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 97 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 98 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 99 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   4,},	/* row 100 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 101 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 102 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 103 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 104 */
	{   0,   0,   0,   6,   0,   0,   0,   0,   0,   0,   0,   0,   0,   5,},	/* row 105 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 106 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 107 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 108 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 109 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 110 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 111 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 112 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 113 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 114 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 115 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 116 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 117 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 118 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 119 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 120 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 121 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 122 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 123 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 124 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 125 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 126 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 127 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 128 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 129 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 130 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 131 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 132 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 133 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 134 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 135 */
	{   1,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   5,},	/* row 136 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 137 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 138 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 139 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 140 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 141 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 142 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 143 */
	{   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   3,},	/* row 144 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 145 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 146 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 147 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 148 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 149 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 150 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 151 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   8,},	/* row 152 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,},	/* row 153 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 154 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 155 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 156 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 157 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 158 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 159 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 160 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 161 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 162 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 163 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 164 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 165 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2,},	/* row 166 */
};
static struct {
	unsigned int f56:2;
	unsigned int f57:2;
	unsigned int f58:2;
	unsigned int f59:2;
	unsigned int f60:2;
	unsigned int f61:2;
	unsigned int f62:2;
	unsigned int f63:2;
	unsigned int f64:2;
	unsigned int f65:3;
	unsigned int f66:3;
	unsigned int f67:2;
	unsigned int f68:2;
	unsigned int f69:2;
	unsigned int f70:2;
} PFfs_plank_5[] = {
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 0 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 1 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 2 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 3 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 4 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 5 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 6 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 7 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 8 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 9 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 10 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 11 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 12 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 13 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 14 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 15 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 16 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 17 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 18 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 19 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 20 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 21 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 22 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 23 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 24 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   1,},	/* row 25 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 26 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   3,   0,   0,   0,   0,},	/* row 27 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 28 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 29 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 30 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 31 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 32 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 33 */
	{   0,   0,   0,   0,   1,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 34 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 35 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   1,   0,   0,},	/* row 36 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 37 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 38 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 39 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   2,   1,   1,   0,   0,   0,},	/* row 40 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 41 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 42 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 43 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 44 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 45 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 46 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 47 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   2,   3,   2,   1,   0,   0,   0,},	/* row 48 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   1,   3,   2,   1,   0,   0,   0,},	/* row 49 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 50 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 51 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 52 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 53 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 54 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 55 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 56 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 57 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 58 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 59 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 60 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 61 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 62 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 63 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 64 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 65 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 66 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 67 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 68 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 69 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 70 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 71 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 72 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 73 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 74 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 75 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 76 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 77 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 78 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 79 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 80 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 81 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 82 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 83 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 84 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 85 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 86 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 87 */
	{   1,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,},	/* row 88 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 89 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 90 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 91 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 92 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 93 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 94 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 95 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 96 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 97 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 98 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 99 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 100 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 101 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 102 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 103 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 104 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   1,   0,},	/* row 105 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 106 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 107 */
	{   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,},	/* row 108 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 109 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 110 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 111 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 112 */
	{   0,   0,   1,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 113 */
	{   0,   0,   0,   1,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 114 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 115 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 116 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 117 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 118 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 119 */
	{   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 120 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 121 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 122 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 123 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 124 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 125 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 126 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 127 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 128 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 129 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 130 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 131 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 132 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 133 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 134 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,},	/* row 135 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 136 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 137 */
	{   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 138 */
	{   0,   1,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 139 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 140 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 141 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 142 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 143 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 144 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 145 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 146 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 147 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 148 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 149 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 150 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 151 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 152 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 153 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 154 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 155 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 156 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 157 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 158 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 159 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 160 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 161 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 162 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 163 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 164 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 165 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   2,   1,   0,   0,   0,},	/* row 166 */
};
static struct {
	unsigned int f40:3;
	unsigned int f41:1;
	unsigned int f42:2;
	unsigned int f43:2;
	unsigned int f44:2;
	unsigned int f45:2;
	unsigned int f46:2;
	unsigned int f47:2;
	unsigned int f48:2;
	unsigned int f49:2;
	unsigned int f50:2;
	unsigned int f51:2;
	unsigned int f52:2;
	unsigned int f53:2;
	unsigned int f54:2;
	unsigned int f55:2;
} PFfs_plank_6[] = {
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 0 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 1 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,},	/* row 2 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,},	/* row 3 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 4 */
	{   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 5 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 6 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 7 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 8 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 9 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 10 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 11 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 12 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 13 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 14 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 15 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 16 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 17 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 18 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 19 */
	{   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 20 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 21 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 22 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 23 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 24 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 25 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 26 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 27 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 28 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 29 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 30 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 31 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 32 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 33 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 34 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 35 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 36 */
	{   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 37 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 38 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 39 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 40 */
	{   4,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 41 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 42 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 43 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 44 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 45 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 46 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 47 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 48 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 49 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 50 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 51 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 52 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 53 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 54 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 55 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 56 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 57 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 58 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 59 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 60 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 61 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 62 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 63 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 64 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 65 */
	{   4,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 66 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 67 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,},	/* row 68 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 69 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 70 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 71 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 72 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 73 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 74 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 75 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 76 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 77 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 78 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 79 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 80 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 81 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 82 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 83 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,},	/* row 84 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 85 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 86 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 87 */
	{   0,   0,   0,   2,   1,   1,   0,   1,   0,   0,   0,   0,   0,   0,   2,   0,},	/* row 88 */
	{   4,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 89 */
	{   4,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 90 */
	{   4,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 91 */
	{   4,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 92 */
	{   4,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 93 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 94 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 95 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 96 */
	{   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 97 */
	{   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 98 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 99 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 100 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 101 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 102 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 103 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 104 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 105 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 106 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 107 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 108 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 109 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 110 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 111 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,},	/* row 112 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 113 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 114 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 115 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 116 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 117 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 118 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 119 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 120 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 121 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 122 */
	{   4,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 123 */
	{   4,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 124 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 125 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,},	/* row 126 */
	{   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,},	/* row 127 */
	{   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,},	/* row 128 */
	{   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,},	/* row 129 */
	{   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,},	/* row 130 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 131 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 132 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 133 */
	{   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 134 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 135 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 136 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 137 */
	{   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 138 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 139 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 140 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 141 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 142 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 143 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 144 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 145 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 146 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 147 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 148 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 149 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 150 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 151 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 152 */
	{   0,   0,   0,   1,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 153 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 154 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 155 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 156 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 157 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,},	/* row 158 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,},	/* row 159 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 160 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 161 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 162 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 163 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 164 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 165 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 166 */
};
static struct {
	unsigned int f3:1;
	unsigned int f25:2;
	unsigned int f26:2;
	unsigned int f27:2;
	unsigned int f28:2;
	unsigned int f29:2;
	unsigned int f30:2;
	unsigned int f31:3;
	unsigned int f32:2;
	unsigned int f33:2;
	unsigned int f34:2;
	unsigned int f35:2;
	unsigned int f36:2;
	unsigned int f37:2;
	unsigned int f38:2;
	unsigned int f39:2;
} PFfs_plank_7[] = {
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 0 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 1 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 2 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 3 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 4 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 5 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 6 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 7 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 8 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 9 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 10 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 11 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 12 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 13 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 14 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 15 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 16 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 17 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 18 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 19 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 20 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 21 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 22 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 23 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 24 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 25 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 26 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 27 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 28 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 29 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 30 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 31 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 32 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 33 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 34 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 35 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 36 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 37 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 38 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 39 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 40 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 41 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 42 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 43 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 44 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 45 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 46 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 47 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 48 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 49 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 50 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 51 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 52 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 53 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 54 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 55 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 56 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 57 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 58 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 59 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 60 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 61 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 62 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 63 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 64 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 65 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 66 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 67 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 68 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 69 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 70 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 71 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 72 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 73 */
	{   0,   0,   0,   0,   0,   1,   0,   1,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 74 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,},	/* row 75 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,},	/* row 76 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 77 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 78 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 79 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 80 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 81 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 82 */
	{   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 83 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 84 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 85 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,},	/* row 86 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 87 */
	{   1,   0,   0,   0,   0,   0,   1,   2,   1,   1,   0,   0,   0,   0,   2,   2,},	/* row 88 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 89 */
	{   0,   0,   0,   0,   0,   0,   0,   4,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 90 */
	{   0,   0,   0,   0,   0,   0,   0,   5,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 91 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 92 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 93 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 94 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 95 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 96 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 97 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 98 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,},	/* row 99 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 100 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 101 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,},	/* row 102 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,},	/* row 103 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 104 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 105 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 106 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,},	/* row 107 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 108 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,},	/* row 109 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 110 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 111 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 112 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 113 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 114 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,},	/* row 115 */
	{   0,   0,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 116 */
	{   0,   0,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 117 */
	{   0,   0,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 118 */
	{   0,   0,   0,   0,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 119 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 120 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,},	/* row 121 */
	{   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 122 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 123 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 124 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 125 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 126 */
	{   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 127 */
	{   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 128 */
	{   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 129 */
	{   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 130 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 131 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 132 */
	{   0,   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 133 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 134 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 135 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 136 */
	{   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 137 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 138 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 139 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 140 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 141 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 142 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 143 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 144 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 145 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 146 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 147 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 148 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 149 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 150 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 151 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 152 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   1,   0,   0,   1,   0,   0,},	/* row 153 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 154 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 155 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 156 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 157 */
	{   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 158 */
	{   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 159 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 160 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 161 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 162 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 163 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 164 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 165 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 166 */
};
static struct {
	unsigned int f10:2;
	unsigned int f11:2;
	unsigned int f12:2;
	unsigned int f13:2;
	unsigned int f14:3;
	unsigned int f15:2;
	unsigned int f16:2;
	unsigned int f17:2;
	unsigned int f18:2;
	unsigned int f19:2;
	unsigned int f20:2;
	unsigned int f21:2;
	unsigned int f22:2;
	unsigned int f23:2;
	unsigned int f24:2;
} PFfs_plank_8[] = {
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 0 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 1 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 2 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 3 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 4 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 5 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 6 */
	{   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 7 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 8 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 9 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 10 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 11 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,},	/* row 12 */
	{   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   1,   0,   0,   0,   0,},	/* row 13 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,},	/* row 14 */
	{   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0,   0,},	/* row 15 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,},	/* row 16 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,},	/* row 17 */
	{   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 18 */
	{   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 19 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 20 */
	{   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 21 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 22 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 23 */
	{   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 24 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 25 */
	{   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 26 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 27 */
	{   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 28 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 29 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 30 */
	{   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 31 */
	{   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 32 */
	{   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 33 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 34 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 35 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 36 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 37 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 38 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 39 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 40 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 41 */
	{   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 42 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 43 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 44 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 45 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 46 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 47 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 48 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 49 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2,   0,},	/* row 50 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 51 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 52 */
	{   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 53 */
	{   0,   0,   0,   0,   2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 54 */
	{   0,   0,   0,   0,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 55 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 56 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 57 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 58 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 59 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 60 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 61 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 62 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 63 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 64 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 65 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 66 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 67 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 68 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 69 */
	{   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 70 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 71 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 72 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 73 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 74 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 75 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 76 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 77 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 78 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 79 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 80 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 81 */
	{   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 82 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 83 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 84 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 85 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 86 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 87 */
	{   0,   0,   1,   0,   0,   1,   1,   0,   1,   2,   1,   0,   0,   0,   1,},	/* row 88 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 89 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 90 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 91 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 92 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 93 */
	{   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 94 */
	{   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 95 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 96 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 97 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 98 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 99 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 100 */
	{   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 101 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 102 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 103 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,},	/* row 104 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 105 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 106 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 107 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,},	/* row 108 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 109 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 110 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 111 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 112 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 113 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 114 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 115 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 116 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 117 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 118 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 119 */
	{   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 120 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 121 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 122 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 123 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 124 */
	{   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 125 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 126 */
	{   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 127 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 128 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 129 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 130 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 131 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 132 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 133 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 134 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 135 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 136 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 137 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 138 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 139 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 140 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 141 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 142 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 143 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 144 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 145 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 146 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 147 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 148 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 149 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 150 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 151 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 152 */
	{   0,   0,   0,   0,   0,   0,   1,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 153 */
	{   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 154 */
	{   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 155 */
	{   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 156 */
	{   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 157 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 158 */
	{   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 159 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 160 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 161 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 162 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 163 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   2,   0,   0,   0,},	/* row 164 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,},	/* row 165 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,   1,   0,   0,   0,   1,   0,},	/* row 166 */
};
static struct {
	unsigned int f0:2;
	unsigned int f1:2;
	unsigned int f2:2;
	unsigned int f4:2;
	unsigned int f5:2;
	unsigned int f6:2;
	unsigned int f7:3;
	unsigned int f8:3;
	unsigned int f9:2;
} PFfs_plank_9[] = {
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 0 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 1 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 2 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 3 */
	{   0,   1,   1,   0,   0,   0,   0,   0,   0,},	/* row 4 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 5 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 6 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 7 */
	{   0,   0,   0,   0,   0,   0,   0,   6,   0,},	/* row 8 */
	{   0,   0,   0,   0,   0,   0,   0,   5,   0,},	/* row 9 */
	{   0,   0,   0,   0,   0,   0,   0,   3,   0,},	/* row 10 */
	{   0,   0,   0,   0,   0,   0,   0,   4,   0,},	/* row 11 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   2,},	/* row 12 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   2,},	/* row 13 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   2,},	/* row 14 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   2,},	/* row 15 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   2,},	/* row 16 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   2,},	/* row 17 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 18 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 19 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 20 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 21 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 22 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 23 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 24 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 25 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 26 */
	{   0,   0,   0,   0,   1,   0,   0,   0,   0,},	/* row 27 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 28 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 29 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 30 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 31 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 32 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 33 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 34 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 35 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 36 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 37 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 38 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 39 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 40 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 41 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 42 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 43 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 44 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 45 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 46 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 47 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 48 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 49 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 50 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 51 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 52 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 53 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 54 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 55 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 56 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 57 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 58 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 59 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 60 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 61 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 62 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 63 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 64 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 65 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 66 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 67 */
	{   0,   0,   0,   0,   0,   0,   0,   1,   0,},	/* row 68 */
	{   0,   0,   0,   0,   0,   0,   0,   2,   0,},	/* row 69 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 70 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 71 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 72 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 73 */
	{   1,   0,   0,   1,   1,   1,   0,   0,   0,},	/* row 74 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 75 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 76 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 77 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 78 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 79 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 80 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 81 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 82 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 83 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 84 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 85 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 86 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 87 */
	{   0,   2,   1,   0,   0,   0,   0,   0,   1,},	/* row 88 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 89 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 90 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 91 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 92 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 93 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 94 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 95 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 96 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 97 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 98 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 99 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 100 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 101 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 102 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 103 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 104 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 105 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 106 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 107 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 108 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 109 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 110 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 111 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 112 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 113 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 114 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 115 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 116 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 117 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 118 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 119 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 120 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 121 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 122 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 123 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 124 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 125 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 126 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 127 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 128 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 129 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 130 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 131 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 132 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 133 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 134 */
	{   0,   0,   0,   1,   0,   0,   0,   0,   0,},	/* row 135 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 136 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 137 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 138 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 139 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 140 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 141 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 142 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 143 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 144 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 145 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 146 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 147 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 148 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 149 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 150 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 151 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 152 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 153 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 154 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 155 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 156 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 157 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 158 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 159 */
	{   0,   0,   0,   0,   0,   0,   3,   0,   0,},	/* row 160 */
	{   0,   0,   0,   0,   0,   0,   4,   0,   0,},	/* row 161 */
	{   0,   0,   0,   0,   0,   0,   1,   0,   0,},	/* row 162 */
	{   0,   0,   0,   0,   0,   0,   2,   0,   0,},	/* row 163 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 164 */
	{   0,   0,   0,   0,   0,   0,   0,   0,   0,},	/* row 165 */
	{   1,   0,   0,   1,   1,   0,   0,   0,   0,},	/* row 166 */
};
static short PFfs_eruleMap[] = {
    0,  127,  126,  125,    0,  104,  105,    0,   90,    0,	/* 0-9 */
  197,    0,  209,  208,  207,    0,   29,    0,   27,    0,	/* 10-19 */
  144,  145,    0,  143,  142,    0,  199,    0,  109,  108,	/* 20-29 */
  107,  106,  119,  118,  117,  116,  122,  115,  114,  121,	/* 30-39 */
  113,  112,  111,  110,    0,   30,    0,  191,  192,  195,	/* 40-49 */
  190,  193,  194,    0,  183,    0,   34,    0,  178,    0,	/* 50-59 */
   23,   11,   12,   13,   14,   15,   16,   17,   18,   19,	/* 60-69 */
   20,   21,  214,   22,   24,    0,    7,   10,    9,    8,	/* 70-79 */
    0,   28,    0,   25,   26,    0,  222,    0,  201,    0,	/* 80-89 */
  179,    0,   97,   96,   98,   99,    0,  196,    0,   33,	/* 90-99 */
    0,   54,   53,    0,   57,   56,   55,   58,   59,  216,	/* 100-109 */
  217,  218,  219,  220,  233,    0,  211,  210,    0,   61,	/* 110-119 */
   60,    0,  161,  160,    0,  187,  186,    0,  184,  185,	/* 120-129 */
    0,  101,    0,  221,    0,  138,  139,    0,  177,    0,	/* 130-139 */
  137,  135,  136,    0,   89,   88,   87,   86,   85,   84,	/* 140-149 */
    0,    4,    0,  172,  173,    0,  157,  158,  155,  156,	/* 150-159 */
    0,    3,    0,    5,    0,   37,    0,  129,  128,  131,	/* 160-169 */
  132,  130,    0,   35,    0,  231,    0,   94,   91,   92,	/* 170-179 */
   93,   95,    0,  175,  174,  176,    0,   65,   62,   63,	/* 180-189 */
   64,   66,    0,   81,   80,    0,  205,  204,  206,    0,	/* 190-199 */
   52,   51,    0,   39,   38,    0,   40,   41,    0,   69,	/* 200-209 */
   68,    0,   45,   46,    0,  236,  235,    0,   71,   72,	/* 210-219 */
    0,   31,    0,  103,  102,    0,   73,   74,    0,  188,	/* 220-229 */
    0,   32,    0,  200,    0,   44,    0,   42,   43,    0,	/* 230-239 */
  182,    0,  153,  154,    0,  213,  212,    0,  164,  171,	/* 240-249 */
  170,  169,  168,  167,  166,  165,  163,    0,    6,    0,	/* 250-259 */
   77,   76,   75,   78,    0,    1,    2,    0,   50,    0,	/* 260-269 */
  123,  124,    0,  234,    0,  224,    0,  223,    0,  215,	/* 270-279 */
    0,   36,    0,  100,    0,  232,   82,  238,    0,   83,	/* 280-289 */
    0,  229,    0,  180,    0,  203,  202,    0,  230,    0,	/* 290-299 */
  198,    0,  228,    0,  226,  227,    0,  225,    0,  141,	/* 300-309 */
  140,    0,   70,    0,   79,    0,  147,  148,  146,    0,	/* 310-319 */
  133,  134,    0,  189,    0,  152,    0,  149,  150,  151,	/* 320-329 */
    0,  181,    0,   49,   48,   47,    0,  237
};
#define PFfs_Query_rule(state)	PFfs_eruleMap[PFfs_plank_0[state].f137 +264]
#define PFfs_MainModule_rule(state)	PFfs_eruleMap[PFfs_plank_0[state].f136 +160]
#define PFfs_LibraryModule_rule(state)	PFfs_eruleMap[PFfs_plank_0[state].f135 +150]
#define PFfs_ModuleDecl_rule(state)	PFfs_eruleMap[PFfs_plank_7[state].f28 +162]
#define PFfs_Prolog_rule(state)	PFfs_eruleMap[PFfs_plank_8[state].f15 +257]
#define PFfs_DeclsImports_rule(state)	PFfs_eruleMap[PFfs_plank_0[state].f134 +75]
#define PFfs_DeclImport_rule(state)	PFfs_eruleMap[PFfs_plank_0[state].f133 +59]
#define PFfs_DefaultNamespaceDecl_rule(state)	PFfs_eruleMap[PFfs_plank_0[state].f132 +82]
#define PFfs_BoundarySpaceDecl_rule(state)	PFfs_eruleMap[PFfs_plank_0[state].f131 +17]
#define PFfs_DefaultCollationDecl_rule(state)	PFfs_eruleMap[PFfs_plank_0[state].f130 +80]
#define PFfs_BaseURIDecl_rule(state)	PFfs_eruleMap[PFfs_plank_0[state].f129 +15]
#define PFfs_ConstructionDecl_rule(state)	PFfs_eruleMap[PFfs_plank_0[state].f128 +44]
#define PFfs_OptionDecl_rule(state)	PFfs_eruleMap[PFfs_plank_0[state].f127 +220]
#define PFfs_OrderingModeDecl_rule(state)	PFfs_eruleMap[PFfs_plank_0[state].f126 +230]
#define PFfs_EmptyOrderingDecl_rule(state)	PFfs_eruleMap[PFfs_plank_0[state].f125 +98]
#define PFfs_CopyNamespacesDecl_rule(state)	PFfs_eruleMap[PFfs_plank_0[state].f124 +55]
#define PFfs_NamespaceDecl_rule(state)	PFfs_eruleMap[PFfs_plank_1[state].f123 +172]
#define PFfs_SchemaImport_rule(state)	PFfs_eruleMap[PFfs_plank_1[state].f122 +280]
#define PFfs_ModuleImport_rule(state)	PFfs_eruleMap[PFfs_plank_1[state].f121 +164]
#define PFfs_OptImportLoc__rule(state)	PFfs_eruleMap[PFfs_plank_1[state].f120 +202]
#define PFfs_OptParamList__rule(state)	PFfs_eruleMap[PFfs_plank_1[state].f119 +205]
#define PFfs_ParamList__rule(state)	PFfs_eruleMap[PFfs_plank_1[state].f119 +236]
#define PFfs_Param_rule(state)	PFfs_eruleMap[PFfs_plank_7[state].f35 +234]
#define PFfs_OptSequenceType__rule(state)	PFfs_eruleMap[PFfs_plank_6[state].f44 +211]
#define PFfs_VariableDeclaration_rule(state)	PFfs_eruleMap[PFfs_plank_1[state].f118 +332]
#define PFfs_QueryBody_rule(state)	PFfs_eruleMap[PFfs_plank_9[state].f0 +267]
#define PFfs_OptExpr_rule(state)	PFfs_eruleMap[PFfs_plank_1[state].f117 +199]
#define PFfs_Expr_rule(state)	PFfs_eruleMap[PFfs_plank_1[state].f116 +100]
#define PFfs_ExprSingle_rule(state)	PFfs_eruleMap[PFfs_plank_1[state].f115 +103]
#define PFfs_FLWORExpr_rule(state)	PFfs_eruleMap[PFfs_plank_1[state].f114 +118]
#define PFfs_OptBindExpr_rule(state)	PFfs_eruleMap[PFfs_plank_1[state].f113 +186]
#define PFfs_OptPosVar_rule(state)	PFfs_eruleMap[PFfs_plank_6[state].f43 +208]
#define PFfs_TypeDeclaration_rule(state)	PFfs_eruleMap[PFfs_plank_7[state].f25 +311]
#define PFfs_OptWhereClause__rule(state)	PFfs_eruleMap[PFfs_plank_1[state].f112 +217]
#define PFfs_OrderSpecList_rule(state)	PFfs_eruleMap[PFfs_plank_1[state].f111 +225]
#define PFfs_QuantifiedExpr_rule(state)	PFfs_eruleMap[PFfs_plank_2[state].f110 +259]
#define PFfs_TypeswitchExpr_rule(state)	PFfs_eruleMap[PFfs_plank_1[state].f109 +313]
#define PFfs_OptCaseVar_rule(state)	PFfs_eruleMap[PFfs_plank_6[state].f43 +192]
#define PFfs_SequenceType_rule(state)	PFfs_eruleMap[PFfs_plank_2[state].f108 +284]
#define PFfs_SingleType_rule(state)	PFfs_eruleMap[PFfs_plank_8[state].f13 +288]
#define PFfs_ItemType_rule(state)	PFfs_eruleMap[PFfs_plank_2[state].f107 +143]
#define PFfs_AtomType_rule(state)	PFfs_eruleMap[PFfs_plank_2[state].f106 +7]
#define PFfs_NodeTest_rule(state)	PFfs_eruleMap[PFfs_plank_2[state].f105 +176]
#define PFfs_ElemOrAttrType_rule(state)	PFfs_eruleMap[PFfs_plank_2[state].f104 +91]
#define PFfs_SchemaType_rule(state)	PFfs_eruleMap[PFfs_plank_2[state].f103 +282]
#define PFfs_IfExpr_rule(state)	PFfs_eruleMap[PFfs_plank_2[state].f102 +130]
#define PFfs_OrExpr_rule(state)	PFfs_eruleMap[PFfs_plank_2[state].f101 +222]
#define PFfs_AndExpr_rule(state)	PFfs_eruleMap[PFfs_plank_2[state].f100 +4]
#define PFfs_ComparisonExpr_rule(state)	PFfs_eruleMap[PFfs_plank_2[state].f99 +27]
#define PFfs_RangeExpr_rule(state)	PFfs_eruleMap[PFfs_plank_2[state].f98 +269]
#define PFfs_AdditiveExpr_rule(state)	PFfs_eruleMap[PFfs_plank_3[state].f97 +0]
#define PFfs_MultiplicativeExpr_rule(state)	PFfs_eruleMap[PFfs_plank_3[state].f96 +166]
#define PFfs_UnionExpr_rule(state)	PFfs_eruleMap[PFfs_plank_3[state].f95 +319]
#define PFfs_IntersectExceptExpr_rule(state)	PFfs_eruleMap[PFfs_plank_3[state].f94 +139]
#define PFfs_InstanceofExpr_rule(state)	PFfs_eruleMap[PFfs_plank_3[state].f93 +134]
#define PFfs_TreatExpr_rule(state)	PFfs_eruleMap[PFfs_plank_3[state].f92 +308]
#define PFfs_CastableExpr_rule(state)	PFfs_eruleMap[PFfs_plank_3[state].f91 +22]
#define PFfs_CastExpr_rule(state)	PFfs_eruleMap[PFfs_plank_3[state].f90 +19]
#define PFfs_UnaryExpr_rule(state)	PFfs_eruleMap[PFfs_plank_3[state].f89 +315]
#define PFfs_ValueExpr_rule(state)	PFfs_eruleMap[PFfs_plank_3[state].f88 +326]
#define PFfs_ValidateExpr_rule(state)	PFfs_eruleMap[PFfs_plank_3[state].f87 +324]
#define PFfs_PathExpr_rule(state)	PFfs_eruleMap[PFfs_plank_7[state].f37 +241]
#define PFfs_LocationPath_rule(state)	PFfs_eruleMap[PFfs_plank_3[state].f86 +155]
#define PFfs_FilterExpr_rule(state)	PFfs_eruleMap[PFfs_plank_3[state].f85 +121]
#define PFfs_PrimaryExpr_rule(state)	PFfs_eruleMap[PFfs_plank_4[state].f84 +247]
#define PFfs_Literal_rule(state)	PFfs_eruleMap[PFfs_plank_4[state].f83 +152]
#define PFfs_NumericLiteral_rule(state)	PFfs_eruleMap[PFfs_plank_4[state].f82 +182]
#define PFfs_IntegerLiteral_rule(state)	PFfs_eruleMap[PFfs_plank_4[state].f81 +137]
#define PFfs_DecimalLiteral_rule(state)	PFfs_eruleMap[PFfs_plank_4[state].f80 +57]
#define PFfs_DoubleLiteral_rule(state)	PFfs_eruleMap[PFfs_plank_4[state].f79 +89]
#define PFfs_StringLiteral_rule(state)	PFfs_eruleMap[PFfs_plank_9[state].f6 +292]
#define PFfs_Var_rule(state)	PFfs_eruleMap[PFfs_plank_7[state].f34 +330]
#define PFfs_ParenthesizedExpr_rule(state)	PFfs_eruleMap[PFfs_plank_9[state].f0 +239]
#define PFfs_ContextItemExpr_rule(state)	PFfs_eruleMap[PFfs_plank_4[state].f78 +53]
#define PFfs_FunctionCall_rule(state)	PFfs_eruleMap[PFfs_plank_4[state].f77 +127]
#define PFfs_FuncArgList_rule(state)	PFfs_eruleMap[PFfs_plank_9[state].f1 +124]
#define PFfs_OrderedExpr_rule(state)	PFfs_eruleMap[PFfs_plank_4[state].f76 +228]
#define PFfs_UnorderedExpr_rule(state)	PFfs_eruleMap[PFfs_plank_4[state].f75 +322]
#define PFfs_Constructor_rule(state)	PFfs_eruleMap[PFfs_plank_4[state].f74 +46]
#define PFfs_ElementConstructor_rule(state)	PFfs_eruleMap[PFfs_plank_4[state].f73 +96]
#define PFfs_AttributeConstructor_rule(state)	PFfs_eruleMap[PFfs_plank_4[state].f72 +9]
#define PFfs_TextConstructor_rule(state)	PFfs_eruleMap[PFfs_plank_4[state].f71 +299]
#define PFfs_CommentConstructor_rule(state)	PFfs_eruleMap[PFfs_plank_5[state].f70 +25]
#define PFfs_PIConstructor_rule(state)	PFfs_eruleMap[PFfs_plank_5[state].f69 +232]
#define PFfs_DocumentConstructor_rule(state)	PFfs_eruleMap[PFfs_plank_5[state].f68 +87]
#define PFfs_TagName_rule(state)	PFfs_eruleMap[PFfs_plank_5[state].f67 +294]
#define PFfs_OptContentExpr_rule(state)	PFfs_eruleMap[PFfs_plank_5[state].f66 +195]
#define PFfs_AttributeValue_rule(state)	PFfs_eruleMap[PFfs_plank_5[state].f65 +11]
#define PFfs_ExtensionExpr_rule(state)	PFfs_eruleMap[PFfs_plank_5[state].f64 +115]
#define PFfs_Pragmas_rule(state)	PFfs_eruleMap[PFfs_plank_5[state].f63 +244]
#define PFfs_RevalidationDecl_rule(state)	PFfs_eruleMap[PFfs_plank_5[state].f62 +278]
#define PFfs_InsertExpr_rule(state)	PFfs_eruleMap[PFfs_plank_5[state].f61 +132]
#define PFfs_DeleteExpr_rule(state)	PFfs_eruleMap[PFfs_plank_5[state].f60 +85]
#define PFfs_ReplaceExpr_rule(state)	PFfs_eruleMap[PFfs_plank_5[state].f59 +276]
#define PFfs_RenameExpr_rule(state)	PFfs_eruleMap[PFfs_plank_5[state].f58 +274]
#define PFfs_TransformExpr_rule(state)	PFfs_eruleMap[PFfs_plank_5[state].f57 +306]
#define PFfs_TransformBindings_rule(state)	PFfs_eruleMap[PFfs_plank_5[state].f56 +303]
#define PFfs_TransformBinding_rule(state)	PFfs_eruleMap[PFfs_plank_6[state].f49 +301]
#define PFfs_SourceExpr_rule(state)	PFfs_eruleMap[PFfs_plank_9[state].f0 +290]
#define PFfs_TargetExpr_rule(state)	PFfs_eruleMap[PFfs_plank_9[state].f0 +297]
#define PFfs_NewNameExpr_rule(state)	PFfs_eruleMap[PFfs_plank_9[state].f0 +174]
#define PFfs_RecursiveExpr_rule(state)	PFfs_eruleMap[PFfs_plank_6[state].f55 +272]
#define PFfs_OptTypeDeclForRec_rule(state)	PFfs_eruleMap[PFfs_plank_6[state].f54 +214]
#define PFfs_XRPCCall_rule(state)	PFfs_eruleMap[PFfs_plank_6[state].f53 +336]

#ifdef __STDC__
int PFfs_rule(int state, int goalnt) {
#else
int PFfs_rule(state, goalnt) int state; int goalnt; {
#endif
	PFfs_assert(state >= 0 && state < 167, PFfs_PANIC("Bad state %d passed to PFfs_rule\n", state));
	switch(goalnt) {
	case 1:
		return PFfs_Query_rule(state);
	case 2:
		return PFfs_MainModule_rule(state);
	case 3:
		return PFfs_LibraryModule_rule(state);
	case 4:
		return PFfs_ModuleDecl_rule(state);
	case 5:
		return PFfs_Prolog_rule(state);
	case 6:
		return PFfs_DeclsImports_rule(state);
	case 7:
		return PFfs_DeclImport_rule(state);
	case 8:
		return PFfs_DefaultNamespaceDecl_rule(state);
	case 9:
		return PFfs_BoundarySpaceDecl_rule(state);
	case 10:
		return PFfs_DefaultCollationDecl_rule(state);
	case 11:
		return PFfs_BaseURIDecl_rule(state);
	case 12:
		return PFfs_ConstructionDecl_rule(state);
	case 13:
		return PFfs_OptionDecl_rule(state);
	case 14:
		return PFfs_OrderingModeDecl_rule(state);
	case 15:
		return PFfs_EmptyOrderingDecl_rule(state);
	case 16:
		return PFfs_CopyNamespacesDecl_rule(state);
	case 17:
		return PFfs_NamespaceDecl_rule(state);
	case 18:
		return PFfs_SchemaImport_rule(state);
	case 19:
		return PFfs_ModuleImport_rule(state);
	case 20:
		return PFfs_OptImportLoc__rule(state);
	case 21:
		return PFfs_OptParamList__rule(state);
	case 22:
		return PFfs_ParamList__rule(state);
	case 23:
		return PFfs_Param_rule(state);
	case 24:
		return PFfs_OptSequenceType__rule(state);
	case 25:
		return PFfs_VariableDeclaration_rule(state);
	case 26:
		return PFfs_QueryBody_rule(state);
	case 27:
		return PFfs_OptExpr_rule(state);
	case 28:
		return PFfs_Expr_rule(state);
	case 29:
		return PFfs_ExprSingle_rule(state);
	case 30:
		return PFfs_FLWORExpr_rule(state);
	case 31:
		return PFfs_OptBindExpr_rule(state);
	case 32:
		return PFfs_OptPosVar_rule(state);
	case 33:
		return PFfs_TypeDeclaration_rule(state);
	case 34:
		return PFfs_OptWhereClause__rule(state);
	case 35:
		return PFfs_OrderSpecList_rule(state);
	case 36:
		return PFfs_QuantifiedExpr_rule(state);
	case 37:
		return PFfs_TypeswitchExpr_rule(state);
	case 38:
		return PFfs_OptCaseVar_rule(state);
	case 39:
		return PFfs_SequenceType_rule(state);
	case 40:
		return PFfs_SingleType_rule(state);
	case 41:
		return PFfs_ItemType_rule(state);
	case 42:
		return PFfs_AtomType_rule(state);
	case 43:
		return PFfs_NodeTest_rule(state);
	case 44:
		return PFfs_ElemOrAttrType_rule(state);
	case 45:
		return PFfs_SchemaType_rule(state);
	case 46:
		return PFfs_IfExpr_rule(state);
	case 47:
		return PFfs_OrExpr_rule(state);
	case 48:
		return PFfs_AndExpr_rule(state);
	case 49:
		return PFfs_ComparisonExpr_rule(state);
	case 50:
		return PFfs_RangeExpr_rule(state);
	case 51:
		return PFfs_AdditiveExpr_rule(state);
	case 52:
		return PFfs_MultiplicativeExpr_rule(state);
	case 53:
		return PFfs_UnionExpr_rule(state);
	case 54:
		return PFfs_IntersectExceptExpr_rule(state);
	case 55:
		return PFfs_InstanceofExpr_rule(state);
	case 56:
		return PFfs_TreatExpr_rule(state);
	case 57:
		return PFfs_CastableExpr_rule(state);
	case 58:
		return PFfs_CastExpr_rule(state);
	case 59:
		return PFfs_UnaryExpr_rule(state);
	case 60:
		return PFfs_ValueExpr_rule(state);
	case 61:
		return PFfs_ValidateExpr_rule(state);
	case 62:
		return PFfs_PathExpr_rule(state);
	case 63:
		return PFfs_LocationPath_rule(state);
	case 64:
		return PFfs_FilterExpr_rule(state);
	case 65:
		return PFfs_PrimaryExpr_rule(state);
	case 66:
		return PFfs_Literal_rule(state);
	case 67:
		return PFfs_NumericLiteral_rule(state);
	case 68:
		return PFfs_IntegerLiteral_rule(state);
	case 69:
		return PFfs_DecimalLiteral_rule(state);
	case 70:
		return PFfs_DoubleLiteral_rule(state);
	case 71:
		return PFfs_StringLiteral_rule(state);
	case 72:
		return PFfs_Var_rule(state);
	case 73:
		return PFfs_ParenthesizedExpr_rule(state);
	case 74:
		return PFfs_ContextItemExpr_rule(state);
	case 75:
		return PFfs_FunctionCall_rule(state);
	case 76:
		return PFfs_FuncArgList_rule(state);
	case 77:
		return PFfs_OrderedExpr_rule(state);
	case 78:
		return PFfs_UnorderedExpr_rule(state);
	case 79:
		return PFfs_Constructor_rule(state);
	case 80:
		return PFfs_ElementConstructor_rule(state);
	case 81:
		return PFfs_AttributeConstructor_rule(state);
	case 82:
		return PFfs_TextConstructor_rule(state);
	case 83:
		return PFfs_CommentConstructor_rule(state);
	case 84:
		return PFfs_PIConstructor_rule(state);
	case 85:
		return PFfs_DocumentConstructor_rule(state);
	case 86:
		return PFfs_TagName_rule(state);
	case 87:
		return PFfs_OptContentExpr_rule(state);
	case 88:
		return PFfs_AttributeValue_rule(state);
	case 89:
		return PFfs_ExtensionExpr_rule(state);
	case 90:
		return PFfs_Pragmas_rule(state);
	case 91:
		return PFfs_RevalidationDecl_rule(state);
	case 92:
		return PFfs_InsertExpr_rule(state);
	case 93:
		return PFfs_DeleteExpr_rule(state);
	case 94:
		return PFfs_ReplaceExpr_rule(state);
	case 95:
		return PFfs_RenameExpr_rule(state);
	case 96:
		return PFfs_TransformExpr_rule(state);
	case 97:
		return PFfs_TransformBindings_rule(state);
	case 98:
		return PFfs_TransformBinding_rule(state);
	case 99:
		return PFfs_SourceExpr_rule(state);
	case 100:
		return PFfs_TargetExpr_rule(state);
	case 101:
		return PFfs_NewNameExpr_rule(state);
	case 102:
		return PFfs_RecursiveExpr_rule(state);
	case 103:
		return PFfs_OptTypeDeclForRec_rule(state);
	case 104:
		return PFfs_XRPCCall_rule(state);
	default:
		PFfs_PANIC("Unknown nonterminal %d in PFfs_rule;\n", goalnt);
		abort();
		return 0;
	}
}

int PFfs_TEMP;
#define PFfs_and_state(l,r)	( (PFfs_TEMP = PFfs_and_transition[PFfs_plank_9[l].f0][PFfs_plank_9[r].f0]) ? PFfs_TEMP + 0 : 0 )
#define PFfs_apply_state(l)	( (PFfs_TEMP = PFfs_plank_9[l].f1) ? PFfs_TEMP + 1 : 0 )
#define PFfs_args_state(l,r)	( (PFfs_TEMP = PFfs_args_transition[PFfs_plank_9[l].f0][PFfs_plank_9[r].f2]) ? PFfs_TEMP + 3 : 0 )
#define PFfs_atom_ty_state(l)	( (PFfs_TEMP = PFfs_plank_7[l].f3) ? PFfs_TEMP + 4 : 0 )
#define PFfs_attr_state(l,r)	( (PFfs_TEMP = PFfs_attr_transition[PFfs_plank_9[l].f4][PFfs_plank_9[r].f5]) ? PFfs_TEMP + 5 : 0 )
#define PFfs_base_uri_state(l)	( (PFfs_TEMP = PFfs_plank_9[l].f6) ? PFfs_TEMP + 6 : 0 )
#define PFfs_bind_state(l,r)	( (PFfs_TEMP = PFfs_bind_transition[PFfs_plank_9[l].f7][PFfs_plank_9[r].f0]) ? PFfs_TEMP + 7 : 0 )
#define PFfs_binds_state(l,r)	( (PFfs_TEMP = PFfs_binds_transition[PFfs_plank_9[l].f8][PFfs_plank_9[r].f9]) ? PFfs_TEMP + 11 : 0 )
#define PFfs_case__state(l,r)	( (PFfs_TEMP = PFfs_case__transition[PFfs_plank_8[l].f10][PFfs_plank_9[r].f0]) ? PFfs_TEMP + 18 : 0 )
#define PFfs_cases_state(l,r)	( (PFfs_TEMP = PFfs_cases_transition[PFfs_plank_8[l].f11][PFfs_plank_8[r].f12]) ? PFfs_TEMP + 19 : 0 )
#define PFfs_cast_state(l,r)	( (PFfs_TEMP = PFfs_cast_transition[PFfs_plank_9[l].f0][PFfs_plank_8[r].f13]) ? PFfs_TEMP + 21 : 0 )
#define PFfs_castable_state(l,r)	( (PFfs_TEMP = PFfs_castable_transition[PFfs_plank_9[l].f0][PFfs_plank_8[r].f13]) ? PFfs_TEMP + 22 : 0 )
#define PFfs_coll_decl_state(l)	( (PFfs_TEMP = PFfs_plank_9[l].f6) ? PFfs_TEMP + 23 : 0 )
#define PFfs_comment_state(l)	( (PFfs_TEMP = PFfs_plank_9[l].f0) ? PFfs_TEMP + 24 : 0 )
#define PFfs_constr_decl_state	26
#define PFfs_contseq_state(l,r)	( (PFfs_TEMP = PFfs_contseq_transition[PFfs_plank_9[l].f5][PFfs_plank_9[r].f5]) ? PFfs_TEMP + 26 : 0 )
#define PFfs_decl_imps_state(l,r)	( (PFfs_TEMP = PFfs_decl_imps_transition[PFfs_plank_8[l].f14][PFfs_plank_8[r].f15]) ? PFfs_TEMP + 28 : 0 )
#define PFfs_def_order_state	32
#define PFfs_default__state(l,r)	( (PFfs_TEMP = PFfs_default__transition[PFfs_plank_8[l].f16][PFfs_plank_9[r].f0]) ? PFfs_TEMP + 32 : 0 )
#define PFfs_div_state(l,r)	( (PFfs_TEMP = PFfs_div_transition[PFfs_plank_9[l].f0][PFfs_plank_9[r].f0]) ? PFfs_TEMP + 34 : 0 )
#define PFfs_doc_state(l)	( (PFfs_TEMP = PFfs_plank_9[l].f0) ? PFfs_TEMP + 35 : 0 )
#define PFfs_dot_state	38
#define PFfs_elem_state(l,r)	( (PFfs_TEMP = PFfs_elem_transition[PFfs_plank_9[l].f4][PFfs_plank_9[r].f5]) ? PFfs_TEMP + 38 : 0 )
#define PFfs_empty_seq_state	40
#define PFfs_empty_ty_state	41
#define PFfs_ens_decl_state(l)	( (PFfs_TEMP = PFfs_plank_9[l].f6) ? PFfs_TEMP + 41 : 0 )
#define PFfs_eq_state(l,r)	( (PFfs_TEMP = PFfs_eq_transition[PFfs_plank_9[l].f0][PFfs_plank_9[r].f0]) ? PFfs_TEMP + 42 : 0 )
#define PFfs_every_state(l,r)	( (PFfs_TEMP = PFfs_every_transition[PFfs_plank_8[l].f17][PFfs_plank_9[r].f0]) ? PFfs_TEMP + 43 : 0 )
#define PFfs_except_state(l,r)	( (PFfs_TEMP = PFfs_except_transition[PFfs_plank_9[l].f0][PFfs_plank_9[r].f0]) ? PFfs_TEMP + 45 : 0 )
#define PFfs_exprseq_state(l,r)	( (PFfs_TEMP = PFfs_exprseq_transition[PFfs_plank_9[l].f0][PFfs_plank_9[r].f0]) ? PFfs_TEMP + 46 : 0 )
#define PFfs_external_state	50
#define PFfs_ext_expr_state(l,r)	( (PFfs_TEMP = PFfs_ext_expr_transition[PFfs_plank_8[l].f18][PFfs_plank_8[r].f19]) ? PFfs_TEMP + 47 : 0 )
#define PFfs_flwr_state(l,r)	( (PFfs_TEMP = PFfs_flwr_transition[PFfs_plank_8[l].f20][PFfs_plank_8[r].f21]) ? PFfs_TEMP + 50 : 0 )
#define PFfs_fns_decl_state(l)	( (PFfs_TEMP = PFfs_plank_9[l].f6) ? PFfs_TEMP + 52 : 0 )
#define PFfs_fun_state(l,r)	( (PFfs_TEMP = PFfs_fun_transition[PFfs_plank_8[l].f22][PFfs_plank_8[r].f23]) ? PFfs_TEMP + 53 : 0 )
#define PFfs_fun_sig_state(l,r)	( (PFfs_TEMP = PFfs_fun_sig_transition[PFfs_plank_8[l].f24][PFfs_plank_7[r].f25]) ? PFfs_TEMP + 55 : 0 )
#define PFfs_ge_state(l,r)	( (PFfs_TEMP = PFfs_ge_transition[PFfs_plank_9[l].f0][PFfs_plank_9[r].f0]) ? PFfs_TEMP + 56 : 0 )
#define PFfs_gt_state(l,r)	( (PFfs_TEMP = PFfs_gt_transition[PFfs_plank_9[l].f0][PFfs_plank_9[r].f0]) ? PFfs_TEMP + 57 : 0 )
#define PFfs_gtgt_state(l,r)	( (PFfs_TEMP = PFfs_gtgt_transition[PFfs_plank_9[l].f0][PFfs_plank_9[r].f0]) ? PFfs_TEMP + 58 : 0 )
#define PFfs_idiv_state(l,r)	( (PFfs_TEMP = PFfs_idiv_transition[PFfs_plank_9[l].f0][PFfs_plank_9[r].f0]) ? PFfs_TEMP + 59 : 0 )
#define PFfs_if__state(l,r)	( (PFfs_TEMP = PFfs_if__transition[PFfs_plank_9[l].f0][PFfs_plank_7[r].f26]) ? PFfs_TEMP + 60 : 0 )
#define PFfs_copy_ns_state	28
#define PFfs_instof_state(l,r)	( (PFfs_TEMP = PFfs_instof_transition[PFfs_plank_9[l].f0][PFfs_plank_7[r].f25]) ? PFfs_TEMP + 62 : 0 )
#define PFfs_intersect_state(l,r)	( (PFfs_TEMP = PFfs_intersect_transition[PFfs_plank_9[l].f0][PFfs_plank_9[r].f0]) ? PFfs_TEMP + 63 : 0 )
#define PFfs_is_state(l,r)	( (PFfs_TEMP = PFfs_is_transition[PFfs_plank_9[l].f0][PFfs_plank_9[r].f0]) ? PFfs_TEMP + 64 : 0 )
#define PFfs_item_ty_state(l)	( (PFfs_TEMP = PFfs_plank_7[l].f3) ? PFfs_TEMP + 65 : 0 )
#define PFfs_le_state(l,r)	( (PFfs_TEMP = PFfs_le_transition[PFfs_plank_9[l].f0][PFfs_plank_9[r].f0]) ? PFfs_TEMP + 66 : 0 )
#define PFfs_let_state(l,r)	( (PFfs_TEMP = PFfs_let_transition[PFfs_plank_7[l].f27][PFfs_plank_9[r].f0]) ? PFfs_TEMP + 67 : 0 )
#define PFfs_lib_mod_state(l,r)	( (PFfs_TEMP = PFfs_lib_mod_transition[PFfs_plank_7[l].f28][PFfs_plank_8[r].f15]) ? PFfs_TEMP + 69 : 0 )
#define PFfs_lit_dbl_state	71
#define PFfs_lit_dec_state	72
#define PFfs_lit_int_state	73
#define PFfs_lit_str_state	74
#define PFfs_locpath_state(l,r)	( (PFfs_TEMP = PFfs_locpath_transition[PFfs_plank_7[l].f29][PFfs_plank_9[r].f0]) ? PFfs_TEMP + 74 : 0 )
#define PFfs_lt_state(l,r)	( (PFfs_TEMP = PFfs_lt_transition[PFfs_plank_9[l].f0][PFfs_plank_9[r].f0]) ? PFfs_TEMP + 76 : 0 )
#define PFfs_ltlt_state(l,r)	( (PFfs_TEMP = PFfs_ltlt_transition[PFfs_plank_9[l].f0][PFfs_plank_9[r].f0]) ? PFfs_TEMP + 77 : 0 )
#define PFfs_main_mod_state(l,r)	( (PFfs_TEMP = PFfs_main_mod_transition[PFfs_plank_8[l].f15][PFfs_plank_9[r].f0]) ? PFfs_TEMP + 78 : 0 )
#define PFfs_minus_state(l,r)	( (PFfs_TEMP = PFfs_minus_transition[PFfs_plank_9[l].f0][PFfs_plank_9[r].f0]) ? PFfs_TEMP + 79 : 0 )
#define PFfs_mod_state(l,r)	( (PFfs_TEMP = PFfs_mod_transition[PFfs_plank_9[l].f0][PFfs_plank_9[r].f0]) ? PFfs_TEMP + 80 : 0 )
#define PFfs_mod_imp_state(l,r)	( (PFfs_TEMP = PFfs_mod_imp_transition[PFfs_plank_9[l].f6][PFfs_plank_7[r].f30]) ? PFfs_TEMP + 81 : 0 )
#define PFfs_mod_ns_state(l)	( (PFfs_TEMP = PFfs_plank_9[l].f6) ? PFfs_TEMP + 82 : 0 )
#define PFfs_mult_state(l,r)	( (PFfs_TEMP = PFfs_mult_transition[PFfs_plank_9[l].f0][PFfs_plank_9[r].f0]) ? PFfs_TEMP + 84 : 0 )
#define PFfs_named_ty_state	86
#define PFfs_ne_state(l,r)	( (PFfs_TEMP = PFfs_ne_transition[PFfs_plank_9[l].f0][PFfs_plank_9[r].f0]) ? PFfs_TEMP + 86 : 0 )
#define PFfs_nil_state	88
#define PFfs_node_ty_state(l)	( (PFfs_TEMP = PFfs_plank_7[l].f31) ? PFfs_TEMP + 88 : 0 )
#define PFfs_ns_decl_state(l)	( (PFfs_TEMP = PFfs_plank_9[l].f6) ? PFfs_TEMP + 93 : 0 )
#define PFfs_option_state(l)	( (PFfs_TEMP = PFfs_plank_9[l].f6) ? PFfs_TEMP + 94 : 0 )
#define PFfs_or_state(l,r)	( (PFfs_TEMP = PFfs_or_transition[PFfs_plank_9[l].f0][PFfs_plank_9[r].f0]) ? PFfs_TEMP + 95 : 0 )
#define PFfs_ord_ret_state(l,r)	( (PFfs_TEMP = PFfs_ord_ret_transition[PFfs_plank_7[l].f32][PFfs_plank_9[r].f0]) ? PFfs_TEMP + 96 : 0 )
#define PFfs_orderby_state(l)	( (PFfs_TEMP = PFfs_plank_7[l].f33) ? PFfs_TEMP + 98 : 0 )
#define PFfs_ordered_state(l)	( (PFfs_TEMP = PFfs_plank_9[l].f0) ? PFfs_TEMP + 99 : 0 )
#define PFfs_ordering_mode_state	101
#define PFfs_orderspecs_state(l,r)	( (PFfs_TEMP = PFfs_orderspecs_transition[PFfs_plank_9[l].f0][PFfs_plank_7[r].f33]) ? PFfs_TEMP + 101 : 0 )
#define PFfs_param_state(l,r)	( (PFfs_TEMP = PFfs_param_transition[PFfs_plank_7[l].f25][PFfs_plank_7[r].f34]) ? PFfs_TEMP + 102 : 0 )
#define PFfs_params_state(l,r)	( (PFfs_TEMP = PFfs_params_transition[PFfs_plank_7[l].f35][PFfs_plank_8[r].f24]) ? PFfs_TEMP + 103 : 0 )
#define PFfs_pi_state(l,r)	( (PFfs_TEMP = PFfs_pi_transition[PFfs_plank_9[l].f0][PFfs_plank_9[r].f5]) ? PFfs_TEMP + 104 : 0 )
#define PFfs_plus_state(l,r)	( (PFfs_TEMP = PFfs_plus_transition[PFfs_plank_9[l].f0][PFfs_plank_9[r].f0]) ? PFfs_TEMP + 105 : 0 )
#define PFfs_pragma_state	107
#define PFfs_pragmas_state(l,r)	( (PFfs_TEMP = PFfs_pragmas_transition[PFfs_plank_7[l].f36][PFfs_plank_8[r].f18]) ? PFfs_TEMP + 107 : 0 )
#define PFfs_pred_state(l,r)	( (PFfs_TEMP = PFfs_pred_transition[PFfs_plank_7[l].f37][PFfs_plank_9[r].f0]) ? PFfs_TEMP + 108 : 0 )
#define PFfs_range_state(l,r)	( (PFfs_TEMP = PFfs_range_transition[PFfs_plank_9[l].f0][PFfs_plank_9[r].f0]) ? PFfs_TEMP + 110 : 0 )
#define PFfs_req_name_state	115
#define PFfs_req_ty_state(l,r)	( (PFfs_TEMP = PFfs_req_ty_transition[PFfs_plank_7[l].f38][PFfs_plank_7[r].f39]) ? PFfs_TEMP + 115 : 0 )
#define PFfs_root__state	121
#define PFfs_schm_ats_state(l,r)	( (PFfs_TEMP = PFfs_schm_ats_transition[PFfs_plank_9[l].f6][PFfs_plank_7[r].f30]) ? PFfs_TEMP + 121 : 0 )
#define PFfs_schm_attr_state	123
#define PFfs_schm_elem_state	124
#define PFfs_schm_imp_state(l,r)	( (PFfs_TEMP = PFfs_schm_imp_transition[PFfs_plank_9[l].f6][PFfs_plank_7[r].f30]) ? PFfs_TEMP + 124 : 0 )
#define PFfs_seq_ty_state(l)	( (PFfs_TEMP = PFfs_plank_6[l].f40) ? PFfs_TEMP + 126 : 0 )
#define PFfs_some_state(l,r)	( (PFfs_TEMP = PFfs_some_transition[PFfs_plank_8[l].f17][PFfs_plank_9[r].f0]) ? PFfs_TEMP + 130 : 0 )
#define PFfs_step_state(l)	( (PFfs_TEMP = PFfs_plank_6[l].f41) ? PFfs_TEMP + 132 : 0 )
#define PFfs_tag_state	135
#define PFfs_text_state(l)	( (PFfs_TEMP = PFfs_plank_9[l].f0) ? PFfs_TEMP + 135 : 0 )
#define PFfs_then_else_state(l,r)	( (PFfs_TEMP = PFfs_then_else_transition[PFfs_plank_9[l].f0][PFfs_plank_9[r].f0]) ? PFfs_TEMP + 136 : 0 )
#define PFfs_treat_state(l,r)	( (PFfs_TEMP = PFfs_treat_transition[PFfs_plank_9[l].f0][PFfs_plank_7[r].f25]) ? PFfs_TEMP + 139 : 0 )
#define PFfs_typeswitch_state(l,r)	( (PFfs_TEMP = PFfs_typeswitch_transition[PFfs_plank_9[l].f0][PFfs_plank_6[r].f42]) ? PFfs_TEMP + 140 : 0 )
#define PFfs_uminus_state(l)	( (PFfs_TEMP = PFfs_plank_9[l].f0) ? PFfs_TEMP + 141 : 0 )
#define PFfs_union__state(l,r)	( (PFfs_TEMP = PFfs_union__transition[PFfs_plank_9[l].f0][PFfs_plank_9[r].f0]) ? PFfs_TEMP + 142 : 0 )
#define PFfs_unordered_state(l)	( (PFfs_TEMP = PFfs_plank_9[l].f0) ? PFfs_TEMP + 143 : 0 )
#define PFfs_uplus_state(l)	( (PFfs_TEMP = PFfs_plank_9[l].f0) ? PFfs_TEMP + 144 : 0 )
#define PFfs_val_eq_state(l,r)	( (PFfs_TEMP = PFfs_val_eq_transition[PFfs_plank_9[l].f0][PFfs_plank_9[r].f0]) ? PFfs_TEMP + 145 : 0 )
#define PFfs_val_ge_state(l,r)	( (PFfs_TEMP = PFfs_val_ge_transition[PFfs_plank_9[l].f0][PFfs_plank_9[r].f0]) ? PFfs_TEMP + 146 : 0 )
#define PFfs_val_gt_state(l,r)	( (PFfs_TEMP = PFfs_val_gt_transition[PFfs_plank_9[l].f0][PFfs_plank_9[r].f0]) ? PFfs_TEMP + 147 : 0 )
#define PFfs_val_le_state(l,r)	( (PFfs_TEMP = PFfs_val_le_transition[PFfs_plank_9[l].f0][PFfs_plank_9[r].f0]) ? PFfs_TEMP + 148 : 0 )
#define PFfs_val_lt_state(l,r)	( (PFfs_TEMP = PFfs_val_lt_transition[PFfs_plank_9[l].f0][PFfs_plank_9[r].f0]) ? PFfs_TEMP + 149 : 0 )
#define PFfs_val_ne_state(l,r)	( (PFfs_TEMP = PFfs_val_ne_transition[PFfs_plank_9[l].f0][PFfs_plank_9[r].f0]) ? PFfs_TEMP + 150 : 0 )
#define PFfs_validate_state(l)	( (PFfs_TEMP = PFfs_plank_9[l].f0) ? PFfs_TEMP + 151 : 0 )
#define PFfs_var_state	153
#define PFfs_var_decl_state(l,r)	( (PFfs_TEMP = PFfs_var_decl_transition[PFfs_plank_7[l].f27][PFfs_plank_8[r].f23]) ? PFfs_TEMP + 153 : 0 )
#define PFfs_var_type_state(l,r)	( (PFfs_TEMP = PFfs_var_type_transition[PFfs_plank_6[l].f43][PFfs_plank_6[r].f44]) ? PFfs_TEMP + 156 : 0 )
#define PFfs_vars_state(l,r)	( (PFfs_TEMP = PFfs_vars_transition[PFfs_plank_7[l].f27][PFfs_plank_6[r].f43]) ? PFfs_TEMP + 159 : 0 )
#define PFfs_where_state(l,r)	( (PFfs_TEMP = PFfs_where_transition[PFfs_plank_6[l].f45][PFfs_plank_6[r].f46]) ? PFfs_TEMP + 163 : 0 )
#define PFfs_boundspc_decl_state	18
#define PFfs_revalid_state	120
#define PFfs_insert_state(l,r)	( (PFfs_TEMP = PFfs_insert_transition[PFfs_plank_9[l].f0][PFfs_plank_9[r].f0]) ? PFfs_TEMP + 61 : 0 )
#define PFfs_delete_state(l)	( (PFfs_TEMP = PFfs_plank_9[l].f0) ? PFfs_TEMP + 33 : 0 )
#define PFfs_replace_state(l,r)	( (PFfs_TEMP = PFfs_replace_transition[PFfs_plank_9[l].f0][PFfs_plank_9[r].f0]) ? PFfs_TEMP + 113 : 0 )
#define PFfs_rename_state(l,r)	( (PFfs_TEMP = PFfs_rename_transition[PFfs_plank_9[l].f0][PFfs_plank_9[r].f0]) ? PFfs_TEMP + 112 : 0 )
#define PFfs_transform_state(l,r)	( (PFfs_TEMP = PFfs_transform_transition[PFfs_plank_6[l].f47][PFfs_plank_6[r].f48]) ? PFfs_TEMP + 138 : 0 )
#define PFfs_modify_state(l,r)	( (PFfs_TEMP = PFfs_modify_transition[PFfs_plank_9[l].f0][PFfs_plank_9[r].f0]) ? PFfs_TEMP + 83 : 0 )
#define PFfs_transbinds_state(l,r)	( (PFfs_TEMP = PFfs_transbinds_transition[PFfs_plank_6[l].f49][PFfs_plank_6[r].f47]) ? PFfs_TEMP + 137 : 0 )
#define PFfs_stmt_ty_state(l)	( (PFfs_TEMP = PFfs_plank_7[l].f3) ? PFfs_TEMP + 133 : 0 )
#define PFfs_recursion_state(l,r)	( (PFfs_TEMP = PFfs_recursion_transition[PFfs_plank_6[l].f50][PFfs_plank_6[r].f51]) ? PFfs_TEMP + 111 : 0 )
#define PFfs_seed_state(l,r)	( (PFfs_TEMP = PFfs_seed_transition[PFfs_plank_9[l].f0][PFfs_plank_9[r].f0]) ? PFfs_TEMP + 125 : 0 )
#define PFfs_xrpc_state(l,r)	( (PFfs_TEMP = PFfs_xrpc_transition[PFfs_plank_9[l].f0][PFfs_plank_6[r].f52]) ? PFfs_TEMP + 165 : 0 )
#define PFfs_docmgmt_ty_state(l)	( (PFfs_TEMP = PFfs_plank_7[l].f3) ? PFfs_TEMP + 36 : 0 )

#ifdef __STDC__
int PFfs_state(int op, int l, int r) {
#else
int PFfs_state(op, l, r) int op; int l; int r; {
#endif
	register int PFfs_TEMP;
#ifndef NDEBUG
	switch (op) {
	case 1:
	case 3:
	case 5:
	case 7:
	case 8:
	case 9:
	case 10:
	case 11:
	case 12:
	case 16:
	case 17:
	case 19:
	case 20:
	case 23:
	case 27:
	case 28:
	case 29:
	case 30:
	case 32:
	case 33:
	case 35:
	case 38:
	case 39:
	case 40:
	case 41:
	case 42:
	case 43:
	case 45:
	case 46:
	case 47:
	case 49:
	case 50:
	case 51:
	case 56:
	case 57:
	case 58:
	case 59:
	case 60:
	case 61:
	case 62:
	case 64:
	case 66:
	case 71:
	case 72:
	case 76:
	case 77:
	case 78:
	case 79:
	case 80:
	case 82:
	case 83:
	case 84:
	case 86:
	case 88:
	case 91:
	case 93:
	case 97:
	case 98:
	case 99:
	case 101:
	case 104:
	case 105:
	case 106:
	case 107:
	case 108:
	case 109:
	case 112:
	case 113:
	case 115:
	case 116:
	case 119:
	case 121:
	case 122:
	case 123:
	case 124:
	case 125:
	case 127:
	case 128:
	case 129:
		PFfs_assert(r >= 0 && r < 167, PFfs_PANIC("Bad state %d passed to PFfs_state\n", r));
		/*FALLTHROUGH*/
	case 2:
	case 4:
	case 6:
	case 13:
	case 14:
	case 21:
	case 26:
	case 34:
	case 48:
	case 63:
	case 68:
	case 69:
	case 70:
	case 73:
	case 74:
	case 92:
	case 94:
	case 96:
	case 100:
	case 102:
	case 103:
	case 110:
	case 120:
	case 126:
	case 130:
		PFfs_assert(l >= 0 && l < 167, PFfs_PANIC("Bad state %d passed to PFfs_state\n", l));
		/*FALLTHROUGH*/
	case 15:
	case 18:
	case 22:
	case 24:
	case 25:
	case 31:
	case 44:
	case 52:
	case 53:
	case 54:
	case 55:
	case 65:
	case 67:
	case 75:
	case 81:
	case 85:
	case 87:
	case 89:
	case 90:
	case 95:
	case 111:
	case 117:
	case 118:
		break;
	}
#endif
	switch (op) {
	default: PFfs_PANIC("Unknown op %d in PFfs_state\n", op); abort(); return 0;
	case 1:
		return PFfs_and_state(l,r);
	case 2:
		return PFfs_apply_state(l);
	case 3:
		return PFfs_args_state(l,r);
	case 4:
		return PFfs_atom_ty_state(l);
	case 5:
		return PFfs_attr_state(l,r);
	case 6:
		return PFfs_base_uri_state(l);
	case 7:
		return PFfs_bind_state(l,r);
	case 8:
		return PFfs_binds_state(l,r);
	case 9:
		return PFfs_case__state(l,r);
	case 10:
		return PFfs_cases_state(l,r);
	case 11:
		return PFfs_cast_state(l,r);
	case 12:
		return PFfs_castable_state(l,r);
	case 13:
		return PFfs_coll_decl_state(l);
	case 14:
		return PFfs_comment_state(l);
	case 15:
		return PFfs_constr_decl_state;
	case 16:
		return PFfs_contseq_state(l,r);
	case 17:
		return PFfs_decl_imps_state(l,r);
	case 18:
		return PFfs_def_order_state;
	case 19:
		return PFfs_default__state(l,r);
	case 20:
		return PFfs_div_state(l,r);
	case 21:
		return PFfs_doc_state(l);
	case 22:
		return PFfs_dot_state;
	case 23:
		return PFfs_elem_state(l,r);
	case 24:
		return PFfs_empty_seq_state;
	case 25:
		return PFfs_empty_ty_state;
	case 26:
		return PFfs_ens_decl_state(l);
	case 27:
		return PFfs_eq_state(l,r);
	case 28:
		return PFfs_every_state(l,r);
	case 29:
		return PFfs_except_state(l,r);
	case 30:
		return PFfs_exprseq_state(l,r);
	case 31:
		return PFfs_external_state;
	case 32:
		return PFfs_ext_expr_state(l,r);
	case 33:
		return PFfs_flwr_state(l,r);
	case 34:
		return PFfs_fns_decl_state(l);
	case 35:
		return PFfs_fun_state(l,r);
	case 38:
		return PFfs_fun_sig_state(l,r);
	case 39:
		return PFfs_ge_state(l,r);
	case 40:
		return PFfs_gt_state(l,r);
	case 41:
		return PFfs_gtgt_state(l,r);
	case 42:
		return PFfs_idiv_state(l,r);
	case 43:
		return PFfs_if__state(l,r);
	case 44:
		return PFfs_copy_ns_state;
	case 45:
		return PFfs_instof_state(l,r);
	case 46:
		return PFfs_intersect_state(l,r);
	case 47:
		return PFfs_is_state(l,r);
	case 48:
		return PFfs_item_ty_state(l);
	case 49:
		return PFfs_le_state(l,r);
	case 50:
		return PFfs_let_state(l,r);
	case 51:
		return PFfs_lib_mod_state(l,r);
	case 52:
		return PFfs_lit_dbl_state;
	case 53:
		return PFfs_lit_dec_state;
	case 54:
		return PFfs_lit_int_state;
	case 55:
		return PFfs_lit_str_state;
	case 56:
		return PFfs_locpath_state(l,r);
	case 57:
		return PFfs_lt_state(l,r);
	case 58:
		return PFfs_ltlt_state(l,r);
	case 59:
		return PFfs_main_mod_state(l,r);
	case 60:
		return PFfs_minus_state(l,r);
	case 61:
		return PFfs_mod_state(l,r);
	case 62:
		return PFfs_mod_imp_state(l,r);
	case 63:
		return PFfs_mod_ns_state(l);
	case 64:
		return PFfs_mult_state(l,r);
	case 65:
		return PFfs_named_ty_state;
	case 66:
		return PFfs_ne_state(l,r);
	case 67:
		return PFfs_nil_state;
	case 68:
		return PFfs_node_ty_state(l);
	case 69:
		return PFfs_ns_decl_state(l);
	case 70:
		return PFfs_option_state(l);
	case 71:
		return PFfs_or_state(l,r);
	case 72:
		return PFfs_ord_ret_state(l,r);
	case 73:
		return PFfs_orderby_state(l);
	case 74:
		return PFfs_ordered_state(l);
	case 75:
		return PFfs_ordering_mode_state;
	case 76:
		return PFfs_orderspecs_state(l,r);
	case 77:
		return PFfs_param_state(l,r);
	case 78:
		return PFfs_params_state(l,r);
	case 79:
		return PFfs_pi_state(l,r);
	case 80:
		return PFfs_plus_state(l,r);
	case 81:
		return PFfs_pragma_state;
	case 82:
		return PFfs_pragmas_state(l,r);
	case 83:
		return PFfs_pred_state(l,r);
	case 84:
		return PFfs_range_state(l,r);
	case 85:
		return PFfs_req_name_state;
	case 86:
		return PFfs_req_ty_state(l,r);
	case 87:
		return PFfs_root__state;
	case 88:
		return PFfs_schm_ats_state(l,r);
	case 89:
		return PFfs_schm_attr_state;
	case 90:
		return PFfs_schm_elem_state;
	case 91:
		return PFfs_schm_imp_state(l,r);
	case 92:
		return PFfs_seq_ty_state(l);
	case 93:
		return PFfs_some_state(l,r);
	case 94:
		return PFfs_step_state(l);
	case 95:
		return PFfs_tag_state;
	case 96:
		return PFfs_text_state(l);
	case 97:
		return PFfs_then_else_state(l,r);
	case 98:
		return PFfs_treat_state(l,r);
	case 99:
		return PFfs_typeswitch_state(l,r);
	case 100:
		return PFfs_uminus_state(l);
	case 101:
		return PFfs_union__state(l,r);
	case 102:
		return PFfs_unordered_state(l);
	case 103:
		return PFfs_uplus_state(l);
	case 104:
		return PFfs_val_eq_state(l,r);
	case 105:
		return PFfs_val_ge_state(l,r);
	case 106:
		return PFfs_val_gt_state(l,r);
	case 107:
		return PFfs_val_le_state(l,r);
	case 108:
		return PFfs_val_lt_state(l,r);
	case 109:
		return PFfs_val_ne_state(l,r);
	case 110:
		return PFfs_validate_state(l);
	case 111:
		return PFfs_var_state;
	case 112:
		return PFfs_var_decl_state(l,r);
	case 113:
		return PFfs_var_type_state(l,r);
	case 115:
		return PFfs_vars_state(l,r);
	case 116:
		return PFfs_where_state(l,r);
	case 117:
		return PFfs_boundspc_decl_state;
	case 118:
		return PFfs_revalid_state;
	case 119:
		return PFfs_insert_state(l,r);
	case 120:
		return PFfs_delete_state(l);
	case 121:
		return PFfs_replace_state(l,r);
	case 122:
		return PFfs_rename_state(l,r);
	case 123:
		return PFfs_transform_state(l,r);
	case 124:
		return PFfs_modify_state(l,r);
	case 125:
		return PFfs_transbinds_state(l,r);
	case 126:
		return PFfs_stmt_ty_state(l);
	case 127:
		return PFfs_recursion_state(l,r);
	case 128:
		return PFfs_seed_state(l,r);
	case 129:
		return PFfs_xrpc_state(l,r);
	case 130:
		return PFfs_docmgmt_ty_state(l);
	}
}
#ifdef PFfs_STATE_LABEL
#define PFfs_INCLUDE_EXTRA
#else
#ifdef STATE_LABEL
#define PFfs_INCLUDE_EXTRA
#define PFfs_STATE_LABEL 	STATE_LABEL
#define PFfs_NODEPTR_TYPE	NODEPTR_TYPE
#define PFfs_LEFT_CHILD  	LEFT_CHILD
#define PFfs_OP_LABEL    	OP_LABEL
#define PFfs_RIGHT_CHILD 	RIGHT_CHILD
#endif /* STATE_LABEL */
#endif /* PFfs_STATE_LABEL */

#ifdef PFfs_INCLUDE_EXTRA

#ifdef __STDC__
int PFfs_label(PFfs_NODEPTR_TYPE n) {
#else
int PFfs_label(n) PFfs_NODEPTR_TYPE n; {
#endif
	PFfs_assert(n, PFfs_PANIC("NULL pointer passed to PFfs_label\n"));
	switch (PFfs_OP_LABEL(n)) {
	default: PFfs_PANIC("Bad op %d in PFfs_label\n", PFfs_OP_LABEL(n)); abort(); return 0;
	case 15:
	case 18:
	case 22:
	case 24:
	case 25:
	case 31:
	case 44:
	case 52:
	case 53:
	case 54:
	case 55:
	case 65:
	case 67:
	case 75:
	case 81:
	case 85:
	case 87:
	case 89:
	case 90:
	case 95:
	case 111:
	case 117:
	case 118:
		return PFfs_STATE_LABEL(n) = PFfs_state(PFfs_OP_LABEL(n), 0, 0);
	case 2:
	case 4:
	case 6:
	case 13:
	case 14:
	case 21:
	case 26:
	case 34:
	case 48:
	case 63:
	case 68:
	case 69:
	case 70:
	case 73:
	case 74:
	case 92:
	case 94:
	case 96:
	case 100:
	case 102:
	case 103:
	case 110:
	case 120:
	case 126:
	case 130:
		return PFfs_STATE_LABEL(n) = PFfs_state(PFfs_OP_LABEL(n), PFfs_label(PFfs_LEFT_CHILD(n)), 0);
	case 1:
	case 3:
	case 5:
	case 7:
	case 8:
	case 9:
	case 10:
	case 11:
	case 12:
	case 16:
	case 17:
	case 19:
	case 20:
	case 23:
	case 27:
	case 28:
	case 29:
	case 30:
	case 32:
	case 33:
	case 35:
	case 38:
	case 39:
	case 40:
	case 41:
	case 42:
	case 43:
	case 45:
	case 46:
	case 47:
	case 49:
	case 50:
	case 51:
	case 56:
	case 57:
	case 58:
	case 59:
	case 60:
	case 61:
	case 62:
	case 64:
	case 66:
	case 71:
	case 72:
	case 76:
	case 77:
	case 78:
	case 79:
	case 80:
	case 82:
	case 83:
	case 84:
	case 86:
	case 88:
	case 91:
	case 93:
	case 97:
	case 98:
	case 99:
	case 101:
	case 104:
	case 105:
	case 106:
	case 107:
	case 108:
	case 109:
	case 112:
	case 113:
	case 115:
	case 116:
	case 119:
	case 121:
	case 122:
	case 123:
	case 124:
	case 125:
	case 127:
	case 128:
	case 129:
		return PFfs_STATE_LABEL(n) = PFfs_state(PFfs_OP_LABEL(n), PFfs_label(PFfs_LEFT_CHILD(n)), PFfs_label(PFfs_RIGHT_CHILD(n)));
	}
}
#ifdef __STDC__
PFfs_NODEPTR_TYPE * PFfs_kids(PFfs_NODEPTR_TYPE p, int rulenumber, PFfs_NODEPTR_TYPE *kids) {
#else
PFfs_NODEPTR_TYPE * PFfs_kids(p, rulenumber, kids) PFfs_NODEPTR_TYPE p; int rulenumber; PFfs_NODEPTR_TYPE *kids; {
#endif
	PFfs_assert(p, PFfs_PANIC("NULL node pointer passed to PFfs_kids\n"));
	PFfs_assert(kids, PFfs_PANIC("NULL kids pointer passed to PFfs_kids\n"));
	switch (rulenumber) {
	default:
		PFfs_PANIC("Unknown Rule %d in PFfs_kids;\n", rulenumber);
		abort();
		/* NOTREACHED */
	case 10:
		kids[0] = PFfs_LEFT_CHILD(PFfs_LEFT_CHILD(PFfs_LEFT_CHILD(p)));
		kids[1] = PFfs_RIGHT_CHILD(PFfs_LEFT_CHILD(PFfs_LEFT_CHILD(p)));
		kids[2] = PFfs_RIGHT_CHILD(p);
		break;
	case 48:
		kids[0] = PFfs_LEFT_CHILD(PFfs_LEFT_CHILD(p));
		kids[1] = PFfs_RIGHT_CHILD(PFfs_LEFT_CHILD(p));
		kids[2] = PFfs_RIGHT_CHILD(p);
		break;
	case 49:
		kids[0] = PFfs_LEFT_CHILD(PFfs_LEFT_CHILD(p));
		kids[1] = PFfs_RIGHT_CHILD(PFfs_LEFT_CHILD(p));
		break;
	case 60:
		kids[0] = PFfs_LEFT_CHILD(p);
		kids[1] = PFfs_LEFT_CHILD(PFfs_RIGHT_CHILD(p));
		kids[2] = PFfs_RIGHT_CHILD(PFfs_RIGHT_CHILD(PFfs_RIGHT_CHILD(p)));
		break;
	case 61:
		kids[0] = PFfs_LEFT_CHILD(p);
		kids[1] = PFfs_LEFT_CHILD(PFfs_RIGHT_CHILD(p));
		kids[2] = PFfs_LEFT_CHILD(PFfs_LEFT_CHILD(PFfs_RIGHT_CHILD(PFfs_RIGHT_CHILD(p))));
		kids[3] = PFfs_RIGHT_CHILD(PFfs_RIGHT_CHILD(PFfs_RIGHT_CHILD(p)));
		break;
	case 63:
		kids[0] = PFfs_LEFT_CHILD(PFfs_LEFT_CHILD(PFfs_LEFT_CHILD(p)));
		kids[1] = PFfs_RIGHT_CHILD(PFfs_LEFT_CHILD(p));
		kids[2] = PFfs_RIGHT_CHILD(p);
		break;
	case 9:
	case 64:
		kids[0] = PFfs_LEFT_CHILD(PFfs_LEFT_CHILD(PFfs_LEFT_CHILD(p)));
		kids[1] = PFfs_RIGHT_CHILD(PFfs_LEFT_CHILD(PFfs_LEFT_CHILD(p)));
		kids[2] = PFfs_RIGHT_CHILD(PFfs_LEFT_CHILD(p));
		kids[3] = PFfs_RIGHT_CHILD(p);
		break;
	case 65:
		kids[0] = PFfs_LEFT_CHILD(PFfs_LEFT_CHILD(PFfs_LEFT_CHILD(PFfs_LEFT_CHILD(p))));
		kids[1] = PFfs_RIGHT_CHILD(PFfs_LEFT_CHILD(PFfs_LEFT_CHILD(p)));
		kids[2] = PFfs_RIGHT_CHILD(PFfs_LEFT_CHILD(p));
		kids[3] = PFfs_RIGHT_CHILD(p);
		break;
	case 66:
		kids[0] = PFfs_LEFT_CHILD(PFfs_LEFT_CHILD(PFfs_LEFT_CHILD(PFfs_LEFT_CHILD(p))));
		kids[1] = PFfs_RIGHT_CHILD(PFfs_LEFT_CHILD(PFfs_LEFT_CHILD(PFfs_LEFT_CHILD(p))));
		kids[2] = PFfs_RIGHT_CHILD(PFfs_LEFT_CHILD(PFfs_LEFT_CHILD(p)));
		kids[3] = PFfs_RIGHT_CHILD(PFfs_LEFT_CHILD(p));
		kids[4] = PFfs_RIGHT_CHILD(p);
		break;
	case 75:
	case 77:
		kids[0] = PFfs_LEFT_CHILD(PFfs_LEFT_CHILD(PFfs_LEFT_CHILD(PFfs_LEFT_CHILD(PFfs_LEFT_CHILD(p)))));
		kids[1] = PFfs_RIGHT_CHILD(PFfs_LEFT_CHILD(PFfs_LEFT_CHILD(p)));
		kids[2] = PFfs_RIGHT_CHILD(p);
		break;
	case 76:
	case 78:
		kids[0] = PFfs_LEFT_CHILD(PFfs_LEFT_CHILD(PFfs_LEFT_CHILD(PFfs_LEFT_CHILD(PFfs_LEFT_CHILD(p)))));
		kids[1] = PFfs_RIGHT_CHILD(PFfs_LEFT_CHILD(PFfs_LEFT_CHILD(PFfs_LEFT_CHILD(PFfs_LEFT_CHILD(p)))));
		kids[2] = PFfs_RIGHT_CHILD(PFfs_LEFT_CHILD(PFfs_LEFT_CHILD(p)));
		kids[3] = PFfs_RIGHT_CHILD(p);
		break;
	case 79:
		kids[0] = PFfs_LEFT_CHILD(p);
		kids[1] = PFfs_LEFT_CHILD(PFfs_LEFT_CHILD(PFfs_LEFT_CHILD(PFfs_RIGHT_CHILD(p))));
		kids[2] = PFfs_RIGHT_CHILD(PFfs_LEFT_CHILD(PFfs_LEFT_CHILD(PFfs_RIGHT_CHILD(p))));
		kids[3] = PFfs_RIGHT_CHILD(PFfs_LEFT_CHILD(PFfs_RIGHT_CHILD(p)));
		kids[4] = PFfs_LEFT_CHILD(PFfs_LEFT_CHILD(PFfs_RIGHT_CHILD(PFfs_RIGHT_CHILD(p))));
		kids[5] = PFfs_RIGHT_CHILD(PFfs_LEFT_CHILD(PFfs_RIGHT_CHILD(PFfs_RIGHT_CHILD(p))));
		break;
	case 94:
		kids[0] = PFfs_LEFT_CHILD(PFfs_LEFT_CHILD(p));
		break;
	case 36:
	case 37:
	case 39:
	case 97:
	case 99:
	case 213:
		kids[0] = PFfs_RIGHT_CHILD(p);
		break;
	case 82:
	case 83:
	case 92:
	case 147:
	case 148:
	case 152:
	case 185:
	case 188:
	case 189:
	case 198:
	case 199:
	case 201:
	case 210:
	case 222:
		kids[0] = PFfs_LEFT_CHILD(p);
		break;
	case 101:
	case 225:
		kids[0] = PFfs_LEFT_CHILD(p);
		kids[1] = PFfs_LEFT_CHILD(PFfs_RIGHT_CHILD(p));
		kids[2] = PFfs_RIGHT_CHILD(PFfs_RIGHT_CHILD(p));
		break;
	case 47:
	case 157:
	case 228:
		kids[0] = PFfs_LEFT_CHILD(PFfs_LEFT_CHILD(p));
		kids[1] = PFfs_RIGHT_CHILD(p);
		break;
	case 234:
		kids[0] = PFfs_LEFT_CHILD(PFfs_LEFT_CHILD(p));
		kids[1] = PFfs_RIGHT_CHILD(PFfs_LEFT_CHILD(p));
		kids[2] = PFfs_LEFT_CHILD(PFfs_RIGHT_CHILD(p));
		kids[3] = PFfs_RIGHT_CHILD(PFfs_RIGHT_CHILD(p));
		break;
	case 1:
	case 6:
	case 11:
	case 12:
	case 13:
	case 14:
	case 15:
	case 16:
	case 17:
	case 18:
	case 19:
	case 20:
	case 21:
	case 22:
	case 23:
	case 24:
	case 41:
	case 46:
	case 50:
	case 52:
	case 54:
	case 55:
	case 56:
	case 57:
	case 58:
	case 59:
	case 69:
	case 70:
	case 72:
	case 81:
	case 85:
	case 102:
	case 104:
	case 106:
	case 123:
	case 125:
	case 128:
	case 133:
	case 135:
	case 138:
	case 140:
	case 142:
	case 144:
	case 146:
	case 149:
	case 150:
	case 151:
	case 153:
	case 154:
	case 160:
	case 163:
	case 164:
	case 165:
	case 166:
	case 167:
	case 168:
	case 169:
	case 170:
	case 171:
	case 172:
	case 173:
	case 174:
	case 175:
	case 176:
	case 182:
	case 190:
	case 191:
	case 192:
	case 193:
	case 194:
	case 195:
	case 203:
	case 204:
	case 207:
	case 214:
	case 216:
	case 217:
	case 218:
	case 219:
	case 220:
	case 229:
	case 230:
	case 231:
	case 233:
	case 236:
		kids[0] = p;
		break;
	case 2:
	case 3:
	case 4:
	case 7:
	case 43:
	case 44:
	case 53:
	case 74:
	case 103:
	case 105:
	case 107:
	case 108:
	case 109:
	case 110:
	case 111:
	case 112:
	case 113:
	case 114:
	case 115:
	case 116:
	case 117:
	case 118:
	case 119:
	case 121:
	case 122:
	case 124:
	case 126:
	case 127:
	case 129:
	case 130:
	case 131:
	case 132:
	case 134:
	case 136:
	case 137:
	case 139:
	case 141:
	case 143:
	case 145:
	case 156:
	case 158:
	case 161:
	case 187:
	case 196:
	case 197:
	case 200:
	case 206:
	case 209:
	case 211:
	case 221:
	case 223:
	case 224:
	case 227:
	case 237:
		kids[0] = PFfs_LEFT_CHILD(p);
		kids[1] = PFfs_RIGHT_CHILD(p);
		break;
	case 5:
	case 8:
	case 25:
	case 26:
	case 27:
	case 28:
	case 29:
	case 30:
	case 31:
	case 32:
	case 33:
	case 34:
	case 35:
	case 38:
	case 40:
	case 42:
	case 45:
	case 51:
	case 62:
	case 68:
	case 71:
	case 73:
	case 80:
	case 84:
	case 86:
	case 87:
	case 88:
	case 89:
	case 90:
	case 91:
	case 93:
	case 95:
	case 96:
	case 98:
	case 100:
	case 155:
	case 177:
	case 178:
	case 179:
	case 180:
	case 181:
	case 183:
	case 184:
	case 186:
	case 202:
	case 205:
	case 208:
	case 212:
	case 215:
	case 226:
	case 232:
	case 235:
	case 238:
		break;
	}
	return kids;
}
#ifdef __STDC__
PFfs_NODEPTR_TYPE PFfs_child(PFfs_NODEPTR_TYPE p, int index) {
#else
PFfs_NODEPTR_TYPE PFfs_child(p, index) PFfs_NODEPTR_TYPE p; int index; {
#endif
	PFfs_assert(p, PFfs_PANIC("NULL pointer passed to PFfs_child\n"));
	switch (index) {
	case 0:
		return PFfs_LEFT_CHILD(p);
	case 1:
		return PFfs_RIGHT_CHILD(p);
	}
	PFfs_PANIC("Bad index %d in PFfs_child;\n", index);
	abort();
	return 0;
}
#ifdef __STDC__
int PFfs_op_label(PFfs_NODEPTR_TYPE p) {
#else
int PFfs_op_label(p) PFfs_NODEPTR_TYPE p; {
#endif
	PFfs_assert(p, PFfs_PANIC("NULL pointer passed to PFfs_op_label\n"));
	return PFfs_OP_LABEL(p);
}
#ifdef __STDC__
int PFfs_state_label(PFfs_NODEPTR_TYPE p) {
#else
int PFfs_state_label(p) PFfs_NODEPTR_TYPE p; {
#endif
	PFfs_assert(p, PFfs_PANIC("NULL pointer passed to PFfs_state_label\n"));
	return PFfs_STATE_LABEL(p);
}
#endif /* PFfs_INCLUDE_EXTRA */
#define PFfs_XRPCCall_NT 104
#define PFfs_OptTypeDeclForRec_NT 103
#define PFfs_RecursiveExpr_NT 102
#define PFfs_NewNameExpr_NT 101
#define PFfs_TargetExpr_NT 100
#define PFfs_SourceExpr_NT 99
#define PFfs_TransformBinding_NT 98
#define PFfs_TransformBindings_NT 97
#define PFfs_TransformExpr_NT 96
#define PFfs_RenameExpr_NT 95
#define PFfs_ReplaceExpr_NT 94
#define PFfs_DeleteExpr_NT 93
#define PFfs_InsertExpr_NT 92
#define PFfs_RevalidationDecl_NT 91
#define PFfs_Pragmas_NT 90
#define PFfs_ExtensionExpr_NT 89
#define PFfs_AttributeValue_NT 88
#define PFfs_OptContentExpr_NT 87
#define PFfs_TagName_NT 86
#define PFfs_DocumentConstructor_NT 85
#define PFfs_PIConstructor_NT 84
#define PFfs_CommentConstructor_NT 83
#define PFfs_TextConstructor_NT 82
#define PFfs_AttributeConstructor_NT 81
#define PFfs_ElementConstructor_NT 80
#define PFfs_Constructor_NT 79
#define PFfs_UnorderedExpr_NT 78
#define PFfs_OrderedExpr_NT 77
#define PFfs_FuncArgList_NT 76
#define PFfs_FunctionCall_NT 75
#define PFfs_ContextItemExpr_NT 74
#define PFfs_ParenthesizedExpr_NT 73
#define PFfs_Var_NT 72
#define PFfs_StringLiteral_NT 71
#define PFfs_DoubleLiteral_NT 70
#define PFfs_DecimalLiteral_NT 69
#define PFfs_IntegerLiteral_NT 68
#define PFfs_NumericLiteral_NT 67
#define PFfs_Literal_NT 66
#define PFfs_PrimaryExpr_NT 65
#define PFfs_FilterExpr_NT 64
#define PFfs_LocationPath_NT 63
#define PFfs_PathExpr_NT 62
#define PFfs_ValidateExpr_NT 61
#define PFfs_ValueExpr_NT 60
#define PFfs_UnaryExpr_NT 59
#define PFfs_CastExpr_NT 58
#define PFfs_CastableExpr_NT 57
#define PFfs_TreatExpr_NT 56
#define PFfs_InstanceofExpr_NT 55
#define PFfs_IntersectExceptExpr_NT 54
#define PFfs_UnionExpr_NT 53
#define PFfs_MultiplicativeExpr_NT 52
#define PFfs_AdditiveExpr_NT 51
#define PFfs_RangeExpr_NT 50
#define PFfs_ComparisonExpr_NT 49
#define PFfs_AndExpr_NT 48
#define PFfs_OrExpr_NT 47
#define PFfs_IfExpr_NT 46
#define PFfs_SchemaType_NT 45
#define PFfs_ElemOrAttrType_NT 44
#define PFfs_NodeTest_NT 43
#define PFfs_AtomType_NT 42
#define PFfs_ItemType_NT 41
#define PFfs_SingleType_NT 40
#define PFfs_SequenceType_NT 39
#define PFfs_OptCaseVar_NT 38
#define PFfs_TypeswitchExpr_NT 37
#define PFfs_QuantifiedExpr_NT 36
#define PFfs_OrderSpecList_NT 35
#define PFfs_OptWhereClause__NT 34
#define PFfs_TypeDeclaration_NT 33
#define PFfs_OptPosVar_NT 32
#define PFfs_OptBindExpr_NT 31
#define PFfs_FLWORExpr_NT 30
#define PFfs_ExprSingle_NT 29
#define PFfs_Expr_NT 28
#define PFfs_OptExpr_NT 27
#define PFfs_QueryBody_NT 26
#define PFfs_VariableDeclaration_NT 25
#define PFfs_OptSequenceType__NT 24
#define PFfs_Param_NT 23
#define PFfs_ParamList__NT 22
#define PFfs_OptParamList__NT 21
#define PFfs_OptImportLoc__NT 20
#define PFfs_ModuleImport_NT 19
#define PFfs_SchemaImport_NT 18
#define PFfs_NamespaceDecl_NT 17
#define PFfs_CopyNamespacesDecl_NT 16
#define PFfs_EmptyOrderingDecl_NT 15
#define PFfs_OrderingModeDecl_NT 14
#define PFfs_OptionDecl_NT 13
#define PFfs_ConstructionDecl_NT 12
#define PFfs_BaseURIDecl_NT 11
#define PFfs_DefaultCollationDecl_NT 10
#define PFfs_BoundarySpaceDecl_NT 9
#define PFfs_DefaultNamespaceDecl_NT 8
#define PFfs_DeclImport_NT 7
#define PFfs_DeclsImports_NT 6
#define PFfs_Prolog_NT 5
#define PFfs_ModuleDecl_NT 4
#define PFfs_LibraryModule_NT 3
#define PFfs_MainModule_NT 2
#define PFfs_Query_NT 1
#define PFfs_NT 104
char * PFfs_opname[] = {
	0, /* 0 */
	"and", /* 1 */
	"apply", /* 2 */
	"args", /* 3 */
	"atom_ty", /* 4 */
	"attr", /* 5 */
	"base_uri", /* 6 */
	"bind", /* 7 */
	"binds", /* 8 */
	"case_", /* 9 */
	"cases", /* 10 */
	"cast", /* 11 */
	"castable", /* 12 */
	"coll_decl", /* 13 */
	"comment", /* 14 */
	"constr_decl", /* 15 */
	"contseq", /* 16 */
	"decl_imps", /* 17 */
	"def_order", /* 18 */
	"default_", /* 19 */
	"div", /* 20 */
	"doc", /* 21 */
	"dot", /* 22 */
	"elem", /* 23 */
	"empty_seq", /* 24 */
	"empty_ty", /* 25 */
	"ens_decl", /* 26 */
	"eq", /* 27 */
	"every", /* 28 */
	"except", /* 29 */
	"exprseq", /* 30 */
	"external", /* 31 */
	"ext_expr", /* 32 */
	"flwr", /* 33 */
	"fns_decl", /* 34 */
	"fun", /* 35 */
	0, /* 36 */
	0, /* 37 */
	"fun_sig", /* 38 */
	"ge", /* 39 */
	"gt", /* 40 */
	"gtgt", /* 41 */
	"idiv", /* 42 */
	"if_", /* 43 */
	"copy_ns", /* 44 */
	"instof", /* 45 */
	"intersect", /* 46 */
	"is", /* 47 */
	"item_ty", /* 48 */
	"le", /* 49 */
	"let", /* 50 */
	"lib_mod", /* 51 */
	"lit_dbl", /* 52 */
	"lit_dec", /* 53 */
	"lit_int", /* 54 */
	"lit_str", /* 55 */
	"locpath", /* 56 */
	"lt", /* 57 */
	"ltlt", /* 58 */
	"main_mod", /* 59 */
	"minus", /* 60 */
	"mod", /* 61 */
	"mod_imp", /* 62 */
	"mod_ns", /* 63 */
	"mult", /* 64 */
	"named_ty", /* 65 */
	"ne", /* 66 */
	"nil", /* 67 */
	"node_ty", /* 68 */
	"ns_decl", /* 69 */
	"option", /* 70 */
	"or", /* 71 */
	"ord_ret", /* 72 */
	"orderby", /* 73 */
	"ordered", /* 74 */
	"ordering_mode", /* 75 */
	"orderspecs", /* 76 */
	"param", /* 77 */
	"params", /* 78 */
	"pi", /* 79 */
	"plus", /* 80 */
	"pragma", /* 81 */
	"pragmas", /* 82 */
	"pred", /* 83 */
	"range", /* 84 */
	"req_name", /* 85 */
	"req_ty", /* 86 */
	"root_", /* 87 */
	"schm_ats", /* 88 */
	"schm_attr", /* 89 */
	"schm_elem", /* 90 */
	"schm_imp", /* 91 */
	"seq_ty", /* 92 */
	"some", /* 93 */
	"step", /* 94 */
	"tag", /* 95 */
	"text", /* 96 */
	"then_else", /* 97 */
	"treat", /* 98 */
	"typeswitch", /* 99 */
	"uminus", /* 100 */
	"union_", /* 101 */
	"unordered", /* 102 */
	"uplus", /* 103 */
	"val_eq", /* 104 */
	"val_ge", /* 105 */
	"val_gt", /* 106 */
	"val_le", /* 107 */
	"val_lt", /* 108 */
	"val_ne", /* 109 */
	"validate", /* 110 */
	"var", /* 111 */
	"var_decl", /* 112 */
	"var_type", /* 113 */
	0, /* 114 */
	"vars", /* 115 */
	"where", /* 116 */
	"boundspc_decl", /* 117 */
	"revalid", /* 118 */
	"insert", /* 119 */
	"delete", /* 120 */
	"replace", /* 121 */
	"rename", /* 122 */
	"transform", /* 123 */
	"modify", /* 124 */
	"transbinds", /* 125 */
	"stmt_ty", /* 126 */
	"recursion", /* 127 */
	"seed", /* 128 */
	"xrpc", /* 129 */
	"docmgmt_ty"
};
char PFfs_arity[] = {
	-1, /* 0 */
	2, /* 1 */
	1, /* 2 */
	2, /* 3 */
	1, /* 4 */
	2, /* 5 */
	1, /* 6 */
	2, /* 7 */
	2, /* 8 */
	2, /* 9 */
	2, /* 10 */
	2, /* 11 */
	2, /* 12 */
	1, /* 13 */
	1, /* 14 */
	0, /* 15 */
	2, /* 16 */
	2, /* 17 */
	0, /* 18 */
	2, /* 19 */
	2, /* 20 */
	1, /* 21 */
	0, /* 22 */
	2, /* 23 */
	0, /* 24 */
	0, /* 25 */
	1, /* 26 */
	2, /* 27 */
	2, /* 28 */
	2, /* 29 */
	2, /* 30 */
	0, /* 31 */
	2, /* 32 */
	2, /* 33 */
	1, /* 34 */
	2, /* 35 */
	-1, /* 36 */
	-1, /* 37 */
	2, /* 38 */
	2, /* 39 */
	2, /* 40 */
	2, /* 41 */
	2, /* 42 */
	2, /* 43 */
	0, /* 44 */
	2, /* 45 */
	2, /* 46 */
	2, /* 47 */
	1, /* 48 */
	2, /* 49 */
	2, /* 50 */
	2, /* 51 */
	0, /* 52 */
	0, /* 53 */
	0, /* 54 */
	0, /* 55 */
	2, /* 56 */
	2, /* 57 */
	2, /* 58 */
	2, /* 59 */
	2, /* 60 */
	2, /* 61 */
	2, /* 62 */
	1, /* 63 */
	2, /* 64 */
	0, /* 65 */
	2, /* 66 */
	0, /* 67 */
	1, /* 68 */
	1, /* 69 */
	1, /* 70 */
	2, /* 71 */
	2, /* 72 */
	1, /* 73 */
	1, /* 74 */
	0, /* 75 */
	2, /* 76 */
	2, /* 77 */
	2, /* 78 */
	2, /* 79 */
	2, /* 80 */
	0, /* 81 */
	2, /* 82 */
	2, /* 83 */
	2, /* 84 */
	0, /* 85 */
	2, /* 86 */
	0, /* 87 */
	2, /* 88 */
	0, /* 89 */
	0, /* 90 */
	2, /* 91 */
	1, /* 92 */
	2, /* 93 */
	1, /* 94 */
	0, /* 95 */
	1, /* 96 */
	2, /* 97 */
	2, /* 98 */
	2, /* 99 */
	1, /* 100 */
	2, /* 101 */
	1, /* 102 */
	1, /* 103 */
	2, /* 104 */
	2, /* 105 */
	2, /* 106 */
	2, /* 107 */
	2, /* 108 */
	2, /* 109 */
	1, /* 110 */
	0, /* 111 */
	2, /* 112 */
	2, /* 113 */
	-1, /* 114 */
	2, /* 115 */
	2, /* 116 */
	0, /* 117 */
	0, /* 118 */
	2, /* 119 */
	1, /* 120 */
	2, /* 121 */
	2, /* 122 */
	2, /* 123 */
	2, /* 124 */
	2, /* 125 */
	1, /* 126 */
	2, /* 127 */
	2, /* 128 */
	2, /* 129 */
	1
};
int PFfs_max_op = 130;
int PFfs_max_state = 166;
#define PFfs_Max_state 166
char *PFfs_string[] = {
	0,
	"Query: MainModule",
	"Query: lib_mod(ModuleDecl, Prolog)",
	"MainModule: main_mod(Prolog, QueryBody)",
	"LibraryModule: lib_mod(ModuleDecl, Prolog)",
	"ModuleDecl: mod_ns(lit_str)",
	"Prolog: DeclsImports",
	"DeclsImports: decl_imps(DeclImport, DeclsImports)",
	"DeclsImports: nil",
	"DeclsImports: decl_imps(fun(fun_sig(OptParamList_, SequenceType), Expr), DeclsImports)",
	"DeclsImports: decl_imps(fun(fun_sig(OptParamList_, SequenceType), external), DeclsImports)",
	"DeclImport: DefaultNamespaceDecl",
	"DeclImport: BoundarySpaceDecl",
	"DeclImport: DefaultCollationDecl",
	"DeclImport: BaseURIDecl",
	"DeclImport: ConstructionDecl",
	"DeclImport: OrderingModeDecl",
	"DeclImport: EmptyOrderingDecl",
	"DeclImport: CopyNamespacesDecl",
	"DeclImport: NamespaceDecl",
	"DeclImport: OptionDecl",
	"DeclImport: SchemaImport",
	"DeclImport: ModuleImport",
	"DeclImport: VariableDeclaration",
	"DeclImport: LibraryModule",
	"DefaultNamespaceDecl: ens_decl(lit_str)",
	"DefaultNamespaceDecl: fns_decl(lit_str)",
	"BoundarySpaceDecl: boundspc_decl",
	"DefaultCollationDecl: coll_decl(lit_str)",
	"BaseURIDecl: base_uri(lit_str)",
	"ConstructionDecl: constr_decl",
	"OptionDecl: option(lit_str)",
	"OrderingModeDecl: ordering_mode",
	"EmptyOrderingDecl: def_order",
	"CopyNamespacesDecl: copy_ns",
	"NamespaceDecl: ns_decl(lit_str)",
	"SchemaImport: schm_imp(lit_str, OptImportLoc_)",
	"ModuleImport: mod_imp(lit_str, OptImportLoc_)",
	"OptImportLoc_: nil",
	"OptImportLoc_: schm_ats(lit_str, OptImportLoc_)",
	"OptParamList_: nil",
	"OptParamList_: ParamList_",
	"ParamList_: nil",
	"ParamList_: params(Param, ParamList_)",
	"Param: param(SequenceType, Var)",
	"OptSequenceType_: nil",
	"OptSequenceType_: SequenceType",
	"VariableDeclaration: var_decl(var_type(Var, nil), OptExpr)",
	"VariableDeclaration: var_decl(var_type(Var, SequenceType), OptExpr)",
	"VariableDeclaration: var_decl(var_type(Var, OptSequenceType_), external)",
	"QueryBody: OptExpr",
	"OptExpr: empty_seq",
	"OptExpr: Expr",
	"Expr: exprseq(OptExpr, OptExpr)",
	"Expr: ExprSingle",
	"ExprSingle: FLWORExpr",
	"ExprSingle: QuantifiedExpr",
	"ExprSingle: TypeswitchExpr",
	"ExprSingle: IfExpr",
	"ExprSingle: OrExpr",
	"FLWORExpr: flwr(OptBindExpr, where(OptWhereClause_, ord_ret(nil, OptExpr)))",
	"FLWORExpr: flwr(OptBindExpr, where(OptWhereClause_, ord_ret(orderby(OrderSpecList), OptExpr)))",
	"OptBindExpr: nil",
	"OptBindExpr: binds(let(var_type(Var, nil), OptExpr), OptBindExpr)",
	"OptBindExpr: binds(let(var_type(Var, TypeDeclaration), OptExpr), OptBindExpr)",
	"OptBindExpr: binds(bind(vars(var_type(Var, nil), OptPosVar), OptExpr), OptBindExpr)",
	"OptBindExpr: binds(bind(vars(var_type(Var, TypeDeclaration), OptPosVar), OptExpr), OptBindExpr)",
	0,
	"OptPosVar: nil",
	"OptPosVar: Var",
	"TypeDeclaration: SequenceType",
	"OptWhereClause_: nil",
	"OptWhereClause_: OptExpr",
	"OrderSpecList: nil",
	"OrderSpecList: orderspecs(OptExpr, OrderSpecList)",
	"QuantifiedExpr: some(binds(bind(vars(var_type(Var, nil), nil), OptExpr), nil), OptExpr)",
	"QuantifiedExpr: some(binds(bind(vars(var_type(Var, TypeDeclaration), nil), OptExpr), nil), OptExpr)",
	"QuantifiedExpr: every(binds(bind(vars(var_type(Var, nil), nil), OptExpr), nil), OptExpr)",
	"QuantifiedExpr: every(binds(bind(vars(var_type(Var, TypeDeclaration), nil), OptExpr), nil), OptExpr)",
	"TypeswitchExpr: typeswitch(OptExpr, cases(case_(var_type(OptCaseVar, SequenceType), OptExpr), cases(default_(OptCaseVar, OptExpr), nil)))",
	"OptCaseVar: nil",
	"OptCaseVar: Var",
	"SequenceType: seq_ty(ItemType)",
	"SingleType: seq_ty(AtomType)",
	"ItemType: empty_ty",
	"ItemType: NodeTest",
	"ItemType: item_ty(nil)",
	"ItemType: atom_ty(nil)",
	"ItemType: schm_elem",
	"ItemType: schm_attr",
	"AtomType: atom_ty(nil)",
	"NodeTest: node_ty(nil)",
	"NodeTest: node_ty(ElemOrAttrType)",
	"NodeTest: node_ty(node_ty(nil))",
	"NodeTest: node_ty(node_ty(ElemOrAttrType))",
	"NodeTest: node_ty(lit_str)",
	"ElemOrAttrType: req_ty(nil, nil)",
	"ElemOrAttrType: req_ty(nil, SchemaType)",
	"ElemOrAttrType: req_ty(req_name, nil)",
	"ElemOrAttrType: req_ty(req_name, SchemaType)",
	"SchemaType: named_ty",
	"IfExpr: if_(OptExpr, then_else(OptExpr, OptExpr))",
	"OrExpr: AndExpr",
	"OrExpr: or(AndExpr, OrExpr)",
	"AndExpr: ComparisonExpr",
	"AndExpr: and(ComparisonExpr, AndExpr)",
	"ComparisonExpr: RangeExpr",
	"ComparisonExpr: eq(RangeExpr, RangeExpr)",
	"ComparisonExpr: ne(RangeExpr, RangeExpr)",
	"ComparisonExpr: lt(RangeExpr, RangeExpr)",
	"ComparisonExpr: le(RangeExpr, RangeExpr)",
	"ComparisonExpr: gt(RangeExpr, RangeExpr)",
	"ComparisonExpr: ge(RangeExpr, RangeExpr)",
	"ComparisonExpr: val_eq(RangeExpr, RangeExpr)",
	"ComparisonExpr: val_ne(RangeExpr, RangeExpr)",
	"ComparisonExpr: val_lt(RangeExpr, RangeExpr)",
	"ComparisonExpr: val_le(RangeExpr, RangeExpr)",
	"ComparisonExpr: val_gt(RangeExpr, RangeExpr)",
	"ComparisonExpr: val_ge(RangeExpr, RangeExpr)",
	"ComparisonExpr: is(RangeExpr, RangeExpr)",
	0,
	"ComparisonExpr: ltlt(RangeExpr, RangeExpr)",
	"ComparisonExpr: gtgt(RangeExpr, RangeExpr)",
	"RangeExpr: AdditiveExpr",
	"RangeExpr: range(RangeExpr, RangeExpr)",
	"AdditiveExpr: MultiplicativeExpr",
	"AdditiveExpr: plus(AdditiveExpr, MultiplicativeExpr)",
	"AdditiveExpr: minus(AdditiveExpr, MultiplicativeExpr)",
	"MultiplicativeExpr: UnionExpr",
	"MultiplicativeExpr: mult(UnionExpr, MultiplicativeExpr)",
	"MultiplicativeExpr: div(UnionExpr, MultiplicativeExpr)",
	"MultiplicativeExpr: idiv(UnionExpr, MultiplicativeExpr)",
	"MultiplicativeExpr: mod(UnionExpr, MultiplicativeExpr)",
	"UnionExpr: IntersectExceptExpr",
	"UnionExpr: union_(IntersectExceptExpr, UnionExpr)",
	"IntersectExceptExpr: InstanceofExpr",
	"IntersectExceptExpr: intersect(InstanceofExpr, IntersectExceptExpr)",
	"IntersectExceptExpr: except(InstanceofExpr, IntersectExceptExpr)",
	"InstanceofExpr: TreatExpr",
	"InstanceofExpr: instof(TreatExpr, SequenceType)",
	"TreatExpr: CastableExpr",
	"TreatExpr: treat(CastableExpr, SequenceType)",
	"CastableExpr: CastExpr",
	"CastableExpr: castable(CastExpr, SingleType)",
	"CastExpr: UnaryExpr",
	"CastExpr: cast(UnaryExpr, SingleType)",
	"UnaryExpr: ValueExpr",
	"UnaryExpr: uminus(ValueExpr)",
	"UnaryExpr: uplus(ValueExpr)",
	"ValueExpr: ValidateExpr",
	"ValueExpr: PathExpr",
	"ValueExpr: ExtensionExpr",
	"ValidateExpr: validate(OptExpr)",
	"PathExpr: FilterExpr",
	"PathExpr: LocationPath",
	"LocationPath: root_",
	"LocationPath: locpath(PathExpr, PathExpr)",
	"LocationPath: locpath(step(NodeTest), PathExpr)",
	"LocationPath: pred(LocationPath, OptExpr)",
	0,
	"FilterExpr: PrimaryExpr",
	"FilterExpr: pred(PathExpr, OptExpr)",
	0,
	"PrimaryExpr: Literal",
	"PrimaryExpr: Var",
	"PrimaryExpr: ParenthesizedExpr",
	"PrimaryExpr: ContextItemExpr",
	"PrimaryExpr: FunctionCall",
	"PrimaryExpr: Constructor",
	"PrimaryExpr: OrderedExpr",
	"PrimaryExpr: UnorderedExpr",
	"PrimaryExpr: XRPCCall",
	"Literal: NumericLiteral",
	"Literal: StringLiteral",
	"NumericLiteral: IntegerLiteral",
	"NumericLiteral: DecimalLiteral",
	"NumericLiteral: DoubleLiteral",
	"IntegerLiteral: lit_int",
	"DecimalLiteral: lit_dec",
	"DoubleLiteral: lit_dbl",
	"StringLiteral: lit_str",
	"Var: var",
	"ParenthesizedExpr: OptExpr",
	"ContextItemExpr: dot",
	"FunctionCall: apply(nil)",
	"FunctionCall: apply(FuncArgList)",
	"FuncArgList: nil",
	"FuncArgList: args(OptExpr, FuncArgList)",
	"OrderedExpr: ordered(OptExpr)",
	"UnorderedExpr: unordered(OptExpr)",
	"Constructor: ElementConstructor",
	"Constructor: AttributeConstructor",
	"Constructor: TextConstructor",
	"Constructor: CommentConstructor",
	"Constructor: PIConstructor",
	"Constructor: DocumentConstructor",
	"ElementConstructor: elem(TagName, OptContentExpr)",
	"AttributeConstructor: attr(TagName, AttributeValue)",
	"TextConstructor: text(OptExpr)",
	"CommentConstructor: comment(OptExpr)",
	"PIConstructor: pi(OptExpr, OptContentExpr)",
	"DocumentConstructor: doc(OptExpr)",
	"TagName: tag",
	"TagName: OptExpr",
	"OptContentExpr: Expr",
	"OptContentExpr: empty_seq",
	"OptContentExpr: contseq(OptContentExpr, OptContentExpr)",
	"AttributeValue: Expr",
	"AttributeValue: empty_seq",
	"AttributeValue: contseq(AttributeValue, AttributeValue)",
	"ExtensionExpr: ext_expr(Pragmas, nil)",
	"ExtensionExpr: ext_expr(Pragmas, Expr)",
	"Pragmas: nil",
	"Pragmas: pragmas(pragma, Pragmas)",
	"DeclImport: RevalidationDecl",
	"RevalidationDecl: revalid",
	"ExprSingle: InsertExpr",
	"ExprSingle: DeleteExpr",
	"ExprSingle: RenameExpr",
	"ExprSingle: ReplaceExpr",
	"ExprSingle: TransformExpr",
	"InsertExpr: insert(SourceExpr, TargetExpr)",
	"DeleteExpr: delete(SourceExpr)",
	"ReplaceExpr: replace(TargetExpr, ExprSingle)",
	"RenameExpr: rename(SourceExpr, NewNameExpr)",
	"TransformExpr: transform(TransformBindings, modify(ExprSingle, ExprSingle))",
	"TransformBindings: nil",
	"TransformBindings: transbinds(TransformBinding, TransformBindings)",
	"TransformBinding: let(var_type(Var, nil), OptExpr)",
	"SourceExpr: ExprSingle",
	"TargetExpr: ExprSingle",
	"NewNameExpr: ExprSingle",
	"SequenceType: seq_ty(stmt_ty(nil))",
	"ExprSingle: RecursiveExpr",
	"RecursiveExpr: recursion(var_type(Var, OptTypeDeclForRec), seed(ExprSingle, ExprSingle))",
	"OptTypeDeclForRec: nil",
	"OptTypeDeclForRec: SequenceType",
	"XRPCCall: xrpc(Expr, FunctionCall)",
	"SequenceType: seq_ty(docmgmt_ty(nil))",
};
int PFfs_max_rule = 238;
#define PFfs_Max_rule 238
short PFfs_rule_descriptor_0[] = { 0, 0 };
short PFfs_rule_descriptor_1[] = {   -1,   -2, };
short PFfs_rule_descriptor_2[] = {   -1,   51,   -4,   -5, };
short PFfs_rule_descriptor_3[] = {   -2,   59,   -5,  -26, };
short PFfs_rule_descriptor_4[] = {   -3,   51,   -4,   -5, };
short PFfs_rule_descriptor_5[] = {   -4,   63,   55, };
short PFfs_rule_descriptor_6[] = {   -5,   -6, };
short PFfs_rule_descriptor_7[] = {   -6,   17,   -7,   -6, };
short PFfs_rule_descriptor_8[] = {   -6,   67, };
short PFfs_rule_descriptor_9[] = {   -6,   17,   35,   38,  -21,  -39,  -28,   -6, };
short PFfs_rule_descriptor_10[] = {   -6,   17,   35,   38,  -21,  -39,   31,   -6, };
short PFfs_rule_descriptor_11[] = {   -7,   -8, };
short PFfs_rule_descriptor_12[] = {   -7,   -9, };
short PFfs_rule_descriptor_13[] = {   -7,  -10, };
short PFfs_rule_descriptor_14[] = {   -7,  -11, };
short PFfs_rule_descriptor_15[] = {   -7,  -12, };
short PFfs_rule_descriptor_16[] = {   -7,  -14, };
short PFfs_rule_descriptor_17[] = {   -7,  -15, };
short PFfs_rule_descriptor_18[] = {   -7,  -16, };
short PFfs_rule_descriptor_19[] = {   -7,  -17, };
short PFfs_rule_descriptor_20[] = {   -7,  -13, };
short PFfs_rule_descriptor_21[] = {   -7,  -18, };
short PFfs_rule_descriptor_22[] = {   -7,  -19, };
short PFfs_rule_descriptor_23[] = {   -7,  -25, };
short PFfs_rule_descriptor_24[] = {   -7,   -3, };
short PFfs_rule_descriptor_25[] = {   -8,   26,   55, };
short PFfs_rule_descriptor_26[] = {   -8,   34,   55, };
short PFfs_rule_descriptor_27[] = {   -9,  117, };
short PFfs_rule_descriptor_28[] = {  -10,   13,   55, };
short PFfs_rule_descriptor_29[] = {  -11,    6,   55, };
short PFfs_rule_descriptor_30[] = {  -12,   15, };
short PFfs_rule_descriptor_31[] = {  -13,   70,   55, };
short PFfs_rule_descriptor_32[] = {  -14,   75, };
short PFfs_rule_descriptor_33[] = {  -15,   18, };
short PFfs_rule_descriptor_34[] = {  -16,   44, };
short PFfs_rule_descriptor_35[] = {  -17,   69,   55, };
short PFfs_rule_descriptor_36[] = {  -18,   91,   55,  -20, };
short PFfs_rule_descriptor_37[] = {  -19,   62,   55,  -20, };
short PFfs_rule_descriptor_38[] = {  -20,   67, };
short PFfs_rule_descriptor_39[] = {  -20,   88,   55,  -20, };
short PFfs_rule_descriptor_40[] = {  -21,   67, };
short PFfs_rule_descriptor_41[] = {  -21,  -22, };
short PFfs_rule_descriptor_42[] = {  -22,   67, };
short PFfs_rule_descriptor_43[] = {  -22,   78,  -23,  -22, };
short PFfs_rule_descriptor_44[] = {  -23,   77,  -39,  -72, };
short PFfs_rule_descriptor_45[] = {  -24,   67, };
short PFfs_rule_descriptor_46[] = {  -24,  -39, };
short PFfs_rule_descriptor_47[] = {  -25,  112,  113,  -72,   67,  -27, };
short PFfs_rule_descriptor_48[] = {  -25,  112,  113,  -72,  -39,  -27, };
short PFfs_rule_descriptor_49[] = {  -25,  112,  113,  -72,  -24,   31, };
short PFfs_rule_descriptor_50[] = {  -26,  -27, };
short PFfs_rule_descriptor_51[] = {  -27,   24, };
short PFfs_rule_descriptor_52[] = {  -27,  -28, };
short PFfs_rule_descriptor_53[] = {  -28,   30,  -27,  -27, };
short PFfs_rule_descriptor_54[] = {  -28,  -29, };
short PFfs_rule_descriptor_55[] = {  -29,  -30, };
short PFfs_rule_descriptor_56[] = {  -29,  -36, };
short PFfs_rule_descriptor_57[] = {  -29,  -37, };
short PFfs_rule_descriptor_58[] = {  -29,  -46, };
short PFfs_rule_descriptor_59[] = {  -29,  -47, };
short PFfs_rule_descriptor_60[] = {  -30,   33,  -31,  116,  -34,   72,   67,  -27, };
short PFfs_rule_descriptor_61[] = {  -30,   33,  -31,  116,  -34,   72,   73,  -35,  -27, };
short PFfs_rule_descriptor_62[] = {  -31,   67, };
short PFfs_rule_descriptor_63[] = {  -31,    8,   50,  113,  -72,   67,  -27,  -31, };
short PFfs_rule_descriptor_64[] = {  -31,    8,   50,  113,  -72,  -33,  -27,  -31, };
short PFfs_rule_descriptor_65[] = {  -31,    8,    7,  115,  113,  -72,   67,  -32,  -27,  -31, };
short PFfs_rule_descriptor_66[] = {  -31,    8,    7,  115,  113,  -72,  -33,  -32,  -27,  -31, };
short PFfs_rule_descriptor_68[] = {  -32,   67, };
short PFfs_rule_descriptor_69[] = {  -32,  -72, };
short PFfs_rule_descriptor_70[] = {  -33,  -39, };
short PFfs_rule_descriptor_71[] = {  -34,   67, };
short PFfs_rule_descriptor_72[] = {  -34,  -27, };
short PFfs_rule_descriptor_73[] = {  -35,   67, };
short PFfs_rule_descriptor_74[] = {  -35,   76,  -27,  -35, };
short PFfs_rule_descriptor_75[] = {  -36,   93,    8,    7,  115,  113,  -72,   67,   67,  -27,   67,  -27, };
short PFfs_rule_descriptor_76[] = {  -36,   93,    8,    7,  115,  113,  -72,  -33,   67,  -27,   67,  -27, };
short PFfs_rule_descriptor_77[] = {  -36,   28,    8,    7,  115,  113,  -72,   67,   67,  -27,   67,  -27, };
short PFfs_rule_descriptor_78[] = {  -36,   28,    8,    7,  115,  113,  -72,  -33,   67,  -27,   67,  -27, };
short PFfs_rule_descriptor_79[] = {  -37,   99,  -27,   10,    9,  113,  -38,  -39,  -27,   10,   19,  -38,  -27,   67, };
short PFfs_rule_descriptor_80[] = {  -38,   67, };
short PFfs_rule_descriptor_81[] = {  -38,  -72, };
short PFfs_rule_descriptor_82[] = {  -39,   92,  -41, };
short PFfs_rule_descriptor_83[] = {  -40,   92,  -42, };
short PFfs_rule_descriptor_84[] = {  -41,   25, };
short PFfs_rule_descriptor_85[] = {  -41,  -43, };
short PFfs_rule_descriptor_86[] = {  -41,   48,   67, };
short PFfs_rule_descriptor_87[] = {  -41,    4,   67, };
short PFfs_rule_descriptor_88[] = {  -41,   90, };
short PFfs_rule_descriptor_89[] = {  -41,   89, };
short PFfs_rule_descriptor_90[] = {  -42,    4,   67, };
short PFfs_rule_descriptor_91[] = {  -43,   68,   67, };
short PFfs_rule_descriptor_92[] = {  -43,   68,  -44, };
short PFfs_rule_descriptor_93[] = {  -43,   68,   68,   67, };
short PFfs_rule_descriptor_94[] = {  -43,   68,   68,  -44, };
short PFfs_rule_descriptor_95[] = {  -43,   68,   55, };
short PFfs_rule_descriptor_96[] = {  -44,   86,   67,   67, };
short PFfs_rule_descriptor_97[] = {  -44,   86,   67,  -45, };
short PFfs_rule_descriptor_98[] = {  -44,   86,   85,   67, };
short PFfs_rule_descriptor_99[] = {  -44,   86,   85,  -45, };
short PFfs_rule_descriptor_100[] = {  -45,   65, };
short PFfs_rule_descriptor_101[] = {  -46,   43,  -27,   97,  -27,  -27, };
short PFfs_rule_descriptor_102[] = {  -47,  -48, };
short PFfs_rule_descriptor_103[] = {  -47,   71,  -48,  -47, };
short PFfs_rule_descriptor_104[] = {  -48,  -49, };
short PFfs_rule_descriptor_105[] = {  -48,    1,  -49,  -48, };
short PFfs_rule_descriptor_106[] = {  -49,  -50, };
short PFfs_rule_descriptor_107[] = {  -49,   27,  -50,  -50, };
short PFfs_rule_descriptor_108[] = {  -49,   66,  -50,  -50, };
short PFfs_rule_descriptor_109[] = {  -49,   57,  -50,  -50, };
short PFfs_rule_descriptor_110[] = {  -49,   49,  -50,  -50, };
short PFfs_rule_descriptor_111[] = {  -49,   40,  -50,  -50, };
short PFfs_rule_descriptor_112[] = {  -49,   39,  -50,  -50, };
short PFfs_rule_descriptor_113[] = {  -49,  104,  -50,  -50, };
short PFfs_rule_descriptor_114[] = {  -49,  109,  -50,  -50, };
short PFfs_rule_descriptor_115[] = {  -49,  108,  -50,  -50, };
short PFfs_rule_descriptor_116[] = {  -49,  107,  -50,  -50, };
short PFfs_rule_descriptor_117[] = {  -49,  106,  -50,  -50, };
short PFfs_rule_descriptor_118[] = {  -49,  105,  -50,  -50, };
short PFfs_rule_descriptor_119[] = {  -49,   47,  -50,  -50, };
short PFfs_rule_descriptor_121[] = {  -49,   58,  -50,  -50, };
short PFfs_rule_descriptor_122[] = {  -49,   41,  -50,  -50, };
short PFfs_rule_descriptor_123[] = {  -50,  -51, };
short PFfs_rule_descriptor_124[] = {  -50,   84,  -50,  -50, };
short PFfs_rule_descriptor_125[] = {  -51,  -52, };
short PFfs_rule_descriptor_126[] = {  -51,   80,  -51,  -52, };
short PFfs_rule_descriptor_127[] = {  -51,   60,  -51,  -52, };
short PFfs_rule_descriptor_128[] = {  -52,  -53, };
short PFfs_rule_descriptor_129[] = {  -52,   64,  -53,  -52, };
short PFfs_rule_descriptor_130[] = {  -52,   20,  -53,  -52, };
short PFfs_rule_descriptor_131[] = {  -52,   42,  -53,  -52, };
short PFfs_rule_descriptor_132[] = {  -52,   61,  -53,  -52, };
short PFfs_rule_descriptor_133[] = {  -53,  -54, };
short PFfs_rule_descriptor_134[] = {  -53,  101,  -54,  -53, };
short PFfs_rule_descriptor_135[] = {  -54,  -55, };
short PFfs_rule_descriptor_136[] = {  -54,   46,  -55,  -54, };
short PFfs_rule_descriptor_137[] = {  -54,   29,  -55,  -54, };
short PFfs_rule_descriptor_138[] = {  -55,  -56, };
short PFfs_rule_descriptor_139[] = {  -55,   45,  -56,  -39, };
short PFfs_rule_descriptor_140[] = {  -56,  -57, };
short PFfs_rule_descriptor_141[] = {  -56,   98,  -57,  -39, };
short PFfs_rule_descriptor_142[] = {  -57,  -58, };
short PFfs_rule_descriptor_143[] = {  -57,   12,  -58,  -40, };
short PFfs_rule_descriptor_144[] = {  -58,  -59, };
short PFfs_rule_descriptor_145[] = {  -58,   11,  -59,  -40, };
short PFfs_rule_descriptor_146[] = {  -59,  -60, };
short PFfs_rule_descriptor_147[] = {  -59,  100,  -60, };
short PFfs_rule_descriptor_148[] = {  -59,  103,  -60, };
short PFfs_rule_descriptor_149[] = {  -60,  -61, };
short PFfs_rule_descriptor_150[] = {  -60,  -62, };
short PFfs_rule_descriptor_151[] = {  -60,  -89, };
short PFfs_rule_descriptor_152[] = {  -61,  110,  -27, };
short PFfs_rule_descriptor_153[] = {  -62,  -64, };
short PFfs_rule_descriptor_154[] = {  -62,  -63, };
short PFfs_rule_descriptor_155[] = {  -63,   87, };
short PFfs_rule_descriptor_156[] = {  -63,   56,  -62,  -62, };
short PFfs_rule_descriptor_157[] = {  -63,   56,   94,  -43,  -62, };
short PFfs_rule_descriptor_158[] = {  -63,   83,  -63,  -27, };
short PFfs_rule_descriptor_160[] = {  -64,  -65, };
short PFfs_rule_descriptor_161[] = {  -64,   83,  -62,  -27, };
short PFfs_rule_descriptor_163[] = {  -65,  -66, };
short PFfs_rule_descriptor_164[] = {  -65,  -72, };
short PFfs_rule_descriptor_165[] = {  -65,  -73, };
short PFfs_rule_descriptor_166[] = {  -65,  -74, };
short PFfs_rule_descriptor_167[] = {  -65,  -75, };
short PFfs_rule_descriptor_168[] = {  -65,  -79, };
short PFfs_rule_descriptor_169[] = {  -65,  -77, };
short PFfs_rule_descriptor_170[] = {  -65,  -78, };
short PFfs_rule_descriptor_171[] = {  -65, -104, };
short PFfs_rule_descriptor_172[] = {  -66,  -67, };
short PFfs_rule_descriptor_173[] = {  -66,  -71, };
short PFfs_rule_descriptor_174[] = {  -67,  -68, };
short PFfs_rule_descriptor_175[] = {  -67,  -69, };
short PFfs_rule_descriptor_176[] = {  -67,  -70, };
short PFfs_rule_descriptor_177[] = {  -68,   54, };
short PFfs_rule_descriptor_178[] = {  -69,   53, };
short PFfs_rule_descriptor_179[] = {  -70,   52, };
short PFfs_rule_descriptor_180[] = {  -71,   55, };
short PFfs_rule_descriptor_181[] = {  -72,  111, };
short PFfs_rule_descriptor_182[] = {  -73,  -27, };
short PFfs_rule_descriptor_183[] = {  -74,   22, };
short PFfs_rule_descriptor_184[] = {  -75,    2,   67, };
short PFfs_rule_descriptor_185[] = {  -75,    2,  -76, };
short PFfs_rule_descriptor_186[] = {  -76,   67, };
short PFfs_rule_descriptor_187[] = {  -76,    3,  -27,  -76, };
short PFfs_rule_descriptor_188[] = {  -77,   74,  -27, };
short PFfs_rule_descriptor_189[] = {  -78,  102,  -27, };
short PFfs_rule_descriptor_190[] = {  -79,  -80, };
short PFfs_rule_descriptor_191[] = {  -79,  -81, };
short PFfs_rule_descriptor_192[] = {  -79,  -82, };
short PFfs_rule_descriptor_193[] = {  -79,  -83, };
short PFfs_rule_descriptor_194[] = {  -79,  -84, };
short PFfs_rule_descriptor_195[] = {  -79,  -85, };
short PFfs_rule_descriptor_196[] = {  -80,   23,  -86,  -87, };
short PFfs_rule_descriptor_197[] = {  -81,    5,  -86,  -88, };
short PFfs_rule_descriptor_198[] = {  -82,   96,  -27, };
short PFfs_rule_descriptor_199[] = {  -83,   14,  -27, };
short PFfs_rule_descriptor_200[] = {  -84,   79,  -27,  -87, };
short PFfs_rule_descriptor_201[] = {  -85,   21,  -27, };
short PFfs_rule_descriptor_202[] = {  -86,   95, };
short PFfs_rule_descriptor_203[] = {  -86,  -27, };
short PFfs_rule_descriptor_204[] = {  -87,  -28, };
short PFfs_rule_descriptor_205[] = {  -87,   24, };
short PFfs_rule_descriptor_206[] = {  -87,   16,  -87,  -87, };
short PFfs_rule_descriptor_207[] = {  -88,  -28, };
short PFfs_rule_descriptor_208[] = {  -88,   24, };
short PFfs_rule_descriptor_209[] = {  -88,   16,  -88,  -88, };
short PFfs_rule_descriptor_210[] = {  -89,   32,  -90,   67, };
short PFfs_rule_descriptor_211[] = {  -89,   32,  -90,  -28, };
short PFfs_rule_descriptor_212[] = {  -90,   67, };
short PFfs_rule_descriptor_213[] = {  -90,   82,   81,  -90, };
short PFfs_rule_descriptor_214[] = {   -7,  -91, };
short PFfs_rule_descriptor_215[] = {  -91,  118, };
short PFfs_rule_descriptor_216[] = {  -29,  -92, };
short PFfs_rule_descriptor_217[] = {  -29,  -93, };
short PFfs_rule_descriptor_218[] = {  -29,  -95, };
short PFfs_rule_descriptor_219[] = {  -29,  -94, };
short PFfs_rule_descriptor_220[] = {  -29,  -96, };
short PFfs_rule_descriptor_221[] = {  -92,  119,  -99, -100, };
short PFfs_rule_descriptor_222[] = {  -93,  120,  -99, };
short PFfs_rule_descriptor_223[] = {  -94,  121, -100,  -29, };
short PFfs_rule_descriptor_224[] = {  -95,  122,  -99, -101, };
short PFfs_rule_descriptor_225[] = {  -96,  123,  -97,  124,  -29,  -29, };
short PFfs_rule_descriptor_226[] = {  -97,   67, };
short PFfs_rule_descriptor_227[] = {  -97,  125,  -98,  -97, };
short PFfs_rule_descriptor_228[] = {  -98,   50,  113,  -72,   67,  -27, };
short PFfs_rule_descriptor_229[] = {  -99,  -29, };
short PFfs_rule_descriptor_230[] = { -100,  -29, };
short PFfs_rule_descriptor_231[] = { -101,  -29, };
short PFfs_rule_descriptor_232[] = {  -39,   92,  126,   67, };
short PFfs_rule_descriptor_233[] = {  -29, -102, };
short PFfs_rule_descriptor_234[] = { -102,  127,  113,  -72, -103,  128,  -29,  -29, };
short PFfs_rule_descriptor_235[] = { -103,   67, };
short PFfs_rule_descriptor_236[] = { -103,  -39, };
short PFfs_rule_descriptor_237[] = { -104,  129,  -28,  -75, };
short PFfs_rule_descriptor_238[] = {  -39,   92,  130,   67, };
/* PFfs_rule_descriptors[0][1] = 1 iff grammar is normal form. */
short * PFfs_rule_descriptors[] = {
	PFfs_rule_descriptor_0,
	PFfs_rule_descriptor_1,
	PFfs_rule_descriptor_2,
	PFfs_rule_descriptor_3,
	PFfs_rule_descriptor_4,
	PFfs_rule_descriptor_5,
	PFfs_rule_descriptor_6,
	PFfs_rule_descriptor_7,
	PFfs_rule_descriptor_8,
	PFfs_rule_descriptor_9,
	PFfs_rule_descriptor_10,
	PFfs_rule_descriptor_11,
	PFfs_rule_descriptor_12,
	PFfs_rule_descriptor_13,
	PFfs_rule_descriptor_14,
	PFfs_rule_descriptor_15,
	PFfs_rule_descriptor_16,
	PFfs_rule_descriptor_17,
	PFfs_rule_descriptor_18,
	PFfs_rule_descriptor_19,
	PFfs_rule_descriptor_20,
	PFfs_rule_descriptor_21,
	PFfs_rule_descriptor_22,
	PFfs_rule_descriptor_23,
	PFfs_rule_descriptor_24,
	PFfs_rule_descriptor_25,
	PFfs_rule_descriptor_26,
	PFfs_rule_descriptor_27,
	PFfs_rule_descriptor_28,
	PFfs_rule_descriptor_29,
	PFfs_rule_descriptor_30,
	PFfs_rule_descriptor_31,
	PFfs_rule_descriptor_32,
	PFfs_rule_descriptor_33,
	PFfs_rule_descriptor_34,
	PFfs_rule_descriptor_35,
	PFfs_rule_descriptor_36,
	PFfs_rule_descriptor_37,
	PFfs_rule_descriptor_38,
	PFfs_rule_descriptor_39,
	PFfs_rule_descriptor_40,
	PFfs_rule_descriptor_41,
	PFfs_rule_descriptor_42,
	PFfs_rule_descriptor_43,
	PFfs_rule_descriptor_44,
	PFfs_rule_descriptor_45,
	PFfs_rule_descriptor_46,
	PFfs_rule_descriptor_47,
	PFfs_rule_descriptor_48,
	PFfs_rule_descriptor_49,
	PFfs_rule_descriptor_50,
	PFfs_rule_descriptor_51,
	PFfs_rule_descriptor_52,
	PFfs_rule_descriptor_53,
	PFfs_rule_descriptor_54,
	PFfs_rule_descriptor_55,
	PFfs_rule_descriptor_56,
	PFfs_rule_descriptor_57,
	PFfs_rule_descriptor_58,
	PFfs_rule_descriptor_59,
	PFfs_rule_descriptor_60,
	PFfs_rule_descriptor_61,
	PFfs_rule_descriptor_62,
	PFfs_rule_descriptor_63,
	PFfs_rule_descriptor_64,
	PFfs_rule_descriptor_65,
	PFfs_rule_descriptor_66,
	PFfs_rule_descriptor_0,
	PFfs_rule_descriptor_68,
	PFfs_rule_descriptor_69,
	PFfs_rule_descriptor_70,
	PFfs_rule_descriptor_71,
	PFfs_rule_descriptor_72,
	PFfs_rule_descriptor_73,
	PFfs_rule_descriptor_74,
	PFfs_rule_descriptor_75,
	PFfs_rule_descriptor_76,
	PFfs_rule_descriptor_77,
	PFfs_rule_descriptor_78,
	PFfs_rule_descriptor_79,
	PFfs_rule_descriptor_80,
	PFfs_rule_descriptor_81,
	PFfs_rule_descriptor_82,
	PFfs_rule_descriptor_83,
	PFfs_rule_descriptor_84,
	PFfs_rule_descriptor_85,
	PFfs_rule_descriptor_86,
	PFfs_rule_descriptor_87,
	PFfs_rule_descriptor_88,
	PFfs_rule_descriptor_89,
	PFfs_rule_descriptor_90,
	PFfs_rule_descriptor_91,
	PFfs_rule_descriptor_92,
	PFfs_rule_descriptor_93,
	PFfs_rule_descriptor_94,
	PFfs_rule_descriptor_95,
	PFfs_rule_descriptor_96,
	PFfs_rule_descriptor_97,
	PFfs_rule_descriptor_98,
	PFfs_rule_descriptor_99,
	PFfs_rule_descriptor_100,
	PFfs_rule_descriptor_101,
	PFfs_rule_descriptor_102,
	PFfs_rule_descriptor_103,
	PFfs_rule_descriptor_104,
	PFfs_rule_descriptor_105,
	PFfs_rule_descriptor_106,
	PFfs_rule_descriptor_107,
	PFfs_rule_descriptor_108,
	PFfs_rule_descriptor_109,
	PFfs_rule_descriptor_110,
	PFfs_rule_descriptor_111,
	PFfs_rule_descriptor_112,
	PFfs_rule_descriptor_113,
	PFfs_rule_descriptor_114,
	PFfs_rule_descriptor_115,
	PFfs_rule_descriptor_116,
	PFfs_rule_descriptor_117,
	PFfs_rule_descriptor_118,
	PFfs_rule_descriptor_119,
	PFfs_rule_descriptor_0,
	PFfs_rule_descriptor_121,
	PFfs_rule_descriptor_122,
	PFfs_rule_descriptor_123,
	PFfs_rule_descriptor_124,
	PFfs_rule_descriptor_125,
	PFfs_rule_descriptor_126,
	PFfs_rule_descriptor_127,
	PFfs_rule_descriptor_128,
	PFfs_rule_descriptor_129,
	PFfs_rule_descriptor_130,
	PFfs_rule_descriptor_131,
	PFfs_rule_descriptor_132,
	PFfs_rule_descriptor_133,
	PFfs_rule_descriptor_134,
	PFfs_rule_descriptor_135,
	PFfs_rule_descriptor_136,
	PFfs_rule_descriptor_137,
	PFfs_rule_descriptor_138,
	PFfs_rule_descriptor_139,
	PFfs_rule_descriptor_140,
	PFfs_rule_descriptor_141,
	PFfs_rule_descriptor_142,
	PFfs_rule_descriptor_143,
	PFfs_rule_descriptor_144,
	PFfs_rule_descriptor_145,
	PFfs_rule_descriptor_146,
	PFfs_rule_descriptor_147,
	PFfs_rule_descriptor_148,
	PFfs_rule_descriptor_149,
	PFfs_rule_descriptor_150,
	PFfs_rule_descriptor_151,
	PFfs_rule_descriptor_152,
	PFfs_rule_descriptor_153,
	PFfs_rule_descriptor_154,
	PFfs_rule_descriptor_155,
	PFfs_rule_descriptor_156,
	PFfs_rule_descriptor_157,
	PFfs_rule_descriptor_158,
	PFfs_rule_descriptor_0,
	PFfs_rule_descriptor_160,
	PFfs_rule_descriptor_161,
	PFfs_rule_descriptor_0,
	PFfs_rule_descriptor_163,
	PFfs_rule_descriptor_164,
	PFfs_rule_descriptor_165,
	PFfs_rule_descriptor_166,
	PFfs_rule_descriptor_167,
	PFfs_rule_descriptor_168,
	PFfs_rule_descriptor_169,
	PFfs_rule_descriptor_170,
	PFfs_rule_descriptor_171,
	PFfs_rule_descriptor_172,
	PFfs_rule_descriptor_173,
	PFfs_rule_descriptor_174,
	PFfs_rule_descriptor_175,
	PFfs_rule_descriptor_176,
	PFfs_rule_descriptor_177,
	PFfs_rule_descriptor_178,
	PFfs_rule_descriptor_179,
	PFfs_rule_descriptor_180,
	PFfs_rule_descriptor_181,
	PFfs_rule_descriptor_182,
	PFfs_rule_descriptor_183,
	PFfs_rule_descriptor_184,
	PFfs_rule_descriptor_185,
	PFfs_rule_descriptor_186,
	PFfs_rule_descriptor_187,
	PFfs_rule_descriptor_188,
	PFfs_rule_descriptor_189,
	PFfs_rule_descriptor_190,
	PFfs_rule_descriptor_191,
	PFfs_rule_descriptor_192,
	PFfs_rule_descriptor_193,
	PFfs_rule_descriptor_194,
	PFfs_rule_descriptor_195,
	PFfs_rule_descriptor_196,
	PFfs_rule_descriptor_197,
	PFfs_rule_descriptor_198,
	PFfs_rule_descriptor_199,
	PFfs_rule_descriptor_200,
	PFfs_rule_descriptor_201,
	PFfs_rule_descriptor_202,
	PFfs_rule_descriptor_203,
	PFfs_rule_descriptor_204,
	PFfs_rule_descriptor_205,
	PFfs_rule_descriptor_206,
	PFfs_rule_descriptor_207,
	PFfs_rule_descriptor_208,
	PFfs_rule_descriptor_209,
	PFfs_rule_descriptor_210,
	PFfs_rule_descriptor_211,
	PFfs_rule_descriptor_212,
	PFfs_rule_descriptor_213,
	PFfs_rule_descriptor_214,
	PFfs_rule_descriptor_215,
	PFfs_rule_descriptor_216,
	PFfs_rule_descriptor_217,
	PFfs_rule_descriptor_218,
	PFfs_rule_descriptor_219,
	PFfs_rule_descriptor_220,
	PFfs_rule_descriptor_221,
	PFfs_rule_descriptor_222,
	PFfs_rule_descriptor_223,
	PFfs_rule_descriptor_224,
	PFfs_rule_descriptor_225,
	PFfs_rule_descriptor_226,
	PFfs_rule_descriptor_227,
	PFfs_rule_descriptor_228,
	PFfs_rule_descriptor_229,
	PFfs_rule_descriptor_230,
	PFfs_rule_descriptor_231,
	PFfs_rule_descriptor_232,
	PFfs_rule_descriptor_233,
	PFfs_rule_descriptor_234,
	PFfs_rule_descriptor_235,
	PFfs_rule_descriptor_236,
	PFfs_rule_descriptor_237,
	PFfs_rule_descriptor_238,
};
short PFfs_cost[][4] = {
	{    0,    0,    0,    0}, /* (none) = 0 */
	{   10,    0,    0,    0}, /* Query: MainModule = 1 */
	{   10,    0,    0,    0}, /* Query: lib_mod(ModuleDecl, Prolog) = 2 */
	{   10,    0,    0,    0}, /* MainModule: main_mod(Prolog, QueryBody) = 3 */
	{   10,    0,    0,    0}, /* LibraryModule: lib_mod(ModuleDecl, Prolog) = 4 */
	{   10,    0,    0,    0}, /* ModuleDecl: mod_ns(lit_str) = 5 */
	{   10,    0,    0,    0}, /* Prolog: DeclsImports = 6 */
	{   10,    0,    0,    0}, /* DeclsImports: decl_imps(DeclImport, DeclsImports) = 7 */
	{   10,    0,    0,    0}, /* DeclsImports: nil = 8 */
	{   10,    0,    0,    0}, /* DeclsImports: decl_imps(fun(fun_sig(OptParamList_, SequenceType), Expr), DeclsImports) = 9 */
	{   10,    0,    0,    0}, /* DeclsImports: decl_imps(fun(fun_sig(OptParamList_, SequenceType), external), DeclsImports) = 10 */
	{   10,    0,    0,    0}, /* DeclImport: DefaultNamespaceDecl = 11 */
	{   10,    0,    0,    0}, /* DeclImport: BoundarySpaceDecl = 12 */
	{   10,    0,    0,    0}, /* DeclImport: DefaultCollationDecl = 13 */
	{   10,    0,    0,    0}, /* DeclImport: BaseURIDecl = 14 */
	{   10,    0,    0,    0}, /* DeclImport: ConstructionDecl = 15 */
	{   10,    0,    0,    0}, /* DeclImport: OrderingModeDecl = 16 */
	{   10,    0,    0,    0}, /* DeclImport: EmptyOrderingDecl = 17 */
	{   10,    0,    0,    0}, /* DeclImport: CopyNamespacesDecl = 18 */
	{   10,    0,    0,    0}, /* DeclImport: NamespaceDecl = 19 */
	{   10,    0,    0,    0}, /* DeclImport: OptionDecl = 20 */
	{   10,    0,    0,    0}, /* DeclImport: SchemaImport = 21 */
	{   10,    0,    0,    0}, /* DeclImport: ModuleImport = 22 */
	{   10,    0,    0,    0}, /* DeclImport: VariableDeclaration = 23 */
	{   10,    0,    0,    0}, /* DeclImport: LibraryModule = 24 */
	{   10,    0,    0,    0}, /* DefaultNamespaceDecl: ens_decl(lit_str) = 25 */
	{   10,    0,    0,    0}, /* DefaultNamespaceDecl: fns_decl(lit_str) = 26 */
	{   10,    0,    0,    0}, /* BoundarySpaceDecl: boundspc_decl = 27 */
	{   10,    0,    0,    0}, /* DefaultCollationDecl: coll_decl(lit_str) = 28 */
	{   10,    0,    0,    0}, /* BaseURIDecl: base_uri(lit_str) = 29 */
	{   10,    0,    0,    0}, /* ConstructionDecl: constr_decl = 30 */
	{   10,    0,    0,    0}, /* OptionDecl: option(lit_str) = 31 */
	{   10,    0,    0,    0}, /* OrderingModeDecl: ordering_mode = 32 */
	{   10,    0,    0,    0}, /* EmptyOrderingDecl: def_order = 33 */
	{   10,    0,    0,    0}, /* CopyNamespacesDecl: copy_ns = 34 */
	{   10,    0,    0,    0}, /* NamespaceDecl: ns_decl(lit_str) = 35 */
	{   10,    0,    0,    0}, /* SchemaImport: schm_imp(lit_str, OptImportLoc_) = 36 */
	{   10,    0,    0,    0}, /* ModuleImport: mod_imp(lit_str, OptImportLoc_) = 37 */
	{   10,    0,    0,    0}, /* OptImportLoc_: nil = 38 */
	{   10,    0,    0,    0}, /* OptImportLoc_: schm_ats(lit_str, OptImportLoc_) = 39 */
	{   10,    0,    0,    0}, /* OptParamList_: nil = 40 */
	{   10,    0,    0,    0}, /* OptParamList_: ParamList_ = 41 */
	{   10,    0,    0,    0}, /* ParamList_: nil = 42 */
	{   10,    0,    0,    0}, /* ParamList_: params(Param, ParamList_) = 43 */
	{   10,    0,    0,    0}, /* Param: param(SequenceType, Var) = 44 */
	{   10,    0,    0,    0}, /* OptSequenceType_: nil = 45 */
	{   10,    0,    0,    0}, /* OptSequenceType_: SequenceType = 46 */
	{   10,    0,    0,    0}, /* VariableDeclaration: var_decl(var_type(Var, nil), OptExpr) = 47 */
	{   10,    0,    0,    0}, /* VariableDeclaration: var_decl(var_type(Var, SequenceType), OptExpr) = 48 */
	{   10,    0,    0,    0}, /* VariableDeclaration: var_decl(var_type(Var, OptSequenceType_), external) = 49 */
	{   10,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{   10,    0,    0,    0}, /* OptExpr: empty_seq = 51 */
	{   10,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{   10,    0,    0,    0}, /* Expr: exprseq(OptExpr, OptExpr) = 53 */
	{   10,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{   10,    0,    0,    0}, /* ExprSingle: FLWORExpr = 55 */
	{   10,    0,    0,    0}, /* ExprSingle: QuantifiedExpr = 56 */
	{   10,    0,    0,    0}, /* ExprSingle: TypeswitchExpr = 57 */
	{   10,    0,    0,    0}, /* ExprSingle: IfExpr = 58 */
	{   10,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{   10,    0,    0,    0}, /* FLWORExpr: flwr(OptBindExpr, where(OptWhereClause_, ord_ret(nil, OptExpr))) = 60 */
	{   10,    0,    0,    0}, /* FLWORExpr: flwr(OptBindExpr, where(OptWhereClause_, ord_ret(orderby(OrderSpecList), OptExpr))) = 61 */
	{   10,    0,    0,    0}, /* OptBindExpr: nil = 62 */
	{   10,    0,    0,    0}, /* OptBindExpr: binds(let(var_type(Var, nil), OptExpr), OptBindExpr) = 63 */
	{   10,    0,    0,    0}, /* OptBindExpr: binds(let(var_type(Var, TypeDeclaration), OptExpr), OptBindExpr) = 64 */
	{   10,    0,    0,    0}, /* OptBindExpr: binds(bind(vars(var_type(Var, nil), OptPosVar), OptExpr), OptBindExpr) = 65 */
	{   10,    0,    0,    0}, /* OptBindExpr: binds(bind(vars(var_type(Var, TypeDeclaration), OptPosVar), OptExpr), OptBindExpr) = 66 */
	{    0,    0,    0,    0}, /* (none) = 67 */
	{   10,    0,    0,    0}, /* OptPosVar: nil = 68 */
	{   10,    0,    0,    0}, /* OptPosVar: Var = 69 */
	{   10,    0,    0,    0}, /* TypeDeclaration: SequenceType = 70 */
	{   10,    0,    0,    0}, /* OptWhereClause_: nil = 71 */
	{   10,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{   10,    0,    0,    0}, /* OrderSpecList: nil = 73 */
	{   10,    0,    0,    0}, /* OrderSpecList: orderspecs(OptExpr, OrderSpecList) = 74 */
	{   10,    0,    0,    0}, /* QuantifiedExpr: some(binds(bind(vars(var_type(Var, nil), nil), OptExpr), nil), OptExpr) = 75 */
	{   10,    0,    0,    0}, /* QuantifiedExpr: some(binds(bind(vars(var_type(Var, TypeDeclaration), nil), OptExpr), nil), OptExpr) = 76 */
	{   10,    0,    0,    0}, /* QuantifiedExpr: every(binds(bind(vars(var_type(Var, nil), nil), OptExpr), nil), OptExpr) = 77 */
	{   10,    0,    0,    0}, /* QuantifiedExpr: every(binds(bind(vars(var_type(Var, TypeDeclaration), nil), OptExpr), nil), OptExpr) = 78 */
	{   10,    0,    0,    0}, /* TypeswitchExpr: typeswitch(OptExpr, cases(case_(var_type(OptCaseVar, SequenceType), OptExpr), cases(default_(OptCaseVar, OptExpr), nil))) = 79 */
	{   10,    0,    0,    0}, /* OptCaseVar: nil = 80 */
	{   10,    0,    0,    0}, /* OptCaseVar: Var = 81 */
	{   10,    0,    0,    0}, /* SequenceType: seq_ty(ItemType) = 82 */
	{   10,    0,    0,    0}, /* SingleType: seq_ty(AtomType) = 83 */
	{   10,    0,    0,    0}, /* ItemType: empty_ty = 84 */
	{   10,    0,    0,    0}, /* ItemType: NodeTest = 85 */
	{   10,    0,    0,    0}, /* ItemType: item_ty(nil) = 86 */
	{   10,    0,    0,    0}, /* ItemType: atom_ty(nil) = 87 */
	{   10,    0,    0,    0}, /* ItemType: schm_elem = 88 */
	{   10,    0,    0,    0}, /* ItemType: schm_attr = 89 */
	{   10,    0,    0,    0}, /* AtomType: atom_ty(nil) = 90 */
	{   10,    0,    0,    0}, /* NodeTest: node_ty(nil) = 91 */
	{   10,    0,    0,    0}, /* NodeTest: node_ty(ElemOrAttrType) = 92 */
	{   10,    0,    0,    0}, /* NodeTest: node_ty(node_ty(nil)) = 93 */
	{   10,    0,    0,    0}, /* NodeTest: node_ty(node_ty(ElemOrAttrType)) = 94 */
	{   10,    0,    0,    0}, /* NodeTest: node_ty(lit_str) = 95 */
	{   10,    0,    0,    0}, /* ElemOrAttrType: req_ty(nil, nil) = 96 */
	{   10,    0,    0,    0}, /* ElemOrAttrType: req_ty(nil, SchemaType) = 97 */
	{   10,    0,    0,    0}, /* ElemOrAttrType: req_ty(req_name, nil) = 98 */
	{   10,    0,    0,    0}, /* ElemOrAttrType: req_ty(req_name, SchemaType) = 99 */
	{   10,    0,    0,    0}, /* SchemaType: named_ty = 100 */
	{   10,    0,    0,    0}, /* IfExpr: if_(OptExpr, then_else(OptExpr, OptExpr)) = 101 */
	{   10,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{   10,    0,    0,    0}, /* OrExpr: or(AndExpr, OrExpr) = 103 */
	{   10,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{   10,    0,    0,    0}, /* AndExpr: and(ComparisonExpr, AndExpr) = 105 */
	{   10,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{   10,    0,    0,    0}, /* ComparisonExpr: eq(RangeExpr, RangeExpr) = 107 */
	{   10,    0,    0,    0}, /* ComparisonExpr: ne(RangeExpr, RangeExpr) = 108 */
	{   10,    0,    0,    0}, /* ComparisonExpr: lt(RangeExpr, RangeExpr) = 109 */
	{   10,    0,    0,    0}, /* ComparisonExpr: le(RangeExpr, RangeExpr) = 110 */
	{   10,    0,    0,    0}, /* ComparisonExpr: gt(RangeExpr, RangeExpr) = 111 */
	{   10,    0,    0,    0}, /* ComparisonExpr: ge(RangeExpr, RangeExpr) = 112 */
	{   10,    0,    0,    0}, /* ComparisonExpr: val_eq(RangeExpr, RangeExpr) = 113 */
	{   10,    0,    0,    0}, /* ComparisonExpr: val_ne(RangeExpr, RangeExpr) = 114 */
	{   10,    0,    0,    0}, /* ComparisonExpr: val_lt(RangeExpr, RangeExpr) = 115 */
	{   10,    0,    0,    0}, /* ComparisonExpr: val_le(RangeExpr, RangeExpr) = 116 */
	{   10,    0,    0,    0}, /* ComparisonExpr: val_gt(RangeExpr, RangeExpr) = 117 */
	{   10,    0,    0,    0}, /* ComparisonExpr: val_ge(RangeExpr, RangeExpr) = 118 */
	{   10,    0,    0,    0}, /* ComparisonExpr: is(RangeExpr, RangeExpr) = 119 */
	{    0,    0,    0,    0}, /* (none) = 120 */
	{   10,    0,    0,    0}, /* ComparisonExpr: ltlt(RangeExpr, RangeExpr) = 121 */
	{   10,    0,    0,    0}, /* ComparisonExpr: gtgt(RangeExpr, RangeExpr) = 122 */
	{   10,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{   10,    0,    0,    0}, /* RangeExpr: range(RangeExpr, RangeExpr) = 124 */
	{   10,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{   10,    0,    0,    0}, /* AdditiveExpr: plus(AdditiveExpr, MultiplicativeExpr) = 126 */
	{   10,    0,    0,    0}, /* AdditiveExpr: minus(AdditiveExpr, MultiplicativeExpr) = 127 */
	{   10,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{   10,    0,    0,    0}, /* MultiplicativeExpr: mult(UnionExpr, MultiplicativeExpr) = 129 */
	{   10,    0,    0,    0}, /* MultiplicativeExpr: div(UnionExpr, MultiplicativeExpr) = 130 */
	{   10,    0,    0,    0}, /* MultiplicativeExpr: idiv(UnionExpr, MultiplicativeExpr) = 131 */
	{   10,    0,    0,    0}, /* MultiplicativeExpr: mod(UnionExpr, MultiplicativeExpr) = 132 */
	{   10,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{   10,    0,    0,    0}, /* UnionExpr: union_(IntersectExceptExpr, UnionExpr) = 134 */
	{   10,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{   10,    0,    0,    0}, /* IntersectExceptExpr: intersect(InstanceofExpr, IntersectExceptExpr) = 136 */
	{   10,    0,    0,    0}, /* IntersectExceptExpr: except(InstanceofExpr, IntersectExceptExpr) = 137 */
	{   10,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{   10,    0,    0,    0}, /* InstanceofExpr: instof(TreatExpr, SequenceType) = 139 */
	{   10,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{   10,    0,    0,    0}, /* TreatExpr: treat(CastableExpr, SequenceType) = 141 */
	{   10,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{   10,    0,    0,    0}, /* CastableExpr: castable(CastExpr, SingleType) = 143 */
	{   10,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{   10,    0,    0,    0}, /* CastExpr: cast(UnaryExpr, SingleType) = 145 */
	{   10,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{   10,    0,    0,    0}, /* UnaryExpr: uminus(ValueExpr) = 147 */
	{   10,    0,    0,    0}, /* UnaryExpr: uplus(ValueExpr) = 148 */
	{   10,    0,    0,    0}, /* ValueExpr: ValidateExpr = 149 */
	{   10,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{   10,    0,    0,    0}, /* ValueExpr: ExtensionExpr = 151 */
	{   10,    0,    0,    0}, /* ValidateExpr: validate(OptExpr) = 152 */
	{   10,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{   10,    0,    0,    0}, /* PathExpr: LocationPath = 154 */
	{   10,    0,    0,    0}, /* LocationPath: root_ = 155 */
	{   10,    0,    0,    0}, /* LocationPath: locpath(PathExpr, PathExpr) = 156 */
	{   10,    0,    0,    0}, /* LocationPath: locpath(step(NodeTest), PathExpr) = 157 */
	{   10,    0,    0,    0}, /* LocationPath: pred(LocationPath, OptExpr) = 158 */
	{    0,    0,    0,    0}, /* (none) = 159 */
	{   10,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   10,    0,    0,    0}, /* FilterExpr: pred(PathExpr, OptExpr) = 161 */
	{    0,    0,    0,    0}, /* (none) = 162 */
	{   10,    0,    0,    0}, /* PrimaryExpr: Literal = 163 */
	{   10,    0,    0,    0}, /* PrimaryExpr: Var = 164 */
	{   10,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{   10,    0,    0,    0}, /* PrimaryExpr: ContextItemExpr = 166 */
	{   10,    0,    0,    0}, /* PrimaryExpr: FunctionCall = 167 */
	{   10,    0,    0,    0}, /* PrimaryExpr: Constructor = 168 */
	{   10,    0,    0,    0}, /* PrimaryExpr: OrderedExpr = 169 */
	{   10,    0,    0,    0}, /* PrimaryExpr: UnorderedExpr = 170 */
	{   10,    0,    0,    0}, /* PrimaryExpr: XRPCCall = 171 */
	{   10,    0,    0,    0}, /* Literal: NumericLiteral = 172 */
	{   10,    0,    0,    0}, /* Literal: StringLiteral = 173 */
	{   10,    0,    0,    0}, /* NumericLiteral: IntegerLiteral = 174 */
	{   10,    0,    0,    0}, /* NumericLiteral: DecimalLiteral = 175 */
	{   10,    0,    0,    0}, /* NumericLiteral: DoubleLiteral = 176 */
	{   10,    0,    0,    0}, /* IntegerLiteral: lit_int = 177 */
	{   10,    0,    0,    0}, /* DecimalLiteral: lit_dec = 178 */
	{   10,    0,    0,    0}, /* DoubleLiteral: lit_dbl = 179 */
	{   10,    0,    0,    0}, /* StringLiteral: lit_str = 180 */
	{   10,    0,    0,    0}, /* Var: var = 181 */
	{   10,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{   10,    0,    0,    0}, /* ContextItemExpr: dot = 183 */
	{   10,    0,    0,    0}, /* FunctionCall: apply(nil) = 184 */
	{   10,    0,    0,    0}, /* FunctionCall: apply(FuncArgList) = 185 */
	{   10,    0,    0,    0}, /* FuncArgList: nil = 186 */
	{   10,    0,    0,    0}, /* FuncArgList: args(OptExpr, FuncArgList) = 187 */
	{   10,    0,    0,    0}, /* OrderedExpr: ordered(OptExpr) = 188 */
	{   10,    0,    0,    0}, /* UnorderedExpr: unordered(OptExpr) = 189 */
	{   10,    0,    0,    0}, /* Constructor: ElementConstructor = 190 */
	{   10,    0,    0,    0}, /* Constructor: AttributeConstructor = 191 */
	{   10,    0,    0,    0}, /* Constructor: TextConstructor = 192 */
	{   10,    0,    0,    0}, /* Constructor: CommentConstructor = 193 */
	{   10,    0,    0,    0}, /* Constructor: PIConstructor = 194 */
	{   10,    0,    0,    0}, /* Constructor: DocumentConstructor = 195 */
	{   10,    0,    0,    0}, /* ElementConstructor: elem(TagName, OptContentExpr) = 196 */
	{   10,    0,    0,    0}, /* AttributeConstructor: attr(TagName, AttributeValue) = 197 */
	{   10,    0,    0,    0}, /* TextConstructor: text(OptExpr) = 198 */
	{   10,    0,    0,    0}, /* CommentConstructor: comment(OptExpr) = 199 */
	{   10,    0,    0,    0}, /* PIConstructor: pi(OptExpr, OptContentExpr) = 200 */
	{   10,    0,    0,    0}, /* DocumentConstructor: doc(OptExpr) = 201 */
	{   10,    0,    0,    0}, /* TagName: tag = 202 */
	{   10,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{   10,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{   10,    0,    0,    0}, /* OptContentExpr: empty_seq = 205 */
	{   10,    0,    0,    0}, /* OptContentExpr: contseq(OptContentExpr, OptContentExpr) = 206 */
	{   10,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{   10,    0,    0,    0}, /* AttributeValue: empty_seq = 208 */
	{   10,    0,    0,    0}, /* AttributeValue: contseq(AttributeValue, AttributeValue) = 209 */
	{   10,    0,    0,    0}, /* ExtensionExpr: ext_expr(Pragmas, nil) = 210 */
	{   10,    0,    0,    0}, /* ExtensionExpr: ext_expr(Pragmas, Expr) = 211 */
	{   10,    0,    0,    0}, /* Pragmas: nil = 212 */
	{   10,    0,    0,    0}, /* Pragmas: pragmas(pragma, Pragmas) = 213 */
	{   10,    0,    0,    0}, /* DeclImport: RevalidationDecl = 214 */
	{   10,    0,    0,    0}, /* RevalidationDecl: revalid = 215 */
	{   10,    0,    0,    0}, /* ExprSingle: InsertExpr = 216 */
	{   10,    0,    0,    0}, /* ExprSingle: DeleteExpr = 217 */
	{   10,    0,    0,    0}, /* ExprSingle: RenameExpr = 218 */
	{   10,    0,    0,    0}, /* ExprSingle: ReplaceExpr = 219 */
	{   10,    0,    0,    0}, /* ExprSingle: TransformExpr = 220 */
	{   10,    0,    0,    0}, /* InsertExpr: insert(SourceExpr, TargetExpr) = 221 */
	{   10,    0,    0,    0}, /* DeleteExpr: delete(SourceExpr) = 222 */
	{   10,    0,    0,    0}, /* ReplaceExpr: replace(TargetExpr, ExprSingle) = 223 */
	{   10,    0,    0,    0}, /* RenameExpr: rename(SourceExpr, NewNameExpr) = 224 */
	{   10,    0,    0,    0}, /* TransformExpr: transform(TransformBindings, modify(ExprSingle, ExprSingle)) = 225 */
	{   10,    0,    0,    0}, /* TransformBindings: nil = 226 */
	{   10,    0,    0,    0}, /* TransformBindings: transbinds(TransformBinding, TransformBindings) = 227 */
	{   10,    0,    0,    0}, /* TransformBinding: let(var_type(Var, nil), OptExpr) = 228 */
	{   10,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{   10,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{   10,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{   10,    0,    0,    0}, /* SequenceType: seq_ty(stmt_ty(nil)) = 232 */
	{   10,    0,    0,    0}, /* ExprSingle: RecursiveExpr = 233 */
	{   10,    0,    0,    0}, /* RecursiveExpr: recursion(var_type(Var, OptTypeDeclForRec), seed(ExprSingle, ExprSingle)) = 234 */
	{   10,    0,    0,    0}, /* OptTypeDeclForRec: nil = 235 */
	{   10,    0,    0,    0}, /* OptTypeDeclForRec: SequenceType = 236 */
	{   10,    0,    0,    0}, /* XRPCCall: xrpc(Expr, FunctionCall) = 237 */
	{   10,    0,    0,    0}, /* SequenceType: seq_ty(docmgmt_ty(nil)) = 238 */
};

short PFfs_delta_cost[167][105][4] = {
{{0}}, /* state 0 */
{ /* state #1: and(dot, dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   50,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{   40,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{   30,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{   20,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   50,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{    0,    0,    0,    0}, /* AndExpr: and(ComparisonExpr, AndExpr) = 105 */
	{  200,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{  190,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  180,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  170,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  160,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  150,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  140,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{  130,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{  120,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{  110,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{  100,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{   90,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   80,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   70,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   60,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   50,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   50,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{   40,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{   40,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   30,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{   30,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{   30,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #2: apply(args(dot, nil)) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  210,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{  200,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{  190,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{  180,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  210,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  170,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{  160,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{  150,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{  140,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  130,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  120,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  110,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  100,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{   90,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{   80,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{   70,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{   60,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{   50,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{   40,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   30,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   20,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   10,    0,    0,    0}, /* PrimaryExpr: FunctionCall = 167 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  210,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* FunctionCall: apply(FuncArgList) = 185 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  210,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{  200,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{  200,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  190,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{  190,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{  190,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #3: apply(nil) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  210,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{  200,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{  190,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{  180,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  210,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  170,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{  160,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{  150,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{  140,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  130,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  120,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  110,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  100,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{   90,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{   80,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{   70,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{   60,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{   50,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{   40,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   30,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   20,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   10,    0,    0,    0}, /* PrimaryExpr: FunctionCall = 167 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  210,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* FunctionCall: apply(nil) = 184 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  210,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{  200,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{  200,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  190,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{  190,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{  190,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #4: args(dot, nil) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* FuncArgList: args(OptExpr, FuncArgList) = 187 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #5: atom_ty(nil) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* ItemType: atom_ty(nil) = 87 */
	{    0,    0,    0,    0}, /* AtomType: atom_ty(nil) = 90 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #6: attr(tag, dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  220,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{  210,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{  200,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{  190,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  220,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  180,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{  170,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{  160,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{  150,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  140,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  130,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  120,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  110,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  100,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{   90,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{   80,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{   70,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{   60,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{   50,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   30,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   20,    0,    0,    0}, /* PrimaryExpr: Constructor = 168 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  220,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* Constructor: AttributeConstructor = 191 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* AttributeConstructor: attr(TagName, AttributeValue) = 197 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  220,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{  210,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{  210,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  200,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{  200,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{  200,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #7: base_uri(lit_str) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* DeclImport: BaseURIDecl = 14 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* BaseURIDecl: base_uri(lit_str) = 29 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #8: bind(vars(var_type(var, seq_ty(schm_attr)), nil), dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #9: bind(vars(var_type(var, seq_ty(schm_attr)), var), dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #10: bind(vars(var_type(var, nil), var), dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #11: bind(vars(var_type(var, nil), nil), dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #12: binds(let(var_type(var, seq_ty(schm_attr)), dot), nil) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* OptBindExpr: binds(let(var_type(Var, TypeDeclaration), OptExpr), OptBindExpr) = 64 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #13: binds(bind(vars(var_type(var, seq_ty(schm_attr)), nil), dot), nil) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   30,    0,    0,    0}, /* OptBindExpr: binds(bind(vars(var_type(Var, TypeDeclaration), OptPosVar), OptExpr), OptBindExpr) = 66 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #14: binds(bind(vars(var_type(var, seq_ty(schm_attr)), var), dot), nil) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* OptBindExpr: binds(bind(vars(var_type(Var, TypeDeclaration), OptPosVar), OptExpr), OptBindExpr) = 66 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #15: binds(bind(vars(var_type(var, nil), nil), dot), nil) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   30,    0,    0,    0}, /* OptBindExpr: binds(bind(vars(var_type(Var, nil), OptPosVar), OptExpr), OptBindExpr) = 65 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #16: binds(bind(vars(var_type(var, nil), var), dot), nil) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* OptBindExpr: binds(bind(vars(var_type(Var, nil), OptPosVar), OptExpr), OptBindExpr) = 65 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #17: binds(let(var_type(var, nil), dot), nil) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* OptBindExpr: binds(let(var_type(Var, nil), OptExpr), OptBindExpr) = 63 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #18: boundspc_decl */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* DeclImport: BoundarySpaceDecl = 12 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* BoundarySpaceDecl: boundspc_decl = 27 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #19: case_(var_type(var, seq_ty(schm_attr)), dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #20: cases(case_(var_type(var, seq_ty(schm_attr)), dot), cases(default_(var, dot), nil)) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #21: cases(default_(var, dot), nil) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #22: cast(dot, seq_ty(atom_ty(nil))) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  150,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{  140,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{  130,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{  120,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  150,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  110,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{  100,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{   90,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{   80,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{   70,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{   60,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{   50,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{   40,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{   30,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{   20,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{   10,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{    0,    0,    0,    0}, /* CastExpr: cast(UnaryExpr, SingleType) = 145 */
	{  200,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{  190,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{  180,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{  170,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{  160,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  150,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  150,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{  140,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{  140,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  130,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{  130,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{  130,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #23: castable(dot, seq_ty(atom_ty(nil))) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  140,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{  130,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{  120,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{  110,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  140,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  100,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{   90,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{   80,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{   70,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{   60,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{   50,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{   40,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{   30,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{   20,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{   10,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{    0,    0,    0,    0}, /* CastableExpr: castable(CastExpr, SingleType) = 143 */
	{  200,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{  190,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{  180,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{  170,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{  160,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{  150,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  140,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  140,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{  130,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{  130,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  120,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{  120,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{  120,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #24: coll_decl(lit_str) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* DeclImport: DefaultCollationDecl = 13 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* DefaultCollationDecl: coll_decl(lit_str) = 28 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #25: comment(dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  220,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{  210,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{  200,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{  190,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  220,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  180,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{  170,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{  160,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{  150,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  140,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  130,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  120,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  110,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  100,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{   90,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{   80,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{   70,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{   60,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{   50,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   30,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   20,    0,    0,    0}, /* PrimaryExpr: Constructor = 168 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  220,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* Constructor: CommentConstructor = 193 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* CommentConstructor: comment(OptExpr) = 199 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  220,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{  210,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{  210,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  200,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{  200,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{  200,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #26: constr_decl */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* DeclImport: ConstructionDecl = 15 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* ConstructionDecl: constr_decl = 30 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #27: contseq(dot, dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* OptContentExpr: contseq(OptContentExpr, OptContentExpr) = 206 */
	{    0,    0,    0,    0}, /* AttributeValue: contseq(AttributeValue, AttributeValue) = 209 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #28: copy_ns */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* DeclImport: CopyNamespacesDecl = 18 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* CopyNamespacesDecl: copy_ns = 34 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #29: decl_imps(revalid, nil) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* Prolog: DeclsImports = 6 */
	{    0,    0,    0,    0}, /* DeclsImports: decl_imps(DeclImport, DeclsImports) = 7 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #30: decl_imps(fun(fun_sig(nil, seq_ty(schm_attr)), external), nil) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* Prolog: DeclsImports = 6 */
	{    0,    0,    0,    0}, /* DeclsImports: decl_imps(fun(fun_sig(OptParamList_, SequenceType), external), DeclsImports) = 10 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #31: decl_imps(fun(fun_sig(nil, seq_ty(schm_attr)), dot), nil) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* Prolog: DeclsImports = 6 */
	{    0,    0,    0,    0}, /* DeclsImports: decl_imps(fun(fun_sig(OptParamList_, SequenceType), Expr), DeclsImports) = 9 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #32: def_order */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* DeclImport: EmptyOrderingDecl = 17 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* EmptyOrderingDecl: def_order = 33 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #33: default_(var, dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #34: delete(dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{   30,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{   20,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{   10,    0,    0,    0}, /* ExprSingle: DeleteExpr = 217 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  210,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{  200,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{  190,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{  180,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  170,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  160,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  150,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  140,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  130,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{  120,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{  110,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{  100,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{   90,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{   80,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   70,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   50,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{   30,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{   30,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* DeleteExpr: delete(SourceExpr) = 222 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   20,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{   20,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{   20,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #35: div(dot, dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   90,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{   80,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{   70,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{   60,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   90,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   50,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{   40,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{   30,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{   20,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{   10,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{    0,    0,    0,    0}, /* MultiplicativeExpr: div(UnionExpr, MultiplicativeExpr) = 130 */
	{  200,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  190,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  180,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{  170,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{  160,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{  150,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{  140,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{  130,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{  120,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{  110,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{  100,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   90,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   90,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{   80,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{   80,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   70,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{   70,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{   70,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #36: doc(dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  220,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{  210,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{  200,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{  190,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  220,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  180,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{  170,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{  160,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{  150,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  140,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  130,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  120,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  110,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  100,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{   90,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{   80,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{   70,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{   60,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{   50,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   30,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   20,    0,    0,    0}, /* PrimaryExpr: Constructor = 168 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  220,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* Constructor: DocumentConstructor = 195 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* DocumentConstructor: doc(OptExpr) = 201 */
	{  220,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{  210,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{  210,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  200,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{  200,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{  200,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #37: docmgmt_ty(nil) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #38: dot */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  210,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{  200,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{  190,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{  180,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  210,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  170,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{  160,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{  150,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{  140,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  130,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  120,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  110,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  100,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{   90,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{   80,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{   70,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{   60,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{   50,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{   40,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   30,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   20,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   10,    0,    0,    0}, /* PrimaryExpr: ContextItemExpr = 166 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  210,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* ContextItemExpr: dot = 183 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  210,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{  200,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{  200,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  190,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{  190,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{  190,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #39: elem(tag, dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  220,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{  210,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{  200,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{  190,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  220,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  180,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{  170,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{  160,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{  150,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  140,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  130,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  120,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  110,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  100,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{   90,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{   80,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{   70,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{   60,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{   50,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   30,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   20,    0,    0,    0}, /* PrimaryExpr: Constructor = 168 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  220,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* Constructor: ElementConstructor = 190 */
	{    0,    0,    0,    0}, /* ElementConstructor: elem(TagName, OptContentExpr) = 196 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  220,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{  210,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{  210,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  200,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{  200,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{  200,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #40: empty_seq */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{    0,    0,    0,    0}, /* OptExpr: empty_seq = 51 */
	{  200,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{  190,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  180,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{  170,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{  160,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{  150,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  140,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  130,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  120,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  110,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  100,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{   90,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{   80,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{   70,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{   60,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{   50,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   30,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   20,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{    0,    0,    0,    0}, /* OptContentExpr: empty_seq = 205 */
	{    0,    0,    0,    0}, /* AttributeValue: empty_seq = 208 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  200,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{  200,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{  200,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #41: empty_ty */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* ItemType: empty_ty = 84 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #42: ens_decl(lit_str) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* DeclImport: DefaultNamespaceDecl = 11 */
	{    0,    0,    0,    0}, /* DefaultNamespaceDecl: ens_decl(lit_str) = 25 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #43: eq(dot, dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{   50,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{   40,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{   30,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   20,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{   10,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{    0,    0,    0,    0}, /* ComparisonExpr: eq(RangeExpr, RangeExpr) = 107 */
	{  200,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  190,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  180,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  170,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  160,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  150,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{  140,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{  130,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{  120,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{  110,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{  100,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   90,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   80,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   70,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{   50,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{   50,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{   40,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{   40,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #44: every(binds(bind(vars(var_type(var, seq_ty(schm_attr)), nil), dot), nil), dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{   30,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{   20,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{   10,    0,    0,    0}, /* ExprSingle: QuantifiedExpr = 56 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* QuantifiedExpr: every(binds(bind(vars(var_type(Var, TypeDeclaration), nil), OptExpr), nil), OptExpr) = 78 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  210,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{  200,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{  190,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{  180,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  170,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  160,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  150,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  140,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  130,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{  120,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{  110,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{  100,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{   90,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{   80,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   70,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   50,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{   30,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{   30,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   20,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{   20,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{   20,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #45: every(binds(bind(vars(var_type(var, nil), nil), dot), nil), dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{   30,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{   20,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{   10,    0,    0,    0}, /* ExprSingle: QuantifiedExpr = 56 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* QuantifiedExpr: every(binds(bind(vars(var_type(Var, nil), nil), OptExpr), nil), OptExpr) = 77 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  210,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{  200,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{  190,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{  180,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  170,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  160,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  150,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  140,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  130,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{  120,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{  110,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{  100,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{   90,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{   80,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   70,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   50,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{   30,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{   30,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   20,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{   20,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{   20,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #46: except(dot, dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  110,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{  100,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{   90,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{   80,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  110,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   70,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{   60,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{   50,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{   40,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{   30,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{   20,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{   10,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{    0,    0,    0,    0}, /* IntersectExceptExpr: except(InstanceofExpr, IntersectExceptExpr) = 137 */
	{  200,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{  190,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{  180,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{  170,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{  160,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{  150,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{  140,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{  130,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{  120,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  110,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  110,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{  100,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{  100,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   90,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{   90,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{   90,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #47: exprseq(dot, dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   20,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{   10,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{    0,    0,    0,    0}, /* Expr: exprseq(OptExpr, OptExpr) = 53 */
	{  200,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   20,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  190,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{  180,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{  170,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{  160,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  150,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  140,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  130,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  120,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  110,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{  100,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{   90,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{   80,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{   70,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{   60,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   50,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   30,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   20,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   20,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{   10,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{   10,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  210,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{  210,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{  210,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #48: ext_expr(nil, nil) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  180,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{  170,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{  160,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{  150,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  180,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  140,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{  130,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{  120,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{  110,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  100,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{   90,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{   80,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{   70,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{   60,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{   50,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{   40,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{   30,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{   20,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{   10,    0,    0,    0}, /* ValueExpr: ExtensionExpr = 151 */
	{    0,    0,    0,    0}, /* (none) */
	{  210,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{  200,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{  190,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  180,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  180,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{  170,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{  170,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* ExtensionExpr: ext_expr(Pragmas, nil) = 210 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  160,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{  160,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{  160,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #49: ext_expr(nil, dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  180,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{  170,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{  160,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{  150,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  180,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  140,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{  130,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{  120,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{  110,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  100,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{   90,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{   80,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{   70,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{   60,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{   50,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{   40,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{   30,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{   20,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{   10,    0,    0,    0}, /* ValueExpr: ExtensionExpr = 151 */
	{    0,    0,    0,    0}, /* (none) */
	{  210,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{  200,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{  190,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  180,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  180,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{  170,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{  170,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* ExtensionExpr: ext_expr(Pragmas, Expr) = 211 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  160,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{  160,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{  160,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #50: external */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #51: flwr(nil, where(dot, ord_ret(orderby(nil), dot))) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{   30,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{   20,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{   10,    0,    0,    0}, /* ExprSingle: FLWORExpr = 55 */
	{    0,    0,    0,    0}, /* FLWORExpr: flwr(OptBindExpr, where(OptWhereClause_, ord_ret(orderby(OrderSpecList), OptExpr))) = 61 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  210,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{  200,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{  190,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{  180,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  170,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  160,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  150,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  140,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  130,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{  120,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{  110,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{  100,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{   90,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{   80,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   70,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   50,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{   30,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{   30,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   20,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{   20,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{   20,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #52: flwr(nil, where(dot, ord_ret(nil, dot))) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{   30,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{   20,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{   10,    0,    0,    0}, /* ExprSingle: FLWORExpr = 55 */
	{    0,    0,    0,    0}, /* FLWORExpr: flwr(OptBindExpr, where(OptWhereClause_, ord_ret(nil, OptExpr))) = 60 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  210,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{  200,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{  190,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{  180,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  170,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  160,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  150,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  140,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  130,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{  120,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{  110,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{  100,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{   90,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{   80,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   70,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   50,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{   30,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{   30,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   20,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{   20,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{   20,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #53: fns_decl(lit_str) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* DeclImport: DefaultNamespaceDecl = 11 */
	{    0,    0,    0,    0}, /* DefaultNamespaceDecl: fns_decl(lit_str) = 26 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #54: fun(fun_sig(nil, seq_ty(schm_attr)), dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #55: fun(fun_sig(nil, seq_ty(schm_attr)), external) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #56: fun_sig(nil, seq_ty(schm_attr)) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #57: ge(dot, dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{   50,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{   40,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{   30,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   20,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{   10,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{    0,    0,    0,    0}, /* ComparisonExpr: ge(RangeExpr, RangeExpr) = 112 */
	{  200,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  190,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  180,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  170,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  160,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  150,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{  140,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{  130,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{  120,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{  110,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{  100,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   90,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   80,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   70,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{   50,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{   50,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{   40,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{   40,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #58: gt(dot, dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{   50,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{   40,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{   30,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   20,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{   10,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{    0,    0,    0,    0}, /* ComparisonExpr: gt(RangeExpr, RangeExpr) = 111 */
	{  200,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  190,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  180,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  170,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  160,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  150,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{  140,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{  130,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{  120,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{  110,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{  100,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   90,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   80,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   70,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{   50,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{   50,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{   40,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{   40,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #59: gtgt(dot, dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{   50,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{   40,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{   30,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   20,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{   10,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{    0,    0,    0,    0}, /* ComparisonExpr: gtgt(RangeExpr, RangeExpr) = 122 */
	{  200,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  190,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  180,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  170,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  160,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  150,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{  140,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{  130,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{  120,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{  110,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{  100,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   90,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   80,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   70,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{   50,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{   50,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{   40,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{   40,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #60: idiv(dot, dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   90,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{   80,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{   70,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{   60,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   90,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   50,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{   40,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{   30,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{   20,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{   10,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{    0,    0,    0,    0}, /* MultiplicativeExpr: idiv(UnionExpr, MultiplicativeExpr) = 131 */
	{  200,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  190,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  180,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{  170,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{  160,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{  150,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{  140,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{  130,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{  120,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{  110,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{  100,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   90,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   90,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{   80,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{   80,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   70,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{   70,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{   70,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #61: if_(dot, then_else(dot, dot)) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{   30,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{   20,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{   10,    0,    0,    0}, /* ExprSingle: IfExpr = 58 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* IfExpr: if_(OptExpr, then_else(OptExpr, OptExpr)) = 101 */
	{  210,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{  200,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{  190,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{  180,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  170,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  160,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  150,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  140,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  130,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{  120,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{  110,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{  100,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{   90,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{   80,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   70,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   50,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{   30,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{   30,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   20,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{   20,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{   20,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #62: insert(dot, dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{   30,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{   20,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{   10,    0,    0,    0}, /* ExprSingle: InsertExpr = 216 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  210,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{  200,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{  190,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{  180,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  170,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  160,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  150,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  140,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  130,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{  120,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{  110,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{  100,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{   90,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{   80,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   70,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   50,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{   30,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{   30,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* InsertExpr: insert(SourceExpr, TargetExpr) = 221 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   20,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{   20,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{   20,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #63: instof(dot, seq_ty(schm_attr)) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  120,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{  110,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{  100,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{   90,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  120,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   80,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{   70,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{   60,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{   50,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{   40,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{   30,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{   20,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{   10,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{    0,    0,    0,    0}, /* InstanceofExpr: instof(TreatExpr, SequenceType) = 139 */
	{  200,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{  190,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{  180,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{  170,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{  160,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{  150,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{  140,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{  130,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  120,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  120,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{  110,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{  110,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  100,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{  100,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{  100,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #64: intersect(dot, dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  110,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{  100,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{   90,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{   80,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  110,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   70,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{   60,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{   50,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{   40,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{   30,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{   20,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{   10,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{    0,    0,    0,    0}, /* IntersectExceptExpr: intersect(InstanceofExpr, IntersectExceptExpr) = 136 */
	{  200,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{  190,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{  180,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{  170,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{  160,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{  150,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{  140,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{  130,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{  120,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  110,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  110,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{  100,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{  100,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   90,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{   90,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{   90,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #65: is(dot, dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{   50,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{   40,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{   30,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   20,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{   10,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{    0,    0,    0,    0}, /* ComparisonExpr: is(RangeExpr, RangeExpr) = 119 */
	{  200,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  190,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  180,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  170,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  160,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  150,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{  140,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{  130,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{  120,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{  110,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{  100,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   90,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   80,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   70,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{   50,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{   50,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{   40,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{   40,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #66: item_ty(nil) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* ItemType: item_ty(nil) = 86 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #67: le(dot, dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{   50,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{   40,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{   30,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   20,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{   10,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{    0,    0,    0,    0}, /* ComparisonExpr: le(RangeExpr, RangeExpr) = 110 */
	{  200,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  190,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  180,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  170,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  160,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  150,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{  140,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{  130,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{  120,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{  110,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{  100,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   90,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   80,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   70,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{   50,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{   50,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{   40,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{   40,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #68: let(var_type(var, nil), dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* TransformBinding: let(var_type(Var, nil), OptExpr) = 228 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #69: let(var_type(var, seq_ty(schm_attr)), dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #70: lib_mod(mod_ns(lit_str), nil) */
	{0},
	{    0,    0,    0,    0}, /* Query: lib_mod(ModuleDecl, Prolog) = 2 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* LibraryModule: lib_mod(ModuleDecl, Prolog) = 4 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* DeclImport: LibraryModule = 24 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #71: lit_dbl */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  230,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{  220,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{  210,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{  200,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  230,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  190,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{  180,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{  170,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{  160,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  150,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  140,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  130,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  120,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  110,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{  100,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{   90,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{   80,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{   70,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{   60,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   50,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   30,    0,    0,    0}, /* PrimaryExpr: Literal = 163 */
	{   20,    0,    0,    0}, /* Literal: NumericLiteral = 172 */
	{   10,    0,    0,    0}, /* NumericLiteral: DoubleLiteral = 176 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* DoubleLiteral: lit_dbl = 179 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  230,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  230,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{  220,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{  220,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  210,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{  210,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{  210,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #72: lit_dec */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  230,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{  220,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{  210,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{  200,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  230,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  190,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{  180,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{  170,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{  160,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  150,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  140,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  130,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  120,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  110,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{  100,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{   90,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{   80,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{   70,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{   60,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   50,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   30,    0,    0,    0}, /* PrimaryExpr: Literal = 163 */
	{   20,    0,    0,    0}, /* Literal: NumericLiteral = 172 */
	{   10,    0,    0,    0}, /* NumericLiteral: DecimalLiteral = 175 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* DecimalLiteral: lit_dec = 178 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  230,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  230,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{  220,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{  220,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  210,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{  210,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{  210,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #73: lit_int */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  230,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{  220,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{  210,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{  200,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  230,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  190,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{  180,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{  170,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{  160,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  150,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  140,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  130,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  120,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  110,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{  100,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{   90,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{   80,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{   70,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{   60,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   50,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   30,    0,    0,    0}, /* PrimaryExpr: Literal = 163 */
	{   20,    0,    0,    0}, /* Literal: NumericLiteral = 172 */
	{   10,    0,    0,    0}, /* NumericLiteral: IntegerLiteral = 174 */
	{    0,    0,    0,    0}, /* IntegerLiteral: lit_int = 177 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  230,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  230,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{  220,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{  220,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  210,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{  210,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{  210,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #74: lit_str */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  230,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{  220,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{  210,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{  200,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  230,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  190,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{  180,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{  170,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{  160,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  150,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  140,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  130,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  120,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  110,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{  100,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{   90,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{   80,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{   70,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{   60,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   50,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   30,    0,    0,    0}, /* PrimaryExpr: Literal = 163 */
	{   20,    0,    0,    0}, /* Literal: StringLiteral = 173 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* StringLiteral: lit_str = 180 */
	{    0,    0,    0,    0}, /* (none) */
	{  230,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  230,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{  220,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{  220,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  210,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{  210,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{  210,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #75: locpath(dot, dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  190,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{  180,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{  170,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{  160,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  190,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  150,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{  140,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{  130,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{  120,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  110,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  100,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{   90,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{   80,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{   70,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{   60,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{   50,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{   40,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{   30,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{   20,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* PathExpr: LocationPath = 154 */
	{    0,    0,    0,    0}, /* LocationPath: locpath(PathExpr, PathExpr) = 156 */
	{  210,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{  200,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  190,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  190,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{  180,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{  180,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  170,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{  170,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{  170,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #76: locpath(step(node_ty(nil)), dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  190,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{  180,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{  170,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{  160,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  190,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  150,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{  140,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{  130,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{  120,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  110,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  100,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{   90,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{   80,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{   70,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{   60,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{   50,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{   40,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{   30,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{   20,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* PathExpr: LocationPath = 154 */
	{    0,    0,    0,    0}, /* LocationPath: locpath(step(NodeTest), PathExpr) = 157 */
	{  210,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{  200,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  190,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  190,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{  180,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{  180,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  170,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{  170,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{  170,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #77: lt(dot, dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{   50,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{   40,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{   30,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   20,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{   10,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{    0,    0,    0,    0}, /* ComparisonExpr: lt(RangeExpr, RangeExpr) = 109 */
	{  200,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  190,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  180,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  170,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  160,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  150,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{  140,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{  130,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{  120,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{  110,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{  100,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   90,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   80,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   70,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{   50,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{   50,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{   40,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{   40,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #78: ltlt(dot, dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{   50,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{   40,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{   30,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   20,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{   10,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{    0,    0,    0,    0}, /* ComparisonExpr: ltlt(RangeExpr, RangeExpr) = 121 */
	{  200,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  190,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  180,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  170,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  160,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  150,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{  140,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{  130,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{  120,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{  110,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{  100,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   90,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   80,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   70,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{   50,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{   50,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{   40,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{   40,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #79: main_mod(nil, dot) */
	{0},
	{   10,    0,    0,    0}, /* Query: MainModule = 1 */
	{    0,    0,    0,    0}, /* MainModule: main_mod(Prolog, QueryBody) = 3 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #80: minus(dot, dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   80,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{   70,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{   60,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{   50,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   80,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{   30,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{   20,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{   10,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{    0,    0,    0,    0}, /* AdditiveExpr: minus(AdditiveExpr, MultiplicativeExpr) = 127 */
	{  200,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  190,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  180,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  170,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{  160,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{  150,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{  140,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{  130,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{  120,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{  110,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{  100,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   90,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   80,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   80,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{   70,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{   70,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{   60,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{   60,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #81: mod(dot, dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   90,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{   80,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{   70,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{   60,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   90,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   50,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{   40,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{   30,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{   20,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{   10,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{    0,    0,    0,    0}, /* MultiplicativeExpr: mod(UnionExpr, MultiplicativeExpr) = 132 */
	{  200,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  190,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  180,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{  170,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{  160,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{  150,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{  140,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{  130,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{  120,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{  110,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{  100,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   90,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   90,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{   80,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{   80,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   70,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{   70,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{   70,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #82: mod_imp(lit_str, nil) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* DeclImport: ModuleImport = 22 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* ModuleImport: mod_imp(lit_str, OptImportLoc_) = 37 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #83: mod_ns(lit_str) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* ModuleDecl: mod_ns(lit_str) = 5 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #84: modify(dot, dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #85: mult(dot, dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   90,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{   80,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{   70,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{   60,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   90,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   50,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{   40,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{   30,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{   20,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{   10,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{    0,    0,    0,    0}, /* MultiplicativeExpr: mult(UnionExpr, MultiplicativeExpr) = 129 */
	{  200,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  190,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  180,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{  170,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{  160,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{  150,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{  140,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{  130,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{  120,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{  110,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{  100,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   90,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   90,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{   80,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{   80,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   70,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{   70,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{   70,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #86: named_ty */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* SchemaType: named_ty = 100 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #87: ne(dot, dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{   50,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{   40,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{   30,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   20,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{   10,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{    0,    0,    0,    0}, /* ComparisonExpr: ne(RangeExpr, RangeExpr) = 108 */
	{  200,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  190,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  180,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  170,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  160,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  150,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{  140,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{  130,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{  120,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{  110,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{  100,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   90,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   80,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   70,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{   50,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{   50,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{   40,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{   40,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #88: nil */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   20,    0,    0,    0}, /* Prolog: DeclsImports = 6 */
	{   10,    0,    0,    0}, /* DeclsImports: nil = 8 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* OptImportLoc_: nil = 38 */
	{   10,    0,    0,    0}, /* OptParamList_: nil = 40 */
	{   10,    0,    0,    0}, /* ParamList_: nil = 42 */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* OptSequenceType_: nil = 45 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* OptBindExpr: nil = 62 */
	{   10,    0,    0,    0}, /* OptPosVar: nil = 68 */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* OptWhereClause_: nil = 71 */
	{   10,    0,    0,    0}, /* OrderSpecList: nil = 73 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* OptCaseVar: nil = 80 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* FuncArgList: nil = 186 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* Pragmas: nil = 212 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* TransformBindings: nil = 226 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* OptTypeDeclForRec: nil = 235 */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #89: node_ty(lit_str) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* ItemType: NodeTest = 85 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* NodeTest: node_ty(lit_str) = 95 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #90: node_ty(nil) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   20,    0,    0,    0}, /* ItemType: NodeTest = 85 */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* NodeTest: node_ty(nil) = 91 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #91: node_ty(req_ty(req_name, named_ty)) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   20,    0,    0,    0}, /* ItemType: NodeTest = 85 */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* NodeTest: node_ty(ElemOrAttrType) = 92 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #92: node_ty(node_ty(nil)) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* ItemType: NodeTest = 85 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* NodeTest: node_ty(node_ty(nil)) = 93 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #93: node_ty(node_ty(req_ty(req_name, named_ty))) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* ItemType: NodeTest = 85 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* NodeTest: node_ty(node_ty(ElemOrAttrType)) = 94 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #94: ns_decl(lit_str) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* DeclImport: NamespaceDecl = 19 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* NamespaceDecl: ns_decl(lit_str) = 35 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #95: option(lit_str) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* DeclImport: OptionDecl = 20 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* OptionDecl: option(lit_str) = 31 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #96: or(dot, dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{   30,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{   20,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{   10,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* OrExpr: or(AndExpr, OrExpr) = 103 */
	{  200,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{  190,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{  180,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  170,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  160,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  150,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  140,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  130,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{  120,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{  110,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{  100,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{   90,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{   80,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   70,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   50,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{   30,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{   30,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   20,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{   20,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{   20,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #97: ord_ret(nil, dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #98: ord_ret(orderby(nil), dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #99: orderby(nil) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #100: ordered(dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  210,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{  200,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{  190,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{  180,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  210,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  170,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{  160,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{  150,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{  140,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  130,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  120,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  110,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  100,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{   90,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{   80,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{   70,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{   60,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{   50,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{   40,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   30,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   20,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   10,    0,    0,    0}, /* PrimaryExpr: OrderedExpr = 169 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  210,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* OrderedExpr: ordered(OptExpr) = 188 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  210,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{  200,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{  200,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  190,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{  190,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{  190,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #101: ordering_mode */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* DeclImport: OrderingModeDecl = 16 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* OrderingModeDecl: ordering_mode = 32 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #102: orderspecs(dot, nil) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* OrderSpecList: orderspecs(OptExpr, OrderSpecList) = 74 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #103: param(seq_ty(schm_attr), var) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* Param: param(SequenceType, Var) = 44 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #104: params(param(seq_ty(schm_attr), var), nil) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* OptParamList_: ParamList_ = 41 */
	{    0,    0,    0,    0}, /* ParamList_: params(Param, ParamList_) = 43 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #105: pi(dot, dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  220,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{  210,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{  200,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{  190,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  220,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  180,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{  170,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{  160,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{  150,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  140,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  130,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  120,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  110,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  100,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{   90,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{   80,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{   70,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{   60,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{   50,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   30,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   20,    0,    0,    0}, /* PrimaryExpr: Constructor = 168 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  220,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* Constructor: PIConstructor = 194 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* PIConstructor: pi(OptExpr, OptContentExpr) = 200 */
	{    0,    0,    0,    0}, /* (none) */
	{  220,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{  210,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{  210,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  200,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{  200,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{  200,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #106: plus(dot, dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   80,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{   70,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{   60,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{   50,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   80,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{   30,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{   20,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{   10,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{    0,    0,    0,    0}, /* AdditiveExpr: plus(AdditiveExpr, MultiplicativeExpr) = 126 */
	{  200,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  190,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  180,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  170,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{  160,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{  150,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{  140,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{  130,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{  120,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{  110,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{  100,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   90,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   80,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   80,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{   70,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{   70,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{   60,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{   60,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #107: pragma */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #108: pragmas(pragma, nil) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* Pragmas: pragmas(pragma, Pragmas) = 213 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #109: pred(root_, dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  190,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{  180,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{  170,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{  160,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  190,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  150,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{  140,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{  130,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{  120,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  110,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  100,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{   90,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{   80,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{   70,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{   60,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{   50,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{   40,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{   30,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{   20,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* PathExpr: LocationPath = 154 */
	{    0,    0,    0,    0}, /* LocationPath: pred(LocationPath, OptExpr) = 158 */
	{   10,    0,    0,    0}, /* FilterExpr: pred(PathExpr, OptExpr) = 161 */
	{  200,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  190,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  190,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{  180,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{  180,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  170,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{  170,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{  170,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #110: pred(dot, dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  190,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{  180,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{  170,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{  160,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  190,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  150,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{  140,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{  130,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{  120,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  110,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  100,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{   90,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{   80,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{   70,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{   60,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{   50,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{   40,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{   30,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{   20,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* FilterExpr: pred(PathExpr, OptExpr) = 161 */
	{  200,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  190,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  190,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{  180,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{  180,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  170,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{  170,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{  170,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #111: range(dot, dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   70,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{   60,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{   50,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{   40,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   70,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   30,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{   20,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{   10,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{    0,    0,    0,    0}, /* RangeExpr: range(RangeExpr, RangeExpr) = 124 */
	{  200,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  190,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  180,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  170,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  160,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{  150,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{  140,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{  130,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{  120,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{  110,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{  100,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   90,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   80,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   70,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   70,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{   60,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{   60,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   50,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{   50,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{   50,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #112: recursion(var_type(var, nil), seed(dot, dot)) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{   30,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{   20,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{   10,    0,    0,    0}, /* ExprSingle: RecursiveExpr = 233 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  210,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{  200,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{  190,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{  180,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  170,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  160,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  150,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  140,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  130,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{  120,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{  110,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{  100,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{   90,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{   80,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   70,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   50,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{   30,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{   30,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   20,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{   20,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{   20,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* RecursiveExpr: recursion(var_type(Var, OptTypeDeclForRec), seed(ExprSingle, ExprSingle)) = 234 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #113: rename(dot, dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{   30,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{   20,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{   10,    0,    0,    0}, /* ExprSingle: RenameExpr = 218 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  210,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{  200,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{  190,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{  180,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  170,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  160,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  150,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  140,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  130,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{  120,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{  110,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{  100,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{   90,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{   80,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   70,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   50,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{   30,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{   30,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* RenameExpr: rename(SourceExpr, NewNameExpr) = 224 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   20,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{   20,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{   20,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #114: replace(dot, dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{   30,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{   20,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{   10,    0,    0,    0}, /* ExprSingle: ReplaceExpr = 219 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  210,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{  200,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{  190,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{  180,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  170,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  160,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  150,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  140,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  130,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{  120,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{  110,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{  100,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{   90,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{   80,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   70,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   50,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{   30,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{   30,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* ReplaceExpr: replace(TargetExpr, ExprSingle) = 223 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   20,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{   20,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{   20,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #115: req_name */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #116: req_ty(req_name, named_ty) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* ElemOrAttrType: req_ty(req_name, SchemaType) = 99 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #117: req_ty(nil, named_ty) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* ElemOrAttrType: req_ty(nil, SchemaType) = 97 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #118: req_ty(nil, nil) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* ElemOrAttrType: req_ty(nil, nil) = 96 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #119: req_ty(req_name, nil) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* ElemOrAttrType: req_ty(req_name, nil) = 98 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #120: revalid */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* DeclImport: RevalidationDecl = 214 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* RevalidationDecl: revalid = 215 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #121: root_ */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  190,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{  180,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{  170,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{  160,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  190,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  150,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{  140,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{  130,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{  120,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  110,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  100,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{   90,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{   80,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{   70,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{   60,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{   50,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{   40,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{   30,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{   20,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* PathExpr: LocationPath = 154 */
	{    0,    0,    0,    0}, /* LocationPath: root_ = 155 */
	{  210,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{  200,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  190,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  190,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{  180,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{  180,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  170,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{  170,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{  170,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #122: schm_ats(lit_str, nil) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* OptImportLoc_: schm_ats(lit_str, OptImportLoc_) = 39 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #123: schm_attr */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* ItemType: schm_attr = 89 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #124: schm_elem */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* ItemType: schm_elem = 88 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #125: schm_imp(lit_str, nil) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* DeclImport: SchemaImport = 21 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* SchemaImport: schm_imp(lit_str, OptImportLoc_) = 36 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #126: seed(dot, dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #127: seq_ty(atom_ty(nil)) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* OptSequenceType_: SequenceType = 46 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* TypeDeclaration: SequenceType = 70 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* SequenceType: seq_ty(ItemType) = 82 */
	{    0,    0,    0,    0}, /* SingleType: seq_ty(AtomType) = 83 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* OptTypeDeclForRec: SequenceType = 236 */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #128: seq_ty(stmt_ty(nil)) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* OptSequenceType_: SequenceType = 46 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* TypeDeclaration: SequenceType = 70 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* SequenceType: seq_ty(stmt_ty(nil)) = 232 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* OptTypeDeclForRec: SequenceType = 236 */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #129: seq_ty(docmgmt_ty(nil)) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* OptSequenceType_: SequenceType = 46 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* TypeDeclaration: SequenceType = 70 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* SequenceType: seq_ty(docmgmt_ty(nil)) = 238 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* OptTypeDeclForRec: SequenceType = 236 */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #130: seq_ty(schm_attr) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* OptSequenceType_: SequenceType = 46 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* TypeDeclaration: SequenceType = 70 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* SequenceType: seq_ty(ItemType) = 82 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* OptTypeDeclForRec: SequenceType = 236 */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #131: some(binds(bind(vars(var_type(var, seq_ty(schm_attr)), nil), dot), nil), dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{   30,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{   20,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{   10,    0,    0,    0}, /* ExprSingle: QuantifiedExpr = 56 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* QuantifiedExpr: some(binds(bind(vars(var_type(Var, TypeDeclaration), nil), OptExpr), nil), OptExpr) = 76 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  210,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{  200,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{  190,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{  180,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  170,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  160,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  150,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  140,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  130,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{  120,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{  110,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{  100,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{   90,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{   80,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   70,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   50,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{   30,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{   30,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   20,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{   20,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{   20,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #132: some(binds(bind(vars(var_type(var, nil), nil), dot), nil), dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{   30,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{   20,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{   10,    0,    0,    0}, /* ExprSingle: QuantifiedExpr = 56 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* QuantifiedExpr: some(binds(bind(vars(var_type(Var, nil), nil), OptExpr), nil), OptExpr) = 75 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  210,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{  200,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{  190,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{  180,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  170,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  160,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  150,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  140,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  130,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{  120,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{  110,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{  100,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{   90,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{   80,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   70,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   50,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{   30,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{   30,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   20,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{   20,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{   20,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #133: step(node_ty(nil)) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #134: stmt_ty(nil) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #135: tag */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* TagName: tag = 202 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #136: text(dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  220,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{  210,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{  200,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{  190,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  220,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  180,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{  170,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{  160,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{  150,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  140,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  130,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  120,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  110,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  100,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{   90,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{   80,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{   70,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{   60,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{   50,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   30,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   20,    0,    0,    0}, /* PrimaryExpr: Constructor = 168 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  220,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* Constructor: TextConstructor = 192 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* TextConstructor: text(OptExpr) = 198 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  220,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{  210,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{  210,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  200,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{  200,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{  200,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #137: then_else(dot, dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #138: transbinds(let(var_type(var, nil), dot), nil) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* TransformBindings: transbinds(TransformBinding, TransformBindings) = 227 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #139: transform(nil, modify(dot, dot)) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{   30,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{   20,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{   10,    0,    0,    0}, /* ExprSingle: TransformExpr = 220 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  210,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{  200,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{  190,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{  180,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  170,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  160,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  150,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  140,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  130,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{  120,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{  110,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{  100,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{   90,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{   80,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   70,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   50,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{   30,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{   30,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* TransformExpr: transform(TransformBindings, modify(ExprSingle, ExprSingle)) = 225 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   20,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{   20,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{   20,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #140: treat(dot, seq_ty(schm_attr)) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  130,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{  120,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{  110,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{  100,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  130,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   90,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{   80,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{   70,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{   60,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{   50,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{   40,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{   30,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{   20,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{   10,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{    0,    0,    0,    0}, /* TreatExpr: treat(CastableExpr, SequenceType) = 141 */
	{  200,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{  190,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{  180,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{  170,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{  160,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{  150,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{  140,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  130,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  130,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{  120,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{  120,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  110,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{  110,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{  110,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #141: typeswitch(dot, cases(case_(var_type(var, seq_ty(schm_attr)), dot), cases(default_(var, dot), nil))) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{   30,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{   20,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{   10,    0,    0,    0}, /* ExprSingle: TypeswitchExpr = 57 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* TypeswitchExpr: typeswitch(OptExpr, cases(case_(var_type(OptCaseVar, SequenceType), OptExpr), cases(default_(OptCaseVar, OptExpr), nil))) = 79 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  210,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{  200,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{  190,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{  180,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  170,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  160,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  150,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  140,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  130,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{  120,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{  110,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{  100,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{   90,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{   80,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   70,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   50,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{   30,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{   30,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   20,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{   20,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{   20,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #142: uminus(dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  160,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{  150,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{  140,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{  130,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  160,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  120,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{  110,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{  100,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{   90,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{   80,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{   70,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{   60,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{   50,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{   40,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{   30,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{   20,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{   10,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{    0,    0,    0,    0}, /* UnaryExpr: uminus(ValueExpr) = 147 */
	{  200,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{  190,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{  180,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{  170,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  160,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  160,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{  150,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{  150,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  140,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{  140,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{  140,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #143: union_(dot, dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  100,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{   90,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{   80,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{   70,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  100,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{   50,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{   40,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{   30,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{   20,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{   10,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{    0,    0,    0,    0}, /* UnionExpr: union_(IntersectExceptExpr, UnionExpr) = 134 */
	{  200,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  190,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{  180,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{  170,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{  160,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{  150,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{  140,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{  130,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{  120,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{  110,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  100,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  100,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{   90,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{   90,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   80,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{   80,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{   80,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #144: unordered(dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  210,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{  200,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{  190,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{  180,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  210,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  170,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{  160,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{  150,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{  140,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  130,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  120,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  110,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  100,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{   90,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{   80,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{   70,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{   60,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{   50,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{   40,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   30,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   20,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   10,    0,    0,    0}, /* PrimaryExpr: UnorderedExpr = 170 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  210,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* UnorderedExpr: unordered(OptExpr) = 189 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  210,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{  200,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{  200,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  190,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{  190,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{  190,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #145: uplus(dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  160,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{  150,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{  140,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{  130,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  160,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  120,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{  110,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{  100,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{   90,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{   80,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{   70,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{   60,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{   50,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{   40,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{   30,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{   20,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{   10,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{    0,    0,    0,    0}, /* UnaryExpr: uplus(ValueExpr) = 148 */
	{  200,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{  190,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{  180,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{  170,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  160,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  160,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{  150,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{  150,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  140,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{  140,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{  140,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #146: val_eq(dot, dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{   50,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{   40,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{   30,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   20,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{   10,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{    0,    0,    0,    0}, /* ComparisonExpr: val_eq(RangeExpr, RangeExpr) = 113 */
	{  200,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  190,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  180,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  170,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  160,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  150,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{  140,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{  130,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{  120,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{  110,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{  100,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   90,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   80,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   70,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{   50,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{   50,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{   40,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{   40,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #147: val_ge(dot, dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{   50,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{   40,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{   30,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   20,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{   10,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{    0,    0,    0,    0}, /* ComparisonExpr: val_ge(RangeExpr, RangeExpr) = 118 */
	{  200,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  190,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  180,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  170,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  160,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  150,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{  140,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{  130,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{  120,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{  110,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{  100,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   90,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   80,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   70,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{   50,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{   50,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{   40,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{   40,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #148: val_gt(dot, dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{   50,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{   40,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{   30,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   20,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{   10,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{    0,    0,    0,    0}, /* ComparisonExpr: val_gt(RangeExpr, RangeExpr) = 117 */
	{  200,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  190,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  180,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  170,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  160,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  150,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{  140,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{  130,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{  120,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{  110,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{  100,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   90,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   80,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   70,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{   50,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{   50,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{   40,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{   40,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #149: val_le(dot, dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{   50,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{   40,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{   30,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   20,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{   10,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{    0,    0,    0,    0}, /* ComparisonExpr: val_le(RangeExpr, RangeExpr) = 116 */
	{  200,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  190,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  180,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  170,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  160,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  150,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{  140,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{  130,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{  120,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{  110,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{  100,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   90,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   80,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   70,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{   50,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{   50,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{   40,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{   40,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #150: val_lt(dot, dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{   50,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{   40,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{   30,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   20,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{   10,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{    0,    0,    0,    0}, /* ComparisonExpr: val_lt(RangeExpr, RangeExpr) = 115 */
	{  200,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  190,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  180,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  170,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  160,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  150,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{  140,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{  130,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{  120,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{  110,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{  100,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   90,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   80,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   70,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{   50,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{   50,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{   40,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{   40,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #151: val_ne(dot, dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{   50,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{   40,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{   30,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   20,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{   10,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{    0,    0,    0,    0}, /* ComparisonExpr: val_ne(RangeExpr, RangeExpr) = 114 */
	{  200,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  190,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  180,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  170,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  160,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{  150,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{  140,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{  130,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{  120,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{  110,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{  100,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   90,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   80,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   70,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   60,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{   50,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{   50,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   40,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{   40,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{   40,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #152: validate(dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  180,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{  170,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{  160,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{  150,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  180,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  140,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{  130,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{  120,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{  110,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  100,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{   90,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{   80,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{   70,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{   60,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{   50,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{   40,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{   30,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{   20,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{   10,    0,    0,    0}, /* ValueExpr: ValidateExpr = 149 */
	{    0,    0,    0,    0}, /* ValidateExpr: validate(OptExpr) = 152 */
	{  210,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{  200,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{  190,    0,    0,    0}, /* PrimaryExpr: ParenthesizedExpr = 165 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  180,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  180,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{  170,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{  170,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  160,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{  160,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{  160,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #153: var */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  210,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{  200,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{  190,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{  180,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* OptPosVar: Var = 69 */
	{    0,    0,    0,    0}, /* (none) */
	{  210,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* OptCaseVar: Var = 81 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  170,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{  160,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{  150,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{  140,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  130,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  120,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  110,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  100,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{   90,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{   80,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{   70,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{   60,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{   50,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{   40,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   30,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   20,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   10,    0,    0,    0}, /* PrimaryExpr: Var = 164 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* Var: var = 181 */
	{  210,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  210,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{  200,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{  200,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  190,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{  190,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{  190,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #154: var_decl(var_type(var, nil), dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* DeclImport: VariableDeclaration = 23 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* VariableDeclaration: var_decl(var_type(Var, nil), OptExpr) = 47 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #155: var_decl(var_type(var, seq_ty(schm_attr)), dot) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* DeclImport: VariableDeclaration = 23 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* VariableDeclaration: var_decl(var_type(Var, SequenceType), OptExpr) = 48 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #156: var_decl(var_type(var, nil), external) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{   10,    0,    0,    0}, /* DeclImport: VariableDeclaration = 23 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* VariableDeclaration: var_decl(var_type(Var, OptSequenceType_), external) = 49 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #157: var_type(nil, seq_ty(schm_attr)) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #158: var_type(var, nil) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #159: var_type(var, seq_ty(schm_attr)) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #160: vars(var_type(var, seq_ty(schm_attr)), var) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #161: vars(var_type(var, seq_ty(schm_attr)), nil) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #162: vars(var_type(var, nil), var) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #163: vars(var_type(var, nil), nil) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #164: where(dot, ord_ret(orderby(nil), dot)) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #165: where(dot, ord_ret(nil, dot)) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
},
{ /* state #166: xrpc(dot, apply(nil)) */
	{0},
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  210,    0,    0,    0}, /* QueryBody: OptExpr = 50 */
	{  200,    0,    0,    0}, /* OptExpr: Expr = 52 */
	{  190,    0,    0,    0}, /* Expr: ExprSingle = 54 */
	{  180,    0,    0,    0}, /* ExprSingle: OrExpr = 59 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  210,    0,    0,    0}, /* OptWhereClause_: OptExpr = 72 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  170,    0,    0,    0}, /* OrExpr: AndExpr = 102 */
	{  160,    0,    0,    0}, /* AndExpr: ComparisonExpr = 104 */
	{  150,    0,    0,    0}, /* ComparisonExpr: RangeExpr = 106 */
	{  140,    0,    0,    0}, /* RangeExpr: AdditiveExpr = 123 */
	{  130,    0,    0,    0}, /* AdditiveExpr: MultiplicativeExpr = 125 */
	{  120,    0,    0,    0}, /* MultiplicativeExpr: UnionExpr = 128 */
	{  110,    0,    0,    0}, /* UnionExpr: IntersectExceptExpr = 133 */
	{  100,    0,    0,    0}, /* IntersectExceptExpr: InstanceofExpr = 135 */
	{   90,    0,    0,    0}, /* InstanceofExpr: TreatExpr = 138 */
	{   80,    0,    0,    0}, /* TreatExpr: CastableExpr = 140 */
	{   70,    0,    0,    0}, /* CastableExpr: CastExpr = 142 */
	{   60,    0,    0,    0}, /* CastExpr: UnaryExpr = 144 */
	{   50,    0,    0,    0}, /* UnaryExpr: ValueExpr = 146 */
	{   40,    0,    0,    0}, /* ValueExpr: PathExpr = 150 */
	{    0,    0,    0,    0}, /* (none) */
	{   30,    0,    0,    0}, /* PathExpr: FilterExpr = 153 */
	{    0,    0,    0,    0}, /* (none) */
	{   20,    0,    0,    0}, /* FilterExpr: PrimaryExpr = 160 */
	{   10,    0,    0,    0}, /* PrimaryExpr: XRPCCall = 171 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  210,    0,    0,    0}, /* ParenthesizedExpr: OptExpr = 182 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  210,    0,    0,    0}, /* TagName: OptExpr = 203 */
	{  200,    0,    0,    0}, /* OptContentExpr: Expr = 204 */
	{  200,    0,    0,    0}, /* AttributeValue: Expr = 207 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{  190,    0,    0,    0}, /* SourceExpr: ExprSingle = 229 */
	{  190,    0,    0,    0}, /* TargetExpr: ExprSingle = 230 */
	{  190,    0,    0,    0}, /* NewNameExpr: ExprSingle = 231 */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* (none) */
	{    0,    0,    0,    0}, /* XRPCCall: xrpc(Expr, FunctionCall) = 237 */
},
};

char * PFfs_state_string[] = {
" not a state", /* state 0 */
	"and(dot, dot)", /* state #1 */
	"apply(args(dot, nil))", /* state #2 */
	"apply(nil)", /* state #3 */
	"args(dot, nil)", /* state #4 */
	"atom_ty(nil)", /* state #5 */
	"attr(tag, dot)", /* state #6 */
	"base_uri(lit_str)", /* state #7 */
	"bind(vars(var_type(var, seq_ty(schm_attr)), nil), dot)", /* state #8 */
	"bind(vars(var_type(var, seq_ty(schm_attr)), var), dot)", /* state #9 */
	"bind(vars(var_type(var, nil), var), dot)", /* state #10 */
	"bind(vars(var_type(var, nil), nil), dot)", /* state #11 */
	"binds(let(var_type(var, seq_ty(schm_attr)), dot), nil)", /* state #12 */
	"binds(bind(vars(var_type(var, seq_ty(schm_attr)), nil), dot), nil)", /* state #13 */
	"binds(bind(vars(var_type(var, seq_ty(schm_attr)), var), dot), nil)", /* state #14 */
	"binds(bind(vars(var_type(var, nil), nil), dot), nil)", /* state #15 */
	"binds(bind(vars(var_type(var, nil), var), dot), nil)", /* state #16 */
	"binds(let(var_type(var, nil), dot), nil)", /* state #17 */
	"boundspc_decl", /* state #18 */
	"case_(var_type(var, seq_ty(schm_attr)), dot)", /* state #19 */
	"cases(case_(var_type(var, seq_ty(schm_attr)), dot), cases(default_(var, dot), nil))", /* state #20 */
	"cases(default_(var, dot), nil)", /* state #21 */
	"cast(dot, seq_ty(atom_ty(nil)))", /* state #22 */
	"castable(dot, seq_ty(atom_ty(nil)))", /* state #23 */
	"coll_decl(lit_str)", /* state #24 */
	"comment(dot)", /* state #25 */
	"constr_decl", /* state #26 */
	"contseq(dot, dot)", /* state #27 */
	"copy_ns", /* state #28 */
	"decl_imps(revalid, nil)", /* state #29 */
	"decl_imps(fun(fun_sig(nil, seq_ty(schm_attr)), external), nil)", /* state #30 */
	"decl_imps(fun(fun_sig(nil, seq_ty(schm_attr)), dot), nil)", /* state #31 */
	"def_order", /* state #32 */
	"default_(var, dot)", /* state #33 */
	"delete(dot)", /* state #34 */
	"div(dot, dot)", /* state #35 */
	"doc(dot)", /* state #36 */
	"docmgmt_ty(nil)", /* state #37 */
	"dot", /* state #38 */
	"elem(tag, dot)", /* state #39 */
	"empty_seq", /* state #40 */
	"empty_ty", /* state #41 */
	"ens_decl(lit_str)", /* state #42 */
	"eq(dot, dot)", /* state #43 */
	"every(binds(bind(vars(var_type(var, seq_ty(schm_attr)), nil), dot), nil), dot)", /* state #44 */
	"every(binds(bind(vars(var_type(var, nil), nil), dot), nil), dot)", /* state #45 */
	"except(dot, dot)", /* state #46 */
	"exprseq(dot, dot)", /* state #47 */
	"ext_expr(nil, nil)", /* state #48 */
	"ext_expr(nil, dot)", /* state #49 */
	"external", /* state #50 */
	"flwr(nil, where(dot, ord_ret(orderby(nil), dot)))", /* state #51 */
	"flwr(nil, where(dot, ord_ret(nil, dot)))", /* state #52 */
	"fns_decl(lit_str)", /* state #53 */
	"fun(fun_sig(nil, seq_ty(schm_attr)), dot)", /* state #54 */
	"fun(fun_sig(nil, seq_ty(schm_attr)), external)", /* state #55 */
	"fun_sig(nil, seq_ty(schm_attr))", /* state #56 */
	"ge(dot, dot)", /* state #57 */
	"gt(dot, dot)", /* state #58 */
	"gtgt(dot, dot)", /* state #59 */
	"idiv(dot, dot)", /* state #60 */
	"if_(dot, then_else(dot, dot))", /* state #61 */
	"insert(dot, dot)", /* state #62 */
	"instof(dot, seq_ty(schm_attr))", /* state #63 */
	"intersect(dot, dot)", /* state #64 */
	"is(dot, dot)", /* state #65 */
	"item_ty(nil)", /* state #66 */
	"le(dot, dot)", /* state #67 */
	"let(var_type(var, nil), dot)", /* state #68 */
	"let(var_type(var, seq_ty(schm_attr)), dot)", /* state #69 */
	"lib_mod(mod_ns(lit_str), nil)", /* state #70 */
	"lit_dbl", /* state #71 */
	"lit_dec", /* state #72 */
	"lit_int", /* state #73 */
	"lit_str", /* state #74 */
	"locpath(dot, dot)", /* state #75 */
	"locpath(step(node_ty(nil)), dot)", /* state #76 */
	"lt(dot, dot)", /* state #77 */
	"ltlt(dot, dot)", /* state #78 */
	"main_mod(nil, dot)", /* state #79 */
	"minus(dot, dot)", /* state #80 */
	"mod(dot, dot)", /* state #81 */
	"mod_imp(lit_str, nil)", /* state #82 */
	"mod_ns(lit_str)", /* state #83 */
	"modify(dot, dot)", /* state #84 */
	"mult(dot, dot)", /* state #85 */
	"named_ty", /* state #86 */
	"ne(dot, dot)", /* state #87 */
	"nil", /* state #88 */
	"node_ty(lit_str)", /* state #89 */
	"node_ty(nil)", /* state #90 */
	"node_ty(req_ty(req_name, named_ty))", /* state #91 */
	"node_ty(node_ty(nil))", /* state #92 */
	"node_ty(node_ty(req_ty(req_name, named_ty)))", /* state #93 */
	"ns_decl(lit_str)", /* state #94 */
	"option(lit_str)", /* state #95 */
	"or(dot, dot)", /* state #96 */
	"ord_ret(nil, dot)", /* state #97 */
	"ord_ret(orderby(nil), dot)", /* state #98 */
	"orderby(nil)", /* state #99 */
	"ordered(dot)", /* state #100 */
	"ordering_mode", /* state #101 */
	"orderspecs(dot, nil)", /* state #102 */
	"param(seq_ty(schm_attr), var)", /* state #103 */
	"params(param(seq_ty(schm_attr), var), nil)", /* state #104 */
	"pi(dot, dot)", /* state #105 */
	"plus(dot, dot)", /* state #106 */
	"pragma", /* state #107 */
	"pragmas(pragma, nil)", /* state #108 */
	"pred(root_, dot)", /* state #109 */
	"pred(dot, dot)", /* state #110 */
	"range(dot, dot)", /* state #111 */
	"recursion(var_type(var, nil), seed(dot, dot))", /* state #112 */
	"rename(dot, dot)", /* state #113 */
	"replace(dot, dot)", /* state #114 */
	"req_name", /* state #115 */
	"req_ty(req_name, named_ty)", /* state #116 */
	"req_ty(nil, named_ty)", /* state #117 */
	"req_ty(nil, nil)", /* state #118 */
	"req_ty(req_name, nil)", /* state #119 */
	"revalid", /* state #120 */
	"root_", /* state #121 */
	"schm_ats(lit_str, nil)", /* state #122 */
	"schm_attr", /* state #123 */
	"schm_elem", /* state #124 */
	"schm_imp(lit_str, nil)", /* state #125 */
	"seed(dot, dot)", /* state #126 */
	"seq_ty(atom_ty(nil))", /* state #127 */
	"seq_ty(stmt_ty(nil))", /* state #128 */
	"seq_ty(docmgmt_ty(nil))", /* state #129 */
	"seq_ty(schm_attr)", /* state #130 */
	"some(binds(bind(vars(var_type(var, seq_ty(schm_attr)), nil), dot), nil), dot)", /* state #131 */
	"some(binds(bind(vars(var_type(var, nil), nil), dot), nil), dot)", /* state #132 */
	"step(node_ty(nil))", /* state #133 */
	"stmt_ty(nil)", /* state #134 */
	"tag", /* state #135 */
	"text(dot)", /* state #136 */
	"then_else(dot, dot)", /* state #137 */
	"transbinds(let(var_type(var, nil), dot), nil)", /* state #138 */
	"transform(nil, modify(dot, dot))", /* state #139 */
	"treat(dot, seq_ty(schm_attr))", /* state #140 */
	"typeswitch(dot, cases(case_(var_type(var, seq_ty(schm_attr)), dot), cases(default_(var, dot), nil)))", /* state #141 */
	"uminus(dot)", /* state #142 */
	"union_(dot, dot)", /* state #143 */
	"unordered(dot)", /* state #144 */
	"uplus(dot)", /* state #145 */
	"val_eq(dot, dot)", /* state #146 */
	"val_ge(dot, dot)", /* state #147 */
	"val_gt(dot, dot)", /* state #148 */
	"val_le(dot, dot)", /* state #149 */
	"val_lt(dot, dot)", /* state #150 */
	"val_ne(dot, dot)", /* state #151 */
	"validate(dot)", /* state #152 */
	"var", /* state #153 */
	"var_decl(var_type(var, nil), dot)", /* state #154 */
	"var_decl(var_type(var, seq_ty(schm_attr)), dot)", /* state #155 */
	"var_decl(var_type(var, nil), external)", /* state #156 */
	"var_type(nil, seq_ty(schm_attr))", /* state #157 */
	"var_type(var, nil)", /* state #158 */
	"var_type(var, seq_ty(schm_attr))", /* state #159 */
	"vars(var_type(var, seq_ty(schm_attr)), var)", /* state #160 */
	"vars(var_type(var, seq_ty(schm_attr)), nil)", /* state #161 */
	"vars(var_type(var, nil), var)", /* state #162 */
	"vars(var_type(var, nil), nil)", /* state #163 */
	"where(dot, ord_ret(orderby(nil), dot))", /* state #164 */
	"where(dot, ord_ret(nil, dot))", /* state #165 */
	"xrpc(dot, apply(nil))", /* state #166 */
};
char *PFfs_ntname[] = {
	"Error: Nonterminals are > 0",
	"Query",
	"MainModule",
	"LibraryModule",
	"ModuleDecl",
	"Prolog",
	"DeclsImports",
	"DeclImport",
	"DefaultNamespaceDecl",
	"BoundarySpaceDecl",
	"DefaultCollationDecl",
	"BaseURIDecl",
	"ConstructionDecl",
	"OptionDecl",
	"OrderingModeDecl",
	"EmptyOrderingDecl",
	"CopyNamespacesDecl",
	"NamespaceDecl",
	"SchemaImport",
	"ModuleImport",
	"OptImportLoc_",
	"OptParamList_",
	"ParamList_",
	"Param",
	"OptSequenceType_",
	"VariableDeclaration",
	"QueryBody",
	"OptExpr",
	"Expr",
	"ExprSingle",
	"FLWORExpr",
	"OptBindExpr",
	"OptPosVar",
	"TypeDeclaration",
	"OptWhereClause_",
	"OrderSpecList",
	"QuantifiedExpr",
	"TypeswitchExpr",
	"OptCaseVar",
	"SequenceType",
	"SingleType",
	"ItemType",
	"AtomType",
	"NodeTest",
	"ElemOrAttrType",
	"SchemaType",
	"IfExpr",
	"OrExpr",
	"AndExpr",
	"ComparisonExpr",
	"RangeExpr",
	"AdditiveExpr",
	"MultiplicativeExpr",
	"UnionExpr",
	"IntersectExceptExpr",
	"InstanceofExpr",
	"TreatExpr",
	"CastableExpr",
	"CastExpr",
	"UnaryExpr",
	"ValueExpr",
	"ValidateExpr",
	"PathExpr",
	"LocationPath",
	"FilterExpr",
	"PrimaryExpr",
	"Literal",
	"NumericLiteral",
	"IntegerLiteral",
	"DecimalLiteral",
	"DoubleLiteral",
	"StringLiteral",
	"Var",
	"ParenthesizedExpr",
	"ContextItemExpr",
	"FunctionCall",
	"FuncArgList",
	"OrderedExpr",
	"UnorderedExpr",
	"Constructor",
	"ElementConstructor",
	"AttributeConstructor",
	"TextConstructor",
	"CommentConstructor",
	"PIConstructor",
	"DocumentConstructor",
	"TagName",
	"OptContentExpr",
	"AttributeValue",
	"ExtensionExpr",
	"Pragmas",
	"RevalidationDecl",
	"InsertExpr",
	"DeleteExpr",
	"ReplaceExpr",
	"RenameExpr",
	"TransformExpr",
	"TransformBindings",
	"TransformBinding",
	"SourceExpr",
	"TargetExpr",
	"NewNameExpr",
	"RecursiveExpr",
	"OptTypeDeclForRec",
	"XRPCCall",
	0
};

short PFfs_closure[105][105] = {
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    1,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    6,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,   24,    0,    0,    0,    0,   11,   12,
	    13,   14,   15,   20,   16,   17,   18,   19,   21,   22,
	     0,    0,    0,    0,    0,   23,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,  214,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,   41,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,   46,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,   50,   50,   50,
	    50,    0,    0,    0,    0,    0,   50,   50,    0,    0,
	     0,    0,    0,    0,    0,    0,   50,   50,   50,   50,
	    50,   50,   50,   50,   50,   50,   50,   50,   50,   50,
	    50,   50,   50,   50,   50,   50,   50,   50,   50,   50,
	    50,   50,   50,   50,   50,   50,    0,   50,   50,   50,
	    50,   50,   50,   50,   50,   50,    0,    0,    0,   50,
	     0,    0,   50,   50,   50,   50,   50,    0,    0,    0,
	     0,    0,   50,    0,   50,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,   52,   52,
	    52,    0,    0,    0,    0,    0,   52,   52,    0,    0,
	     0,    0,    0,    0,    0,    0,   52,   52,   52,   52,
	    52,   52,   52,   52,   52,   52,   52,   52,   52,   52,
	    52,   52,   52,   52,   52,   52,   52,   52,   52,   52,
	    52,   52,   52,   52,   52,   52,    0,   52,   52,   52,
	    52,   52,   52,   52,   52,   52,    0,    0,    0,   52,
	     0,    0,   52,   52,   52,   52,   52,    0,    0,    0,
	     0,    0,   52,    0,   52,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,   54,    0,   54,
	    54,    0,    0,    0,    0,    0,   54,   54,    0,    0,
	     0,    0,    0,    0,    0,    0,   54,   54,   54,   54,
	    54,   54,   54,   54,   54,   54,   54,   54,   54,   54,
	    54,   54,   54,   54,   54,   54,   54,   54,   54,   54,
	    54,   54,   54,   54,   54,   54,    0,   54,   54,   54,
	    54,   54,   54,   54,   54,   54,    0,    0,    0,   54,
	     0,    0,   54,   54,   54,   54,   54,    0,    0,    0,
	     0,    0,   54,    0,   54,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,   59,   59,    0,
	    55,    0,    0,    0,    0,    0,   56,   57,    0,    0,
	     0,    0,    0,    0,    0,    0,   58,   59,   59,   59,
	    59,   59,   59,   59,   59,   59,   59,   59,   59,   59,
	    59,   59,   59,   59,   59,   59,   59,   59,   59,   59,
	    59,   59,   59,   59,   59,   59,    0,   59,   59,   59,
	    59,   59,   59,   59,   59,   59,    0,    0,    0,   59,
	     0,    0,  216,  217,  219,  218,  220,    0,    0,    0,
	     0,    0,  233,    0,   59,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,   69,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,   70,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,   72,   72,   72,
	    72,    0,    0,    0,    0,    0,   72,   72,    0,    0,
	     0,    0,    0,    0,    0,    0,   72,   72,   72,   72,
	    72,   72,   72,   72,   72,   72,   72,   72,   72,   72,
	    72,   72,   72,   72,   72,   72,   72,   72,   72,   72,
	    72,   72,   72,   72,   72,   72,    0,   72,   72,   72,
	    72,   72,   72,   72,   72,   72,    0,    0,    0,   72,
	     0,    0,   72,   72,   72,   72,   72,    0,    0,    0,
	     0,    0,   72,    0,   72,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,   81,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,   85,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,  102,  102,  102,
	   102,    0,    0,    0,    0,    0,  102,  102,    0,    0,
	     0,    0,    0,    0,    0,    0,  102,    0,  102,  102,
	   102,  102,  102,  102,  102,  102,  102,  102,  102,  102,
	   102,  102,  102,  102,  102,  102,  102,  102,  102,  102,
	   102,  102,  102,  102,  102,  102,    0,  102,  102,  102,
	   102,  102,  102,  102,  102,  102,    0,    0,    0,  102,
	     0,    0,  102,  102,  102,  102,  102,    0,    0,    0,
	     0,    0,  102,    0,  102,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,  104,  104,  104,
	   104,    0,    0,    0,    0,    0,  104,  104,    0,    0,
	     0,    0,    0,    0,    0,    0,  104,  104,    0,  104,
	   104,  104,  104,  104,  104,  104,  104,  104,  104,  104,
	   104,  104,  104,  104,  104,  104,  104,  104,  104,  104,
	   104,  104,  104,  104,  104,  104,    0,  104,  104,  104,
	   104,  104,  104,  104,  104,  104,    0,    0,    0,  104,
	     0,    0,  104,  104,  104,  104,  104,    0,    0,    0,
	     0,    0,  104,    0,  104,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,  106,  106,  106,
	   106,    0,    0,    0,    0,    0,  106,  106,    0,    0,
	     0,    0,    0,    0,    0,    0,  106,  106,  106,    0,
	   106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
	   106,  106,  106,  106,  106,  106,  106,  106,  106,  106,
	   106,  106,  106,  106,  106,  106,    0,  106,  106,  106,
	   106,  106,  106,  106,  106,  106,    0,    0,    0,  106,
	     0,    0,  106,  106,  106,  106,  106,    0,    0,    0,
	     0,    0,  106,    0,  106,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,  123,  123,  123,
	   123,    0,    0,    0,    0,    0,  123,  123,    0,    0,
	     0,    0,    0,    0,    0,    0,  123,  123,  123,  123,
	     0,  123,  123,  123,  123,  123,  123,  123,  123,  123,
	   123,  123,  123,  123,  123,  123,  123,  123,  123,  123,
	   123,  123,  123,  123,  123,  123,    0,  123,  123,  123,
	   123,  123,  123,  123,  123,  123,    0,    0,    0,  123,
	     0,    0,  123,  123,  123,  123,  123,    0,    0,    0,
	     0,    0,  123,    0,  123,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,  125,  125,  125,
	   125,    0,    0,    0,    0,    0,  125,  125,    0,    0,
	     0,    0,    0,    0,    0,    0,  125,  125,  125,  125,
	   125,    0,  125,  125,  125,  125,  125,  125,  125,  125,
	   125,  125,  125,  125,  125,  125,  125,  125,  125,  125,
	   125,  125,  125,  125,  125,  125,    0,  125,  125,  125,
	   125,  125,  125,  125,  125,  125,    0,    0,    0,  125,
	     0,    0,  125,  125,  125,  125,  125,    0,    0,    0,
	     0,    0,  125,    0,  125,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,  128,  128,  128,
	   128,    0,    0,    0,    0,    0,  128,  128,    0,    0,
	     0,    0,    0,    0,    0,    0,  128,  128,  128,  128,
	   128,  128,    0,  128,  128,  128,  128,  128,  128,  128,
	   128,  128,  128,  128,  128,  128,  128,  128,  128,  128,
	   128,  128,  128,  128,  128,  128,    0,  128,  128,  128,
	   128,  128,  128,  128,  128,  128,    0,    0,    0,  128,
	     0,    0,  128,  128,  128,  128,  128,    0,    0,    0,
	     0,    0,  128,    0,  128,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,  133,  133,  133,
	   133,    0,    0,    0,    0,    0,  133,  133,    0,    0,
	     0,    0,    0,    0,    0,    0,  133,  133,  133,  133,
	   133,  133,  133,    0,  133,  133,  133,  133,  133,  133,
	   133,  133,  133,  133,  133,  133,  133,  133,  133,  133,
	   133,  133,  133,  133,  133,  133,    0,  133,  133,  133,
	   133,  133,  133,  133,  133,  133,    0,    0,    0,  133,
	     0,    0,  133,  133,  133,  133,  133,    0,    0,    0,
	     0,    0,  133,    0,  133,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,  135,  135,  135,
	   135,    0,    0,    0,    0,    0,  135,  135,    0,    0,
	     0,    0,    0,    0,    0,    0,  135,  135,  135,  135,
	   135,  135,  135,  135,    0,  135,  135,  135,  135,  135,
	   135,  135,  135,  135,  135,  135,  135,  135,  135,  135,
	   135,  135,  135,  135,  135,  135,    0,  135,  135,  135,
	   135,  135,  135,  135,  135,  135,    0,    0,    0,  135,
	     0,    0,  135,  135,  135,  135,  135,    0,    0,    0,
	     0,    0,  135,    0,  135,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,  138,  138,  138,
	   138,    0,    0,    0,    0,    0,  138,  138,    0,    0,
	     0,    0,    0,    0,    0,    0,  138,  138,  138,  138,
	   138,  138,  138,  138,  138,    0,  138,  138,  138,  138,
	   138,  138,  138,  138,  138,  138,  138,  138,  138,  138,
	   138,  138,  138,  138,  138,  138,    0,  138,  138,  138,
	   138,  138,  138,  138,  138,  138,    0,    0,    0,  138,
	     0,    0,  138,  138,  138,  138,  138,    0,    0,    0,
	     0,    0,  138,    0,  138,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,  140,  140,  140,
	   140,    0,    0,    0,    0,    0,  140,  140,    0,    0,
	     0,    0,    0,    0,    0,    0,  140,  140,  140,  140,
	   140,  140,  140,  140,  140,  140,    0,  140,  140,  140,
	   140,  140,  140,  140,  140,  140,  140,  140,  140,  140,
	   140,  140,  140,  140,  140,  140,    0,  140,  140,  140,
	   140,  140,  140,  140,  140,  140,    0,    0,    0,  140,
	     0,    0,  140,  140,  140,  140,  140,    0,    0,    0,
	     0,    0,  140,    0,  140,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,  142,  142,  142,
	   142,    0,    0,    0,    0,    0,  142,  142,    0,    0,
	     0,    0,    0,    0,    0,    0,  142,  142,  142,  142,
	   142,  142,  142,  142,  142,  142,  142,    0,  142,  142,
	   142,  142,  142,  142,  142,  142,  142,  142,  142,  142,
	   142,  142,  142,  142,  142,  142,    0,  142,  142,  142,
	   142,  142,  142,  142,  142,  142,    0,    0,    0,  142,
	     0,    0,  142,  142,  142,  142,  142,    0,    0,    0,
	     0,    0,  142,    0,  142,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,  144,  144,  144,
	   144,    0,    0,    0,    0,    0,  144,  144,    0,    0,
	     0,    0,    0,    0,    0,    0,  144,  144,  144,  144,
	   144,  144,  144,  144,  144,  144,  144,  144,    0,  144,
	   144,  144,  144,  144,  144,  144,  144,  144,  144,  144,
	   144,  144,  144,  144,  144,  144,    0,  144,  144,  144,
	   144,  144,  144,  144,  144,  144,    0,    0,    0,  144,
	     0,    0,  144,  144,  144,  144,  144,    0,    0,    0,
	     0,    0,  144,    0,  144,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,  146,  146,  146,
	   146,    0,    0,    0,    0,    0,  146,  146,    0,    0,
	     0,    0,    0,    0,    0,    0,  146,  146,  146,  146,
	   146,  146,  146,  146,  146,  146,  146,  146,  146,    0,
	   146,  146,  146,  146,  146,  146,  146,  146,  146,  146,
	   146,  146,  146,  146,  146,  146,    0,  146,  146,  146,
	   146,  146,  146,  146,  146,  146,    0,    0,    0,  146,
	     0,    0,  146,  146,  146,  146,  146,    0,    0,    0,
	     0,    0,  146,    0,  146,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,  150,  150,  150,
	   150,    0,    0,    0,    0,    0,  150,  150,    0,    0,
	     0,    0,    0,    0,    0,    0,  150,  150,  150,  150,
	   150,  150,  150,  150,  150,  150,  150,  150,  150,  150,
	     0,  149,  150,  150,  150,  150,  150,  150,  150,  150,
	   150,  150,  150,  150,  150,  150,    0,  150,  150,  150,
	   150,  150,  150,  150,  150,  150,    0,    0,    0,  151,
	     0,    0,  150,  150,  150,  150,  150,    0,    0,    0,
	     0,    0,  150,    0,  150,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,  153,  153,  153,
	   153,    0,    0,    0,    0,    0,  153,  153,    0,    0,
	     0,    0,    0,    0,    0,    0,  153,  153,  153,  153,
	   153,  153,  153,  153,  153,  153,  153,  153,  153,  153,
	   153,  153,    0,  154,  153,  153,  153,  153,  153,  153,
	   153,  153,  153,  153,  153,  153,    0,  153,  153,  153,
	   153,  153,  153,  153,  153,  153,    0,    0,    0,  153,
	     0,    0,  153,  153,  153,  153,  153,    0,    0,    0,
	     0,    0,  153,    0,  153,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,  160,  160,  160,
	   160,    0,    0,    0,    0,    0,  160,  160,    0,    0,
	     0,    0,    0,    0,    0,    0,  160,  160,  160,  160,
	   160,  160,  160,  160,  160,  160,  160,  160,  160,  160,
	   160,  160,  160,  160,    0,  160,  160,  160,  160,  160,
	   160,  160,  160,  160,  160,  160,    0,  160,  160,  160,
	   160,  160,  160,  160,  160,  160,    0,    0,    0,  160,
	     0,    0,  160,  160,  160,  160,  160,    0,    0,    0,
	     0,    0,  160,    0,  160,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,  165,  165,  165,
	   165,    0,    0,    0,    0,    0,  165,  165,    0,    0,
	     0,    0,    0,    0,    0,    0,  165,  165,  165,  165,
	   165,  165,  165,  165,  165,  165,  165,  165,  165,  165,
	   165,  165,  165,  165,  165,    0,  163,  163,  163,  163,
	   163,  163,  164,  165,  166,  167,    0,  169,  170,  168,
	   168,  168,  168,  168,  168,  168,    0,    0,    0,  165,
	     0,    0,  165,  165,  165,  165,  165,    0,    0,    0,
	     0,    0,  165,    0,  171,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,  172,  172,  172,
	   172,  173,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,  174,  175,
	   176,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,  182,  182,  182,
	   182,    0,    0,    0,    0,    0,  182,  182,    0,    0,
	     0,    0,    0,    0,    0,    0,  182,  182,  182,  182,
	   182,  182,  182,  182,  182,  182,  182,  182,  182,  182,
	   182,  182,  182,  182,  182,  182,  182,  182,  182,  182,
	   182,  182,  182,    0,  182,  182,    0,  182,  182,  182,
	   182,  182,  182,  182,  182,  182,    0,    0,    0,  182,
	     0,    0,  182,  182,  182,  182,  182,    0,    0,    0,
	     0,    0,  182,    0,  182,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	   190,  191,  192,  193,  194,  195,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,  203,  203,  203,
	   203,    0,    0,    0,    0,    0,  203,  203,    0,    0,
	     0,    0,    0,    0,    0,    0,  203,  203,  203,  203,
	   203,  203,  203,  203,  203,  203,  203,  203,  203,  203,
	   203,  203,  203,  203,  203,  203,  203,  203,  203,  203,
	   203,  203,  203,  203,  203,  203,    0,  203,  203,  203,
	   203,  203,  203,  203,  203,  203,    0,    0,    0,  203,
	     0,    0,  203,  203,  203,  203,  203,    0,    0,    0,
	     0,    0,  203,    0,  203,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,  204,  204,  204,
	   204,    0,    0,    0,    0,    0,  204,  204,    0,    0,
	     0,    0,    0,    0,    0,    0,  204,  204,  204,  204,
	   204,  204,  204,  204,  204,  204,  204,  204,  204,  204,
	   204,  204,  204,  204,  204,  204,  204,  204,  204,  204,
	   204,  204,  204,  204,  204,  204,    0,  204,  204,  204,
	   204,  204,  204,  204,  204,  204,    0,    0,    0,  204,
	     0,    0,  204,  204,  204,  204,  204,    0,    0,    0,
	     0,    0,  204,    0,  204,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,  207,  207,  207,
	   207,    0,    0,    0,    0,    0,  207,  207,    0,    0,
	     0,    0,    0,    0,    0,    0,  207,  207,  207,  207,
	   207,  207,  207,  207,  207,  207,  207,  207,  207,  207,
	   207,  207,  207,  207,  207,  207,  207,  207,  207,  207,
	   207,  207,  207,  207,  207,  207,    0,  207,  207,  207,
	   207,  207,  207,  207,  207,  207,    0,    0,    0,  207,
	     0,    0,  207,  207,  207,  207,  207,    0,    0,    0,
	     0,    0,  207,    0,  207,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,  229,  229,  229,
	   229,    0,    0,    0,    0,    0,  229,  229,    0,    0,
	     0,    0,    0,    0,    0,    0,  229,  229,  229,  229,
	   229,  229,  229,  229,  229,  229,  229,  229,  229,  229,
	   229,  229,  229,  229,  229,  229,  229,  229,  229,  229,
	   229,  229,  229,  229,  229,  229,    0,  229,  229,  229,
	   229,  229,  229,  229,  229,  229,    0,    0,    0,  229,
	     0,    0,  229,  229,  229,  229,  229,    0,    0,    0,
	     0,    0,  229,    0,  229,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,  230,  230,  230,
	   230,    0,    0,    0,    0,    0,  230,  230,    0,    0,
	     0,    0,    0,    0,    0,    0,  230,  230,  230,  230,
	   230,  230,  230,  230,  230,  230,  230,  230,  230,  230,
	   230,  230,  230,  230,  230,  230,  230,  230,  230,  230,
	   230,  230,  230,  230,  230,  230,    0,  230,  230,  230,
	   230,  230,  230,  230,  230,  230,    0,    0,    0,  230,
	     0,    0,  230,  230,  230,  230,  230,    0,    0,    0,
	     0,    0,  230,    0,  230,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,  231,  231,  231,
	   231,    0,    0,    0,    0,    0,  231,  231,    0,    0,
	     0,    0,    0,    0,    0,    0,  231,  231,  231,  231,
	   231,  231,  231,  231,  231,  231,  231,  231,  231,  231,
	   231,  231,  231,  231,  231,  231,  231,  231,  231,  231,
	   231,  231,  231,  231,  231,  231,    0,  231,  231,  231,
	   231,  231,  231,  231,  231,  231,    0,    0,    0,  231,
	     0,    0,  231,  231,  231,  231,  231,    0,    0,    0,
	     0,    0,  231,    0,  231,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,  236,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
	{    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
	     0,    0,    0,    0,    0,},
};
#line 637 "fs.brg"


/** Access the Core representation of any node */
#define C(p) ((p)->core)

/** Maximum number of pattern leaves */
#define MAX_KIDS 10

/** mnemonic XQuery Core constructors */
#include "core_mnemonic.h"

/** global query information */
static PFquery_t *PFquery;

/**
 * The current context items $fs:dot, $fs:position, and $fs:last
 * (see W3C XQuery Formal Semantics 3.1.2)
 */
static PFvar_t *fs_dot;
static PFvar_t *fs_position;
static PFvar_t *fs_last;

/**
 * The top of this stack holds a reference to the current function
 * (whose argument list is currently compiled).
 */
static PFarray_t *funs;
/**
 * The top of this stack holds a representation of the 
 * argument list to be passed to the currently compiled function application:
 *
 *               arg (e1, arg (e2, ..., arg (en, nil)...))
 */
static PFarray_t *args;
/**
 * The top of this stack holds the function argument number
 * that is currently compiled.
 */
static PFarray_t *arg_num;

/**
 * Collect function signatures here. The will be available as the
 * left child of the root node of the resulting Core tree.
 */
static PFcnode_t *fun_sigs;

/**
 * Collect variable declarations from the query prolog in this
 * root/hole pair. (see rule for VariableDeclaration)
 */
static struct {
    PFcnode_t *root;
    PFcnode_t *hole;
} var_decls;

/* element/attribute constructor and content */
static PFty_t (*elem_attr) (PFqname_t, PFty_t);
static PFty_t any;

static PFty_t *parameter_types (PFpnode_t *paramlist);

static PFcnode_t *function_conversion (const PFcnode_t *e, PFty_t expected);
static void comparison_worker (PFpnode_t *p, PFfun_t *fun);
static void val_comparison_worker (PFpnode_t *p, PFfun_t *fun);
static void calculation_worker (PFpnode_t *p, PFfun_t *fun);
static void two_arg_fun_worker (PFpnode_t *p, PFfun_t *fun);

static void reduce (PFpnode_t * p, int goalnt);

/**
 * Watch out for function declarations in the abstract syntax tree
 * and call reduce() for them.
 *
 * Compilation of expressions into Core requires the full signatures
 * of all functions to be readily available.  Function checking
 * (functions.c) has only set the number of function arguments, but
 * not their types (and neither the function return type).
 *
 * reduce() contains logic to derive #PFty_t types from abstract
 * syntax tree fragments.  We thus call reduce() for the function
 * signature whenever we find a function declaration.  Otherwise
 * we follow all those children that might contain further function
 * declarations.  (Note that we only handle the function signature
 * here, not its body.)
 */
static void
process_fun_sigs (PFpnode_t *p, int goalnt)
{
    int           rule;           /* rule number that matches for this node */
    short        *nts;            /* target non-terminals for the leaf nodes of
                                     the current rule */
    PFpnode_t    *kids[MAX_KIDS]; /* leaf nodes of this rule */

    /* guard against too dep recursion */
    PFrecursion_fence();

    /* determine rule that matches for this non-terminal */
    rule = PFfs_rule (STATE_LABEL (p), goalnt);

    /* PFinfo (OOPS_NOTICE, "match for rule %i", rule); */
    assert (rule);

    /* initialize the kids[] vector */
    for (unsigned short i = 0; i < MAX_KIDS; i++)
        kids[i] = NULL;

    /*
     * prepare recursive traversal: get information on leaf nodes of
     * this rule
     */
    nts = PFfs_nts[rule];
    PFfs_kids (p, rule, kids);

    switch (rule) {

        /* Query:                  lib_mod (ModuleDecl, Prolog) */
        case 2:
        /* LibraryModule:          lib_mod (ModuleDecl, Prolog) */
        case 4:

            /* we are heading for function signatures in the prolog only */
            process_fun_sigs (kids[1], nts[1]);
            break;

        /* MainModule:             main_mod (Prolog, QueryBody) */
        case 3:

            /* we are heading for function signatures in the prolog only */
            process_fun_sigs (kids[0], nts[0]);
            break;

        /* DeclsImports:           decl_imps (fun (fun_sig (OptParamList_,
                                                            SequenceType),
                                                   Expr),
                                              DeclsImports) */
        case 9:
            /* fall through */

        /* DeclsImports:           decls_imps (fun (fun_sig (OptParamList_,
                                                             SequenceType),
                                                    external),
                                               DeclsImports) */
        case 10:
            /*
             * Parse tree situation for function declarations:
             *
             *                   decls_imps
             *                  /          \
             *               fun        (other decls)
             *              /   \
             *        fun_sig    BODY
             *        /     \
             *  ParamList  ReturnType
             */

            /* debugging only
            PFinfo (OOPS_NOTICE, "processing `%s'",
                    PFqname_str (L(p)->sem.fun->qname));
            */

            /*
             * This is the situation that process_fun_sigs() actually
             * looks out for.  We call reduce() for function signatures.
             * reduce() contains code to build up the Core equivalent
             * of the function signature and---most importantly---code
             * to derive #PFty_t types from parse tree fragments.  We
             * use the latter to fill in the missing parts into the
             * #PFfun_t struct.
             */
            reduce (kids[0], nts[0]);
            reduce (kids[1], nts[1]);

            L(p)->sem.fun->sig_count = 1;
            L(p)->sem.fun->sigs =  
                (PFfun_sig_t *) PFmalloc (sizeof (PFfun_sig_t));
	    
            L(p)->sem.fun->sigs[0].par_ty = parameter_types (LLL(p));
            L(p)->sem.fun->sigs[0].ret_ty = C(LLR(p))->sem.type;

            /* Look out for more function declarations */
            process_fun_sigs (kids[3], nts[3]);

            break;


        /* Query:                  MainModule */
        case 1:

        /* Prolog:                 DeclsImports */
        case 6:

        /* DeclsImports:           decl_imps (DeclImport, DeclsImports) */
        case 7:

        /* DeclImport:             LibraryModule */
        case 24:

            /* These are the interesting paths to follow */
            for (unsigned short i = 0; nts[i]; i++)
                process_fun_sigs (kids[i], nts[i]);
            break;

        default:
            break;
    }
}

/**
 * Reducer function. This is the heart of this source file. It
 * contains all the action code for the above burg patterns.
 */
static void reduce_rest (PFpnode_t * p, int rule, short * nts, PFpnode_t *kids[MAX_KIDS]);
static void
reduce (PFpnode_t * p, int goalnt)
{
    int           rule;           /* rule number that matches for this node */
    short        *nts;            /* target non-terminals for the leaf nodes of
                                     the current rule */
    PFpnode_t    *kids[MAX_KIDS]; /* leaf nodes of this rule */
    bool          topdown;        /* is this a top-down rule? */

    /* guard against too dep recursion */
    PFrecursion_fence();

    /* determine rule that matches for this non-terminal */
    rule = PFfs_rule (STATE_LABEL (p), goalnt);

    /* PFinfo (OOPS_NOTICE, "match for rule %i", rule); */
    assert (rule);

    /* initialize the kids[] vector */
    for (unsigned short i = 0; i < MAX_KIDS; i++)
        kids[i] = NULL;

    /*
     * prepare recursive traversal: get information on leaf nodes of
     * this rule
     */
    nts = PFfs_nts[rule];
    PFfs_kids (p, rule, kids);

    /* PFinfo (OOPS_NOTICE, "in rule %u", rule); */

    switch (rule) {
        /* DeclsImports:           decl_imps (fun (fun_sig (OptParamList_,
                                                            SequenceType),
                                                   Expr),
                                              DeclsImports) */
        case 9:
        /* DeclsImports:           decls_imps (fun (fun_sig (OptParamList_,
                                                             SequenceType),
                                                    external),
                                               DeclsImports) */
        case 10:
        /* NodeTest:               node_ty (ElemOrAttrType) */
        case 92:
        /* NodeTest:               node_ty (node_ty (ElemOrAttrType)) */
        case 94:
        /* LocationPath:           locpath (PathExpr, PathExpr) */
        case 156:
        /* LocationPath:           pred (LocationPath, OptExpr) */
        case 158:
        /* FilterExpr:             pred (PathExpr, Expr) */
        case 161:
        /* FunctionCall:           apply (nil) */
        case 184:
        /* FunctionCall:           apply (FuncArgList) */
        case 185:
        /* FuncArgList:            args (Expr, FuncArgList) */
        case 187:
            topdown = true;
            break;

        default:
            topdown = false;
    }

    /*
     * Recursively invoke compilation.  This means bottom-up compilation.
     */
    if (!topdown)
        for (unsigned short i = 0; nts[i]; i++)
            reduce (kids[i], nts[i]);

    reduce_rest(p, rule, nts, kids);
}

static void
reduce_rest (PFpnode_t * p, int rule, short * nts, PFpnode_t *kids[MAX_KIDS])
{
    PFcnode_t    *c;              /* temporary helper variable */

    /* guard against too dep recursion */
    PFrecursion_fence();

    /*
    PFinfo (OOPS_NOTICE, "processing rule %i", rule);
    */

    switch (rule) {

        /* Query:                  MainModule */
        case 1:
            break;

        /* Query:                  lib_mod (ModuleDecl, Prolog) */
        case 2:
            /*
             * Catch the case where we got an XQuery module
             * directly as the input.
             *
             * Don't know what the best is to do here. For now,
             * we pick Peter's approach and just "simulate" a
             * query that returns just the empty sequence.
             */

            /*
             * We may have collected variable bindings from the
             * prolog. Put them on top of the query.
             */
            if (var_decls.root) {
                var_decls.hole->child[1] = nil ();
                c = flwr (var_decls.root, empty ());
            }
            else
                c = empty ();

            /*
             * Make a `main' node the root of the resulting tree.
             * Left child lists all the function signatures. Right
             * child is the query itself.
             */
            C(p) = PFcore_main (fun_sigs, c);

            break;

        /* MainModule:             main_mod (Prolog, QueryBody) */
        case 3:
            /*
             * We may have collected variable bindings from the
             * prolog. Put them on top of the query.
             */
            if (var_decls.root) {
                var_decls.hole->child[1] = nil ();
                c = flwr (var_decls.root, C(R(p)));
            }
            else
                c = C(R(p));

            /*
             * Make a `main' node the root of the resulting tree.
             * Left child lists all the function signatures. Right
             * child is the query itself.
             */
            C(p) = PFcore_main (fun_sigs, c);

            break;

        /* LibraryModule:          lib_mod (ModuleDecl, Prolog) */
        case 4:
            break;

        /* ModuleDecl:             mod_ns (lit_str) */
        case 5:
            break;

        /* Prolog:                 DeclsImports */
        case 6:
            break;

        /* DeclsImports:           decl_imps (DeclImport, DeclsImports) */
        case 7:
            break;

        /* DeclsImports:           nil */
        case 8:
            break;

        /* DeclsImports:           decl_imps (fun (fun_sig (OptParamList_,
                                                            SequenceType),
                                                   Expr),
                                              DeclsImports) */
        case 9:
        {   /* TOPDOWN */

            /*
             * User-defined functions:
             *
             * (1) Create a unique PFfun_t data structure, containing
             *     type information, number of arguments, and an argument
             *     list (PFvar_t pointers).
             *     This has already happened during function checking.
             *
             * (2) Set function signatures
             *     This has been done already during process_fun_sigs().
             * 
             * (3) Translate function body.
             *
             * (4) Store the Core equivalent of the function body in
             *     the PFfun_t data structure. Parameter variables will
             *     appear free in this Core expression.
             *     Step (3) is what we do here.
             *
             * [(5) Before executing the Core code for this function,
             *      the back-end (e.g., algebra translation) needs to
             *      fill the variable environment with bindings for all
             *      the function parameters.
             *      This has to be done in the back-end.]
             *
             * Parse tree situation for function declarations:
             *
             *                   decls_imps
             *                  /          \
             *               fun        (other decls)
             *              /   \
             *        fun_sig    BODY
             *        /     \
             *  ParamList  ReturnType
             */

            /*
             * No need to compile function signature.  This has been
             * done already in process_fun_sigs().
             */

            /* recurse */
            reduce (kids[2], nts[2]);
            reduce (kids[3], nts[3]);

            /*
             * Apply function conversion to the function result
             * (see W3C XQuery 3.1.5 or function_conversion()
             */

            C(LR(p)) = function_conversion (C(LR(p)),
                                            C(LLR(p))->sem.type);
            L(p)->sem.fun->core = C(LR(p));

            fun_sigs = fun_decls (fun_decl (L(p)->sem.fun, C(LLL(p)), C(LR(p))),
                                  fun_sigs);

        } break;


        /* DeclsImports:           decls_imps (fun (fun_sig (OptParamList_,
                                                             SequenceType),
                                                    external),
                                               DeclsImports) */
        case 10:
        {   /* TOPDOWN */

            /*
             * Parse tree situation for function declarations:
             *
             *                   decls_imps
             *                  /          \
             *               fun        (other decls)
             *              /   \
             *        fun_sig    external
             *        /     \
             *  ParamList  ReturnType
             */

            /*
             * We have already handled the function signature, so
             * nothing left to do for external functions.
             *
             * just recurse and handle other declarations
             */
            reduce (kids[2], nts[2]);

        } break;

        /* DeclImport:             DefaultNamespaceDecl */
        case 11:
            break;

        /* DeclImport:             XMLSpaceDecl */
        case 12:
            break;

        /* DeclImport:             DefaultCollationDecl */
        case 13:
            break;

        /* DeclImport:             BaseURIDecl */
        case 14:
            break;

        /* DeclImport:             ConstructionDecl */
        case 15:
            break;

        /* DeclImport:             OrderingModeDecl */
        case 16:
            break;

        /* DeclImport:             EmptyOrderingDecl */
        case 17:
            break;

        /* DeclImport:             InheritNamespacesDecl */
        case 18:
            break;

        /* DeclImport:             NamespaceDecl */
        case 19:
            break;

        /* DeclImport:             OptionDecl */
        case 20:
            break;

        /* DeclImport:             SchemaImport */
        case 21:
            break;

        /* DeclImport:             ModuleImport */
        case 22:
            break;

        /* DeclImport:             VariableDeclaration */
        case 23:
            break;

        /* DeclImport:             LibraryModule */
        case 24:
            break;

        /* DefaultNamespaceDecl:   ens_decl (lit_str) */
        case 25:
            break;

        /* DefaultNamespaceDecl:   fns_decl (lit_str) */
        case 26:
            break;

        /* XMLSpaceDecl:           xmls_decl */
        case 27:
            break;

        /* DefaultCollationDecl:   coll_decl (lit_str) */
        case 28:
            PFinfo_loc (OOPS_WARN_NOTSUPPORTED, p->loc,
                        "default collation declaration will be ignored");
            break;

        /* BaseURIDecl:            base_uri (lit_str) */
        case 29:
            PFinfo_loc (OOPS_WARN_NOTSUPPORTED, p->loc,
                        "base-uri declaration will be ignored");
            break;

        /* ConstructionDecl:       constr_decl */
        case 30:
            PFinfo_loc (OOPS_WARN_NOTSUPPORTED, p->loc,
                        "construction mode declaration will be ignored");
            break;

        /* OptionDecl:             option (lit_str) */
        case 31:
            /*
             * Options have already been extracted and entered into
             * PFoptions.  No need to do anything special here.
             */
            break;

        /* OrderingModeDecl:       ordering_mode */
        case 32:
            PFquery->ordering = p->sem.tru;
            break;

        /* EmptyOrderingDecl:      def_order */
        case 33:
            PFquery->empty_order = p->sem.mode.empty==p_greatest?greatest:least;
            break;

        /* InheritNamespacesDecl:  inherit_ns */
        case 34:
            PFquery->inherit_ns = p->sem.tru;
            break;

        /* NamespaceDecl:          ns_decl (lit_str) */
        case 35:
            break;

        /* SchemaImport:           schm_imp (lit_str, OptImportLoc_) */
        case 36:
            break;

        /* ModuleImport:           mod_imp (lit_str, OptImportLoc_) */
        case 37:
            /*
             * module imports have already been handled by the parser
             */
            break;

        /* OptImportLoc_:          nil */
        case 38:
            break;

        /* OptImportLoc_:          schm_ats (lit_str, OptImportLoc_) */
        case 39:
            break;

        /* OptParamList_:          nil */
        case 40:
            C(p) = nil ();
            break;

        /* OptParamList_:          ParamList_ */
        case 41:
            break;

        /* ParamList_:             nil */
        case 42:
            C(p) = nil ();
            break;

        /* ParamList_:             params (Param, ParamList_) */
        case 43:
            C(p) = params (C(L(p)), C(R(p)));
            break;

        /* Param:                  param (SequenceType, Var) */
        case 44:
            C(p) = param (C(L(p)), C(R(p)));
            break;

        /* OptSequenceType_:       nil */
        case 45:
            break;

        /* OptSequenceType_:       SequenceType */
        case 46:
            break;

        /* VariableDeclaration:    var_decl (var_type (Var,
                                                       nil),
                                             OptExpr) */
        case 47:
            /*
             * We translate
             *
             *  declare variable $a := ea ;
             *  declare variable $b := eb ;
             *  e0
             *
             * as
             *
             *  let $a := ea return
             *    let $b := eb return
             *      e0  .
             *
             * So while traversing the Query Prolog (this is what we're
             * currently doing), we build up a chain of nested let clauses,
             * with the return part as a hole.
             */
            c = let (letbind (C(LL(p)), C(R(p))), nil ());

	    /* set global information */
	    LL(c)->sem.var->global = true;

            if (var_decls.root)
                var_decls.hole = var_decls.hole->child[1] = c;
            else
                var_decls.root = var_decls.hole = c;

            break;

        /* VariableDeclaration:    var_decl (var_type (Var,
                                                       SequenceType),
                                             OptExpr) */
        case 48:
        {
            /*
             * We translate
             *
             *  declare variable $a as t1 := ea ;
             *  declare variable $b as t2 := eb ;
             *  e0
             *
             * as
             *
             *  let $a as t1 := ea return
             *    let $b as t2 := eb return
             *      e0  .
             *
             * So while traversing the Query Prolog (this is what we're
             * currently doing), we build up a chain of nested let clauses,
             * with the return part as a hole.
             */
            PFvar_t *v1 = new_var (NULL);

            c = let (letbind (var (v1), C(R(p))),
                     let (letbind (C(LL(p)),
                                   proof (subty (var (v1), C(LR(p))),
                                          seqcast (C(LR(p)), var (v1)))),
                          nil ()));

	    /* set global information */
	    LL(R(c))->sem.var->global = true;

            if (var_decls.root)
                var_decls.hole->child[1] = c;
            else
                var_decls.root = c;

            var_decls.hole = c->child[1];

        } break;

        /* VariableDeclaration:    var_decl (var_type (Var,
                                                       OptSequenceType_),
                                             external) */
        case 49:
            /* FIXME: What to do here? */
            PFoops_loc (OOPS_NOTSUPPORTED, p->loc,
                        "external variables not yet implemented");
            break;

        /* QueryBody:              OptExpr */
        case 50:
            assert (C(p));
            break;

        /* OptExpr:                empty_seq */
        case 51:
            C(p) = empty ();
            break;

        /* OptExpr:                Expr */
        case 52:
            assert (C(p));
            break;

        /* Expr:                   exprseq (OptExpr, OptExpr) */
        case 53:
            C(p) = seq (C(L(p)), C(R(p)));
            break;

        /* Expr:                   ExprSingle */
        case 54:
            assert (C(p));
            break;

        /* ExprSingle:             FLWORExpr */
        case 55:
            break;

        /* ExprSingle:             QuantifiedExpr */
        case 56:
            break;

        /* ExprSingle:             TypeswitchExpr */
        case 57:
            break;

        /* ExprSingle:             IfExpr */
        case 58:
            break;

        /* ExprSingle:             OrExpr */
        case 59:
            break;

        /* FLWORExpr:              flwr (OptBindExpr,
                                         where (OptWhereClause_,
                                                ord_ret (nil, OptExpr))) */
        case 60:
            C(p) = flwr (C(L(p)),
                         where (C(RL(p)), C(RRR(p))));
            break;

        /* FLWORExpr:              flwr (OptBindExpr,
                                         where (OptWhereClause_,
                                                ord_ret (orderby (
                                                            OrderSpecList),
                                                    OptExpr))) */
        case 61:
            C(p) = flwr (C(L(p)),
                         where (C(RL(p)),
                             orderby (RRL(p)->sem.tru,    /* stable or not? */
                                      C(RRLL(p)),
                                      C(RRR(p)))));
            break;

        /* OptBindExpr:            nil */
        case 62:
            C(p) = nil ();
            break;

        /* OptBindExpr:            binds (let (var_type (Var,
                                                         nil),
                                               OptExpr),
                                          OptBindExpr) */
        case 63:
        {
            C(p) = let (letbind (C(LLL(p)), C(LR(p))), C(R(p)));

        } break;

        /* OptBindExpr:            binds (let (var_type (Var,
                                                         TypeDeclaration),
                                               OptExpr),
                                          OptBindExpr) */
        case 64:
        {
            PFvar_t *v = new_var (NULL);

            C(p) = let (letbind (var (v), C(LR(p))),
                        let (letbind (C(LLL(p)),
                                      proof (subty (var (v), C(LLR(p))),
                                             seqcast (C(LLR(p)), 
                                                      var (v)))),
                             C(R(p))));
                        
        } break;

        /* OptBindExpr:            binds (bind (vars (var_type (Var,
                                                                nil),
                                                      OptPosVar),
                                                OptExpr),
                                          OptBindExpr) */
        case 65:
            C(p) = for_ (forbind (forvars (C(LLLL(p)), C(LLR(p))),
                                  C(LR(p))),
                         C(R(p)));
            break;

        /* OptBindExpr:            binds (bind (vars (var_type (
                                                          Var,
                                                          TypeDeclaration),
                                                      OptPosVar),
                                                OptExpr),
                                          OptBindExpr) */
        case 66:
        {
            PFvar_t *v = new_var (NULL);

            C(p) = for_ (forbind (forvars (var (v), C(LLR(p))),
                                  C(LR(p))),
                         let (letbind (C(LLLL(p)),
                                       proof (subty (var (v),
                                                     C(LLLR(p))),
                                              seqcast (C(LLLR(p)),
                                                       var (v)))),
                              C(R(p))));
        } break;

        /* OptPosVar:              nil */
        case 68:
            C(p) = nil ();
            break;

        /* OptPosVar:              Var */
        case 69:
            break;

        /* TypeDeclaration:        SequenceType */
        case 70:
            break;

        /* OptWhereClause_:        nil */
        case 71:
            C(p) = true_ ();
            break;

        /* OptWhereClause_:        OptExpr */
        case 72:
            C(p) = ebv (C(p));
            break;

        /* OrderSpecList:          nil */
        case 73:
            C(p) = nil ();
            break;

        /* OrderSpecList:          orderspecs (OptExpr, OrderSpecList) */
        case 74:
            C(p) = orderspecs (p->sem.mode, 
                               fs_convert_op_by_type (fn_data (C(L(p))), 
                                                      PFty_xs_string ()),
                               C(R(p)));
            break;

        /* QuantifiedExpr:         some (binds (bind (vars (var_type (Var,
                                                                      nil),
                                                            nil),
                                                      OptExpr),
                                                nil),
                                         OptExpr) */
        case 75:
            C(p) = some (C(LLLLL(p)), C(LLR(p)), C(R(p)));
            break;

        /* QuantifiedExpr:         some (binds (bind (vars (var_type (Var,
                                                                      TypeDecl),
                                                            nil),
                                                      OptExpr),
                                                nil),
                                         OptExpr) */
        case 76:
        {
            PFvar_t *v1 = new_var (NULL);

            C(p) = some (var (v1), C(LLR(p)),
                         flwr (let (letbind (C(LLLLL(p)),
                                             proof (subty (var (v1),
                                                           C(LLLLR(p))),
                                                    seqcast (C(LLLLR(p)), 
                                                             var (v1)))),
                                    nil ()),
                               C(R(p))));

        } break;


        /* QuantifiedExpr:         every (binds (bind (vars (var_type (Var,
                                                                       nil),
                                                             nil),
                                                       OptExpr),
                                                 nil),
                                          OptExpr) */
        case 77:
        {
            /* every $a in e1 satisfies e2
             *  ==
             * not (some $a in e1 satisfies not (e2))
             *  ==
             * not (boolean (some $a in e1 satisfies not (boolean (e2))))
             *
             * (As not() uses boolean() to ensure that all items are
             *  present and we only want to support the more specific
             *  variant of not() we introduce additional calls 
             *  to boolean().)
             */
            PFfun_t *fn_not  = function (PFqname (PFns_fn, "not"));
            PFfun_t *fn_bool = function (PFqname (PFns_fn, "boolean"));

            C(p) = APPLY (fn_not, 
                          APPLY (fn_bool,
                                 some (
                                     C(LLLLL(p)),
                                     C(LLR(p)),
                                     APPLY (fn_not, 
                                            APPLY (fn_bool, C(R(p)))))));
        } break;

        /* QuantifiedExpr:         every (binds (bind (vars (var_type (Var,
                                                                       TypeDec),
                                                             nil),
                                                       OptExpr),
                                                 nil),
                                          OptExpr) */
        case 78:
        {
            /* every $a as t in e1 satisfies e2
             *  ==
             * not (some $a as t in e1 satisfies not (e2))
             *  ==
             * not (boolean (some $a as t in e1 satisfies not (boolean (e2))))
             *  ==
             * not (boolean (
             *   some $v in e1 satisfies
             *     let $a := proof ($v, t, seqcast (t, $v))
             *     return not (boolean (e2))))
             *
             * (As not() uses boolean() to ensure that all items are
             *  present and we only want to support the more specific
             *  variant of not() we introduce additional calls 
             *  to boolean().)
             */
            PFvar_t *v = new_var (NULL);
            PFfun_t *fn_not  = function (PFqname (PFns_fn, "not"));
            PFfun_t *fn_bool = function (PFqname (PFns_fn, "boolean"));

            C(p) = APPLY (fn_not,
                          APPLY (fn_bool, 
                                 some (
                                     var (v),
                                     C(LLR(p)),
                                     flwr (let (
                                               letbind (
                                                   C(LLLLL(p)),
                                                   proof (subty (var (v),
                                                                 C(LLLLR(p))),
                                                          seqcast (C(LLLLR(p)),
                                                                   var (v)))),
                                               nil ()),
                                           APPLY (fn_not,
                                                  APPLY (fn_bool,
                                                         C(R(p))))))));
        } break;

        /* TypeswitchExpr:         typeswitch (
                                       Expr,
                                       cases (case_ (var_type (OptCaseVar,
                                                               SequenceType),
                                                     OptExpr),
                                              cases (default_ (OptCaseVar,
                                                               OptExpr),
                                                     nil))) */
        case 79:
        {
            PFvar_t *v = new_var (NULL);

            C(p) = flwr (let (letbind (var (v), C(L(p))),
                              nil ()),
                         typeswitch (
                             var (v),
                             cases (
                                 case_ (
                                     C(RLLR(p)),
                                     flwr (let (letbind (C(RLLL(p)),
                                                         seqcast (C(RLLR(p)),
                                                                  var (v))),
                                                nil ()),
                                           C(RLR(p)))),
                                 default_ (
                                     flwr (let (letbind (C(RRLL(p)), var (v)),
                                                nil ()),
                                           C(RRLR(p)))))));
        } break;

        /* OptCaseVar:             nil */
        case 80:
            /*
             * Generate new variable in case no variable has been
             * specified for case/default branch of typeswitch.
             */
            C(p) = var (new_var ("tsw"));
            break;

        /* OptCaseVar:             Var */
        case 81:
            assert (C(p));
            break;

        /* SequenceType:           seq_ty (ItemType) */
        case 82:
            /* fall through */

        /* SingleType:             seq_ty (AtomType) */
        case 83:
            switch (p->sem.oci) {
                case p_one:
                    C(p) = C(L(p));
                    break;
                case p_zero_or_one:
                    C(p) = seqtype (PFty_opt (C(L(p))->sem.type));
                    break;
                case p_zero_or_more:
                    C(p) = seqtype (PFty_star (C(L(p))->sem.type));
                    break;
                case p_one_or_more:
                    C(p) = seqtype (PFty_plus (C(L(p))->sem.type));
                    break;
                default:
                    PFoops_loc (OOPS_FATAL, p->loc,
                                "illegal occurrence indicator (%d) in "
                                "sequence type", p->sem.oci);
                    break;
            } break;
            
        /* ItemType:               empty_ty */
        case 84:
            C(p) = seqtype (PFty_empty ());
            break;

        /* ItemType:               NodeTest */
        case 85:
            assert (C(p));
            break;

        /* ItemType:               item_ty (nil) */
        case 86:
            C(p) = seqtype (PFty_xs_anyItem ());
            break;

        /* ItemType:               atom_ty (nil) */
        case 87:
            /* is the referenced type a known schema type? */
            if (! PFty_schema (PFty_named (p->sem.qname)))
                PFoops_loc (OOPS_TYPENOTDEF, p->loc,
                        "``%s''",
                        PFqname_str (p->sem.qname));

            C(p) = seqtype (PFty_named (p->sem.qname));
            break;

        /* ItemType:               schm_elem */
        case 88:
            /* FIXME: This probably requires work in the type system */
            PFoops_loc (OOPS_NOTSUPPORTED, p->loc,
                        "schema-element types are not supported");
            break;

        /* ItemType:               schm_attr */
        case 89:
            /* FIXME: This probably requires work in the type system */
            PFoops_loc (OOPS_NOTSUPPORTED, p->loc,
                        "schema-attribute types are not supported");
            break;

        /* AtomType:               atom_ty (nil) */
        case 90:
            /* is the referenced type a known schema type? */
            if (! PFty_schema (PFty_named (p->sem.qname)))
                PFoops_loc (OOPS_TYPENOTDEF, p->loc,
                        "``%s''",
                        PFqname_str (p->sem.qname));

            C(p) = seqtype (PFty_named (p->sem.qname));
            break;

        /* NodeTest:               node_ty (nil) */
        case 91:
            switch (p->sem.kind) {
                case p_kind_node:
                    /* node */
                    C(p) = seqtype (PFty_xs_anyNode ());
                    break;
                case p_kind_comment:
                    /* comment */
                    C(p) = seqtype (PFty_comm ());
                    break;
                case p_kind_text:
                    /* text */
                    C(p) = seqtype (PFty_text ());
                    break;
                case p_kind_pi:
                    /* processing-instruction */
                    C(p) = seqtype (PFty_pi (NULL));
                    break;
                case p_kind_doc:
                    /* document */
                    C(p) = seqtype (PFty_doc (PFty_xs_anyType ()));
                    break;
                case p_kind_elem:
                    /* element */
                    C(p) = seqtype (PFty_elem (PFqname (PFns_wild, NULL),
                                PFty_xs_anyType ()));
                    break;
                case p_kind_attr:
                    /* attribute */
                    C(p) = seqtype (PFty_attr (PFqname (PFns_wild, NULL),
                                               PFty_star (PFty_atomic ())));
                    break;
                default:
                    PFoops_loc (OOPS_FATAL, p->loc,
                                "illegal node kind in node type");
            }
            break;

        /* NodeTest:               node_ty (ElemOrAttrType) */
        case 92:
            /* TOPDOWN */

            switch (p->sem.kind) {
                case p_kind_elem:
                    elem_attr = PFty_elem;
                    any = PFty_xs_anyType ();
                    break;
                case p_kind_attr:
                    elem_attr = PFty_attr;
                    any = PFty_star (PFty_atomic ());
                    break;
                default:
                    PFoops_loc (OOPS_FATAL, p->loc,
                                "illegal node kind in NodeTest");
            }

            reduce (kids[0], nts[0]);

            C(p) = C(L(p));
            break;

        /* NodeTest:               node_ty (node_ty (nil)) */
        case 93:
            assert (p->sem.kind == p_kind_doc);
            assert (L(p)->sem.kind == p_kind_elem);

            C(p) = seqtype (PFty_doc (PFty_xs_anyElement ()));
            break;

        /* NodeTest:               node_ty (node_ty (ElemOrAttrType)) */
        case 94:
            /* TOPDOWN */

            assert (p->sem.kind == p_kind_doc);
            assert (L(p)->sem.kind == p_kind_elem);

            elem_attr = PFty_elem;
            any = PFty_xs_anyType ();

            reduce (kids[0], nts[0]);

            C(p) = seqtype (PFty_doc (C(LL(p))->sem.type));
            break;

        /* NodeTest:               node_ty (lit_str) */
        case 95:
            /*
             * processing-instruction tests may have a string
             * argument (restricting the pi target)
             */
            assert (p->sem.kind == p_kind_pi);
            C(p) = seqtype (PFty_pi (L(p)->sem.str));
            break;

        /* ElemOrAttrType:         req_ty (nil, nil) */
        case 96:
            /* element/attribute: any name, any type */
            C(p) = seqtype (elem_attr (PFqname (PFns_wild, NULL), any));
            break;
     
        /* ElemOrAttrType:         req_ty (nil, SchemaType) */
        case 97:
            /* element/attribute: any name, given type */
            C(p) = seqtype (elem_attr (PFqname (PFns_wild, NULL),
                                       C(R(p))->sem.type));
            break;

        /* ElemOrAttrType:         req_ty (req_name, nil) */
        case 98:
            /* element/attribute: given name, any type */
            C(p) = seqtype (elem_attr (L(p)->sem.qname, any));
            break;


        /* ElemOrAttrType:         req_ty (req_name, SchemaType) */
        case 99:
            /* element/attribute: given name, given type */
            C(p) = seqtype (elem_attr (L(p)->sem.qname, C(R(p))->sem.type));
            break;

        /* SchemaType:             named_ty */
        case 100:
            /* is the referenced type a known schema type definition? */
            if (! PFty_schema (PFty_named (p->sem.qname)))
                PFoops_loc (OOPS_TYPENOTDEF, p->loc,
                            "``%s''",
                            PFqname_str (p->sem.qname));

            C(p) = seqtype (PFty_named (p->sem.qname));
            break;

        /* IfExpr:                 if_ (Expr,
                                        then_else (OptExpr, OptExpr)) */
        case 101:
            C(p) = if_ (ebv (C(L(p))), then_else (C(RL(p)), C(RR(p))));
            break;

        /* OrExpr:                 AndExpr */
        case 102:
            break;

        /* OrExpr:                 or (AndExpr, OrExpr) */
        case 103:
        {
            PFfun_t *op_or = function (PFqname (PFns_op, "or"));

            C(p) = APPLY (op_or, ebv (C(L(p))), ebv (C(R(p))));
        } break;

        /* AndExpr:                ComparisonExpr */
        case 104:
            break;

        /* AndExpr:                and (ComparisonExpr, AndExpr) */
        case 105:
        {
            PFfun_t *op_and = function (PFqname (PFns_op, "and"));

            C(p) = APPLY (op_and, ebv (C(L(p))), ebv (C(R(p))));
        } break;

        /* ComparisonExpr:         RangeExpr */
        case 106:
            break;

        /* ComparisonExpr:         eq (RangeExpr, RangeExpr) */
        case 107:
        {
            PFfun_t *op_eq = function (PFqname (PFns_op, "eq"));
            comparison_worker (p, op_eq);
            
        } break;

        /* ComparisonExpr:         ne (RangeExpr, RangeExpr) */
        case 108:
        {
            PFfun_t *op_ne = function (PFqname (PFns_op, "ne"));
            comparison_worker (p, op_ne);

        } break;

        /* ComparisonExpr:         lt (RangeExpr, RangeExpr) */
        case 109:
        {
            PFfun_t *op_lt = function (PFqname (PFns_op, "lt"));
            comparison_worker (p, op_lt);

        } break;

        /* ComparisonExpr:         le (RangeExpr, RangeExpr) */
        case 110:
        {
            PFfun_t *op_le = function (PFqname (PFns_op, "le"));
            comparison_worker (p, op_le);

        } break;

        /* ComparisonExpr:         gt (RangeExpr, RangeExpr) */
        case 111:
        {
            PFfun_t *op_gt = function (PFqname (PFns_op, "gt"));
            comparison_worker (p, op_gt);

        } break;

        /* ComparisonExpr:         ge (RangeExpr, RangeExpr) */
        case 112:
        {
            PFfun_t *op_ge = function (PFqname (PFns_op, "ge"));
            comparison_worker (p, op_ge);

        } break;

        /* ComparisonExpr:         val_eq (RangeExpr, RangeExpr) */
        case 113:
        {
            PFfun_t *op_eq = function (PFqname (PFns_op, "eq"));
            val_comparison_worker (p, op_eq);

        } break;

        /* ComparisonExpr:         val_ne (RangeExpr, RangeExpr) */
        case 114:
        {
            PFfun_t *op_ne = function (PFqname (PFns_op, "ne"));
            val_comparison_worker (p, op_ne);

        } break;

        /* ComparisonExpr:         val_lt (RangeExpr, RangeExpr) */
        case 115:
        {
            PFfun_t *op_lt = function (PFqname (PFns_op, "lt"));
            val_comparison_worker (p, op_lt);

        } break;

        /* ComparisonExpr:         val_le (RangeExpr, RangeExpr) */
        case 116:
        {
            PFfun_t *op_le = function (PFqname (PFns_op, "le"));
            val_comparison_worker (p, op_le);

        } break;

        /* ComparisonExpr:         val_gt (RangeExpr, RangeExpr) */
        case 117:
        {
            PFfun_t *op_gt = function (PFqname (PFns_op, "gt"));
            val_comparison_worker (p, op_gt);

        } break;

        /* ComparisonExpr:         val_ge (RangeExpr, RangeExpr) */
        case 118:
        {
            PFfun_t *op_ge = function (PFqname (PFns_op, "ge"));
            val_comparison_worker (p, op_ge);

        } break;

        /* ComparisonExpr:         is (RangeExpr, RangeExpr) */
        case 119:
        {
            PFfun_t *op_is = function (PFqname (PFns_op, "is-same-node"));
            two_arg_fun_worker (p, op_is);

        } break;

        /* ComparisonExpr:         ltlt (RangeExpr, RangeExpr) */
        case 121:
        {
            PFfun_t *op_ltlt = function (PFqname (PFns_op, "node-before"));
            two_arg_fun_worker (p, op_ltlt);

        } break;

        /* ComparisonExpr:         gtgt (RangeExpr, RangeExpr) */
        case 122:
        {
            PFfun_t *op_gtgt = function (PFqname (PFns_op, "node-after"));
            two_arg_fun_worker (p, op_gtgt);

        } break;

        /* RangeExpr:              AdditiveExpr */
        case 123:
            break;

        /* RangeExpr:              range (RangeExpr, RangeExpr) */
        case 124:
        {
            PFfun_t *op_to = function (PFqname (PFns_op, "to"));
            two_arg_fun_worker (p, op_to);

        } break;

        /* AdditiveExpr:           MultiplicativeExpr */
        case 125:
            break;

        /* AdditiveExpr:           plus (AdditiveExpr, MultiplicativeExpr) */
        case 126:
        {
            PFfun_t *op_plus = function (PFqname (PFns_op, "plus"));
            calculation_worker (p, op_plus);

        } break;

        /* AdditiveExpr:           minus (AdditiveExpr, MultiplicativeExpr) */
        case 127:
        {
            PFfun_t *op_minus = function (PFqname (PFns_op, "minus"));
            calculation_worker (p, op_minus);

        } break;

        /* MultiplicativeExpr:     UnionExpr */
        case 128:
            break;

        /* MultiplicativeExpr:     mult (UnionExpr, MultiplicativeExpr) */
        case 129:
        {
            PFfun_t *op_times = function (PFqname (PFns_op, "times"));
            calculation_worker (p, op_times);

        } break;

        /* MultiplicativeExpr:     div (UnionExpr, MultiplicativeExpr) */
        case 130:
        {
            PFfun_t *op_div = function (PFqname (PFns_op, "div"));
            calculation_worker (p, op_div);

        } break;

        /* MultiplicativeExpr:     idiv (UnionExpr, MultiplicativeExpr) */
        case 131:
        {
            PFfun_t *op_idiv = function (PFqname (PFns_op, "idiv"));

            C(p) =  APPLY (op_idiv, 
                           fs_convert_op_by_type (fn_data (C(L(p))),
                                                  PFty_xs_integer ()),
                           fs_convert_op_by_type (fn_data (C(R(p))),
                                                  PFty_xs_integer ()));
        } break;

        /* MultiplicativeExpr:     mod (UnionExpr, MultiplicativeExpr) */
        case 132:
        {
            PFfun_t *op_mod = function (PFqname (PFns_op, "mod"));
            calculation_worker (p, op_mod);

        } break;

        /* UnionExpr:              IntersectExceptExpr */
        case 133:
            break;

        /* UnionExpr:              union_ (IntersectExceptExpr, UnionExpr) */
        case 134:
        {
            PFfun_t *op_union = function (PFqname (PFns_op, "union"));
            two_arg_fun_worker (p, op_union);

        } break;

        /* IntersectExceptExpr:    InstanceofExpr */
        case 135:
            break;

        /* IntersectExceptExpr:    intersect (InstanceofExpr,
                                              IntersectExceptExpr) */
        case 136:
        {
            PFfun_t *op_intersect = function (PFqname (PFns_op, "intersect"));
            two_arg_fun_worker (p, op_intersect);

        } break;

        /* IntersectExceptExpr:    except (InstanceofExpr,
                                           IntersectExceptExpr) */
        case 137:
        {
            PFfun_t *op_except = function (PFqname (PFns_op, "except"));
            two_arg_fun_worker (p, op_except);

        } break;

        /* InstanceofExpr:         TreatExpr */
        case 138:
            break;

        /* InstanceofExpr:         instof (TreatExpr, SequenceType) */
        case 139:
            C(p) = typeswitch (
                       C(L(p)),
                       cases (
                           case_ (C(R(p)), true_ ()),
                           default_ (false_ ())));
            break;

        /* TreatExpr:              CastableExpr */
        case 140:
            break;

        /* TreatExpr:              treat (CastableExpr, SequenceType) */
        case 141:
        {
            PFvar_t *v1 = new_var (NULL);
            PFfun_t *fn_error = function (PFqname (PFns_fn, "error"));

            C(p) = flwr (let (letbind (var (v1), C(L(p))), nil ()),
                         typeswitch (
                             var (v1),
                             cases (case_ (C(R(p)), var (v1)),
                                    default_ (APPLY (fn_error,
                                                     str("err:XP0050"))))));
        } break;

        /* CastableExpr:           CastExpr */
        case 142:
            break;

        /* CastableExpr:           castable (CastExpr, SingleType) */
        case 143:
            PFoops_loc (OOPS_NOTSUPPORTED, p->loc,
                        "`castable' keyword is not supported");
            break;

        /* CastExpr:               UnaryExpr */
        case 144:
            break;

        /* CastExpr:               cast (UnaryExpr, SingleType) */
        case 145:
            C(p) = cast (C(R(p)), fn_data (C(L(p))));
            break;

        /* UnaryExpr:              ValueExpr */
        case 146:
            break;

        /* UnaryExpr:              uminus (ValueExpr) */
        case 147:
        {
            PFfun_t *op_minus = function (PFqname (PFns_op, "minus"));

            C(p) = APPLY (op_minus,
                          num (0), 
                          fs_convert_op_by_type (
                              fn_data (C(L(p))),
                              PFty_xs_double ()));
        } break;

        /* UnaryExpr:              uplus (ValueExpr) */
        case 148:
        {
            PFfun_t *op_plus = function (PFqname (PFns_op, "plus"));

            C(p) = APPLY (op_plus,
                          num (0), 
                          fs_convert_op_by_type (
                              fn_data (C(L(p))),
                              PFty_xs_double ()));
        } break;

        /* ValueExpr:              ValidateExpr */
        case 149:
            break;

        /* ValueExpr:              PathExpr */
        case 150:
            break;

        /* ValueExpr:              ExtensionExpr */
        case 151:
            break;

        /* ValidateExpr:           validate (Expr) */
        case 152:
            PFoops_loc (OOPS_NOTSUPPORTED, p->loc, "validation not supported");
            break;

        /* PathExpr:               PathExpr */
        case 153:
            assert (C(p));
            break;

        /* PathExpr:               LocationPath */
        case 154:
        {   /* guarantee document order (even for reverse axis steps) */
            assert (C(p));

            PFfun_t *ddo_as   = function (
                                    PFqname (
                                        PFns_pf, 
                                        "distinct-doc-order-or-atomic-sequence"
                                            ));
                
            C(p) = APPLY (ddo_as, C(p));
        } break;

        /* LocationPath:           root_ */
        case 155:
        {
            PFfun_t *_root = function (PFqname (PFns_fn, "root"));

            if (fs_dot)
                C(p) = APPLY (_root, var (fs_dot));
            else
                PFoops_loc (OOPS_NOCONTEXT, p->loc, "``.'' is unbound");

#if 0
    FIXME:
        The code below is actually closer to the standard that
        demands to not only translate the '/' operator into a
        call to fn:root(.), but into

          fn:root(.) treat as document .

        However, milprint_summer cannot evaluate the typeswitch
        involved, making '/' fail in any case.  We do not want to
        break too much, but rather be slightly off the standard
        for the time being.
                                                 Jens (26.10.05)

            PFvar_t *v1 = new_var (NULL);
            PFfun_t *_root = function (PFqname (PFns_fn, "root"));
            PFfun_t *fn_error = function (PFqname (PFns_fn, "error"));

            if (fs_dot)
                C(p) = flwr (let (letbind (var (v1), APPLY (_root, 
                                                            var (fs_dot))),
                                  nil ()),
                             typeswitch (
                                 var (v1),
                                 cases (
                                     case_ (
                                         seqtype (PFty_doc (
                                             PFty_plus (PFty_xs_anyNode ()))),
                                         var (v1)),
                                     default_ (
                                         APPLY (fn_error,
                                                str("err:XP0050"))))));
            else
                PFoops_loc (OOPS_NOCONTEXT, p->loc, "``.'' is unbound");
#endif
        } break;

        /* LocationPath:           locpath (PathExpr, PathExpr) */
        case 156:
        {
            /* TOPDOWN */

            /*
             * A path expression E1/E2 is to be evaluated as follows:
             * (see XQuery Oct 2004 WD, Section 3.2)
             *
             * (1) Expression E1 is evaluated, and if the result is not
             *     a (possibly empty) sequence of nodes, a type error
             *     is raised.
             *
             * (2) Each node resulting from the evaluation of E1 then
             *     serves in turn to provide an inner focus for an
             *     evaluation of E2.
             *
             * (3) The sequences resulting from all the evaluations of
             *     E2 are combined as follows:
             *
             *     -- If every evaluation of E2 returns a (possibly empty)
             *        sequence of nodes, these sequences are combined,
             *        and duplicate nodes are eliminated. If ordering mode
             *        is ordered, the resulting node sequence is returned
             *        in document order.
             *     -- If every evaluation of E2 returns a (possibly empty)
             *        sequence of atomic values, these sequences are
             *        concatenated, in order, and returned.
             *     -- Otherwise (that is, if the multiple evaluations of
             *        E2 return at least one node and at least one atomic
             *        value), a type error is raised.
             *
             * In XQuery Core:
             *
             * let $v := E1 return
             *   typeswitch ($v)
             *     case node* return
             *       fs:distinct-doc-order-or-atomic-sequence (
             *           let $fs:last := fn:count ($v) returni
             *               for $fs:dot at $fs:position in $v return E2)
             *     default error (err:XP0019)
             */

            PFvar_t *dot;
            PFvar_t *position;
            PFvar_t *last;
            PFvar_t *v = new_var (NULL);

            PFfun_t *fn_error = function (PFqname (PFns_fn, "error"));
            PFfun_t *count    = function (PFqname (PFns_fn, "count"));
            PFfun_t *ddo_as   = function (
                                    PFqname (
                                        PFns_pf, 
                                        "distinct-doc-order-or-atomic-sequence"
                                            ));

            /* compile context first */
            reduce (kids[1], nts[1]);

            /*
             * save context items $fs:dot, $fs:position, $fs:last
             * and establish new context item
             */
            dot = fs_dot;
            fs_dot = new_var ("dot");
            position = fs_position;
            fs_position = new_var ("pos");
            last = fs_last;
            fs_last = new_var ("lst");

            /* compile step in this new context */
            reduce (kids[0], nts[0]);

            C(p) =
                flwr (let (letbind (var (v), C(R(p))), nil ()),
                  typeswitch (
                    var (v),
                    cases (
                      case_ (
                        seqtype (PFty_star (PFty_xs_anyNode ())),
                        APPLY (ddo_as,
                               flwr (
                                 let (
                                   letbind (var (fs_last), 
                                            APPLY (count, var (v))),
                                   nil ()),
                                 flwr (for_ (
                                     forbind (forvars (var (fs_dot),
                                                       var (fs_position)),
                                              var (v)),
                                     nil ()),
                                 C(L(p)))))),
                      default_ (
                        APPLY (fn_error,
                               str ("err:XP0019: Argument to location step "
                                    "must be a node sequence."))))));

            /* restore context items */
            fs_dot = dot;
            fs_position = position;
            fs_last = last;

        } break;

        /* LocationPath:           locpath (step (NodeTest), PathExpr) */
        case 157:
        {   /*
             * For regular location steps, we don't actually need the
             * complicated stuff in the above rule.
             *
             * let $v := PathExpr return
             *   typeswitch ($v)
             *     case node* return
             *         fs:distinct-doc-order ($v/LocationStep)
             *     default error (err:XP0019)
             */
            PFvar_t *v = new_var (NULL);
            PFfun_t *ddo = function (PFqname (PFns_pf, "distinct-doc-order"));
            PFfun_t *fn_error = function (PFqname (PFns_fn, "error"));

            assert (C(LL(p)));
            
            C(L(p)) = step (L(p)->sem.axis, C(LL(p)));
            C(p) = flwr (let (letbind (var (v), C(R(p))),
                              nil ()),
                         typeswitch (
                           var (v),
                           cases (
                             case_ (seqtype (PFty_star (PFty_xs_anyNode ())),
                               APPLY (ddo, locsteps (C(L(p)),var (v)))),
                             default_ (
                               APPLY (fn_error,
                                      str ("err:XP0019: Argument to location "
                                           "step must be a node sequence."))))));

            /* Return the result of the path step in reverse document order
               if the axis is a reverse axis. This enables predicates to lookup
               the values at the correct position. */
            if (L(p)->sem.axis == p_ancestor ||
                L(p)->sem.axis == p_ancestor_or_self ||
                L(p)->sem.axis == p_parent ||
                L(p)->sem.axis == p_preceding ||
                L(p)->sem.axis == p_preceding_sibling) {
                
                PFvar_t *v = new_var (NULL);
                PFfun_t *fn_rev = function (PFqname (PFns_fn, "reverse"));
                
                C(p) = flwr (let (letbind (var (v), C(p)),
                                  nil ()),
                             APPLY (fn_rev, var (v)));
            }
        } break;

        /* LocationPath:           pred (LocationPath, OptExpr) */
        case 158:
        /* FilterExpr:             pred (PathExpr, Expr) */
        case 161:
        {   /* TOPDOWN */

            /*
             * Predicate evaluation:
             *
             * e1[e2]
             *  ==
             * let $v1 := e1 return
             *   let $fs:last := fn:count ($v1) return
             *     for $fs:dot at $fs:position in $v1 return
             *       let $v2 := e2 return
             *         let $v3 :=
             *           typeswitch ($v2)
             *             case $v8 as numeric return
             *               typeswitch ($v2)
             *                 case $v4 as xs:integer
             *                   return fn:eq ($v4, $fs:position)
             *                 case $v5 as xs:decimal
             *                   return fn:eq ($v5, $fs:position)
             *                 case $v6 as xs:double
             *                   return fn:eq ($v6, $fs:position)
             *                 default return   (: should actually not happen :)
             *                   let $v7 := $v8 cast as xs:double return
             *                     fn:eq ($v7, $fs:position)
             *             default return ebv ($v2)
             *           return
             *             if ($v3) then $fs:dot else ()
             *
             * Note:
             *   We explicitly line up the variants of type `numeric' here.
             *   (a simple `case $v4 as numeric return fn:eq ($v4, $fs:pos)'
             *   instead of the inner typeswitch would loose important
             *   static type information: $v4 would get static type numeric,
             *   and we would loose the (possibly more specific) type we
             *   inferred so far.)
             */
            PFvar_t *v1 = new_var (NULL);
            PFvar_t *v2 = new_var (NULL);
            PFvar_t *v3 = new_var (NULL);
            PFvar_t *v4 = new_var (NULL);
            PFvar_t *v5 = new_var (NULL);
            PFvar_t *v6 = new_var (NULL);
            PFvar_t *v7 = new_var (NULL);
            PFvar_t *v8 = new_var (NULL);
            PFvar_t *dot;
            PFvar_t *position;
            PFvar_t *last;
            PFfun_t *count = function (PFqname (PFns_fn, "count"));
            PFfun_t *op_eq = function (PFqname (PFns_op, "eq"));

            /* compile e1 */
            reduce (kids[0], nts[0]);

            /*
             * save context items $fs:dot, $fs:position, $fs:last
             * and establish new context item
             */
            dot = fs_dot;
            fs_dot = new_var ("dot");
            position = fs_position;
            fs_position = new_var ("pos");
            last = fs_last;
            fs_last = new_var ("lst");

            /* compile predicate expression in this new context */
            reduce (kids[1], nts[1]);

            C(p) = flwr (
                     let (letbind (var (v1), C(L(p))), nil ()),
                     flwr (
                       let (letbind (var (fs_last), 
                                     APPLY (count, var (v1))), nil ()),
                       flwr (
                         for_ (forbind (forvars (var (fs_dot),
                                                 var (fs_position)),
                                        var (v1)), nil ()),
                         flwr (
                           let (letbind (var (v2), C(R(p))), nil ()),
                           flwr (
                             let (
                               letbind (
                                 var (v3),
                                 typeswitch (
                                   var (v2),
                                   cases (
                                     case_ (
                                       seqtype (PFty_numeric ()),
                                       flwr (
                                         let (letbind (var (v8),
                                                       seqcast (
                                                         seqtype (
                                                           PFty_numeric ()),
                                                       var (v2))), nil ()),
                                         typeswitch (
                                           var (v2),
                                           cases (
                                             case_ (
                                               seqtype (PFty_xs_integer ()),
                                               flwr (
                                                 let (
                                                   letbind (
                                                     var (v4),
                                                     seqcast (
                                                       seqtype (
                                                         PFty_xs_integer()),
                                                       var (v2))), nil ()),
                                                 APPLY (op_eq, var (v4),
                                                        var (fs_position)))),
                                             default_ (
                                         /* <------ */
                                         typeswitch (
                                           var (v2),
                                           cases (
                                             case_ (
                                               seqtype (PFty_xs_decimal ()),
                                               flwr (
                                                 let (
                                                   letbind (
                                                     var (v5),
                                                     seqcast (
                                                       seqtype (
                                                         PFty_xs_decimal ()),
                                                       var (v2))), nil ()),
                                                 APPLY (op_eq, var (v5),
                                                        var (fs_position)))),
                                             default_ (
                                         /* <------ */
                                         typeswitch (
                                           var (v2),
                                           cases (
                                             case_ (
                                               seqtype (PFty_xs_double ()),
                                               flwr (
                                                 let (
                                                   letbind (
                                                     var (v6),
                                                     seqcast (
                                                       seqtype (
                                                         PFty_xs_double()),
                                                       var (v2))), nil ()),
                                                 APPLY (op_eq, var (v6),
                                                        var (fs_position)))),
                                             default_ (
                                         /* <------ */
                                         flwr (
                                           let (
                                             letbind (
                                               var (v7),
                                               cast (seqtype (PFty_xs_double()),
                                                     var (v8))), nil ()),
                                           APPLY (op_eq,
                                                  var (v7),
                                                  var (fs_position)))
                                         /* ------> */
                                                      )))
                                         /* ------> */
                                                      )))
                                         /* ------> */
                                                      )))
                                       )
                                     ),
                                   default_ (ebv (var (v2)))))), nil ()),
                             if_ (var (v3),
                              then_else (var (fs_dot), empty ())))))));

            /* restore context items */
            fs_dot = dot;
            fs_position = position;
            fs_last = last;
        } break;

        /* FilterExpr:             PrimaryExpr */
        case 160:
            assert (C(p));
            break;

        /* PrimaryExpr:            Literal */
        case 163:
            assert (C(p));
            break;

        /* PrimaryExpr:            Var */
        case 164:
            assert (C(p));
            break;

        /* PrimaryExpr:            ParenthesizedExpr */
        case 165:
            assert (C(p));
            break;

        /* PrimaryExpr:            ContextItemExpr */
        case 166:
            assert (C(p));
            break;

        /* PrimaryExpr:            FunctionCall */
        case 167:
            assert (C(p));
            break;

        /* PrimaryExpr:            Constructor */
        case 168:
            assert (C(p));
            break;

        /* PrimaryExpr:            OrderedExpr */
        case 169:
            assert (C(p));
            break;

        /* PrimaryExpr:            UnorderedExpr */
        case 170:
            assert (C(p));
            break;
            
        /* PrimaryExpr:            XRPCCall */
        case 171:
            assert (C(p));
            break;

        /* Literal:                NumericLiteral */
        case 172:
            break;

        /* Literal:                StringLiteral */
        case 173:
            break;

        /* NumericLiteral:         IntegerLiteral */
        case 174:
            break;

        /* NumericLiteral:         DecimalLiteral */
        case 175:
            break;

        /* NumericLiteral:         DoubleLiteral */
        case 176:
            break;

        /* IntegerLiteral:         lit_int */
        case 177:
            C(p) = num (p->sem.num);
            break;

        /* DecimalLiteral:         lit_dec */
        case 178:
            C(p) = dec (p->sem.dec);
            break;

        /* DoubleLiteral:          lit_dbl */
        case 179:
            C(p) = dbl (p->sem.dbl);
            break;

        /* StringLiteral:          lit_str */
        case 180:
            C(p) = str (p->sem.str);
            break;

        /* Var:                    var */
        case 181:
            C(p) = var (p->sem.var);
            break;

        /* ParenthesizedExpr:      Expr */
        case 182:
            assert (C(p));
            break;

        /* ContextItemExpr:        dot */
        case 183:
            if (fs_dot)
                C(p) = var (fs_dot);
            else
                PFoops_loc (OOPS_NOCONTEXT, p->loc, "``.'' is unbound");
            break;

        /* FunctionCall:           apply (nil) */
        case 184:
            /*
             * There are some built-in functions with special
             * implicit semantics.
             */

            /*
             * Functions `fn:string', `fn:name', `fn:local-name', 
             * `fn:namespace-uri', `fn:root' with no arguments
             * implicitly get the current
             * context node `.' as their argument.
             */
            if ( (!PFqname_eq (p->sem.fun->qname, PFqname (PFns_fn, "string")))
                || (!PFqname_eq (p->sem.fun->qname, PFqname (PFns_fn, "name")))
                || (!PFqname_eq (p->sem.fun->qname,
                                 PFqname (PFns_fn, "local-name")))
                || (!PFqname_eq (p->sem.fun->qname,
                                 PFqname (PFns_fn, "namespace-uri")))
                || (!PFqname_eq (p->sem.fun->qname, PFqname (PFns_fn, "root")))
               ) {
                if (fs_dot)
                    C(p) = APPLY (p->sem.fun, var (fs_dot));
                else
                    PFoops_loc (OOPS_NOCONTEXT, p->loc, "``.'' is unbound");

                break;
            }

            /*
             * Functions `fn:string-length', `fn:normalize-space',
             * `fn:number'
             * implicitly get the string-value of the current context
             * node `.' as their argument if called without any
             * arguments.
             */
            if ((!PFqname_eq (p->sem.fun->qname,
                              PFqname (PFns_fn, "string-length")))
                || (!PFqname_eq (p->sem.fun->qname,
                                 PFqname (PFns_fn, "normalize-space")))
                || (!PFqname_eq (p->sem.fun->qname,
                                 PFqname (PFns_fn, "number")))
               ) {
                if (fs_dot) {
                    C(p) = APPLY (p->sem.fun, 
                                  APPLY (function (PFqname (PFns_fn, "string")),
                                         var (fs_dot)));
                }
                else
                    PFoops_loc (OOPS_NOCONTEXT, p->loc, "``.'' is unbound");

                break;
            }

            /*
             * `fn:position' is available as variable $fs:position.
             */
            if (!PFqname_eq (p->sem.fun->qname, PFqname (PFns_fn, "position"))){
                if (fs_position)
                    C(p) = var (fs_position);
                else
                    PFoops_loc (OOPS_NOCONTEXT, p->loc,
                                "position information not available");
                break;
            }

            /*
             * `fn:last' is available as variable $fs:last.
             */
            if (!PFqname_eq (p->sem.fun->qname, PFqname (PFns_fn, "last"))) {
                if (fs_last)
                    C(p) = var (fs_last);
                else
                    PFoops_loc (OOPS_NOCONTEXT, p->loc,
                                "position information not available");
                break;
            }

            /*
             * In all other cases just call the respective function
             */
            C(p) = apply (p->sem.fun, nil ());
            break;

        /* FunctionCall:           apply (FuncArgList) */
        case 185:
            /* TOPDOWN */

            assert (PFqname_eq (p->sem.fun->qname,
                                PFqname (PFns_fn, "position")));

            *(PFfun_t **) PFarray_add (funs) = p->sem.fun;
            *(PFcnode_t **) PFarray_add (args) = nil ();

            *(short *) PFarray_add (arg_num) = 0;

            /*
             * Functions `fn:id', `fn:idref' with one argument
             * implicitly get the current context node `.' as
             * their second/last argument.
             */
            if (p->sem.fun->arity == 1 &&
                ((!PFqname_eq (p->sem.fun->qname,
                               PFqname (PFns_fn, "id")))
                || (!PFqname_eq (p->sem.fun->qname,
                                 PFqname (PFns_fn, "idref")))
                )
               ) {
                if (fs_dot)
                    *(PFcnode_t **) PFarray_top (args)
                        = arg (var (fs_dot), *(PFcnode_t **) PFarray_top (args));
                else
                    PFoops_loc (OOPS_NOCONTEXT, p->loc, "``.'' is unbound");
            }

            reduce (kids[0], nts[0]);

            PFarray_del (funs);
            PFarray_del (args);
            PFarray_del (arg_num);

            /*
             * Rewrite type cast functions into "cast as"
             */
            if (p->sem.fun->arity == 1) {
                PFty_t tpe;
                if (PFty_atomic_lookup (PFqname_uri(p->sem.fun->qname),
                                        PFqname_loc(p->sem.fun->qname),
                                        &tpe)) {
                    C(p) = cast ( seqtype (tpe), fn_data( L(L(C(L(p))))));
                    break;
                }
            }

            C(p) = C(L(p));
            break;

        /* FuncArgList:            nil */
        case 186:
        {
            PFfun_t *fun = *(PFfun_t **) PFarray_top (funs);

            C(p) = apply (fun, *(PFcnode_t **) PFarray_top (args));

        } break;

        /* FuncArgList:            args (Expr, FuncArgList) */
        case 187:
        {   /* TOPDOWN */
            
            /*
             * Apply ``function conversion rules'' (W3C XQuery 3.1.5).
             * See also helper function function_conversion().
             */
            PFcnode_t    *dummy = nil ();
            PFfun_t      *fun;
            PFty_t        expected;
            int  param_num;
            PFfun_t      *implicit = NULL;
            PFcnode_t    *converted = NULL;

            /* translate function argument */
            reduce (kids[0], nts[0]);

            /* build up argument list -- fill in a dummy placeholder
               which will be replaced at the end by the real content */
            *(PFcnode_t **) PFarray_top (args)
                = arg (dummy, *(PFcnode_t **) PFarray_top (args));

            /* translate other arguments */
            reduce (kids[1], nts[1]);

            /* count arguments (parse tree is in reverse order
               - therefore increase argument after other arguments) */
            param_num = (*(short*) PFarray_top (arg_num))++;


            /* What function are we actually talking about? */
            fun = *(PFfun_t **) PFarray_top (funs);

            /* no dynamic overloading here */
            assert (fun->sig_count == 1);

            /*
             * There's exactly one function that allows an arbitrary
             * number of arguments: fn:concat(). Looking at the
             * parameter types here doesn't really make sense, we
             * use atomic? instead.
             */
            if (PFqname_eq (fun->qname,
                            PFqname (PFns_fn, "concat")) == 0) {
                expected = PFty_opt (PFty_atomic ());
            } else if (param_num == -1) {
                /* this is the magic leading URL string parameter of an
                 * XRPC function call, its expected type is string. */
                expected = PFty_opt (PFty_xs_string ());
            } else {
                /* look up expected argument type */
                expected = fun->sigs[0].par_ty[param_num];
            }

            /*
             * Apply function conversion rules (XQuery 3.1.5.)
             * to this function argument, but leave the *literal*
             * empty sequence alone.
             *
             * (Account for specific issues around the empty sequence
             * here.  In particular, function conversion would try to
             * iterate over the empty sequence, which resulted in an
             * undefined type for the iteration variable.)
             */
            if (C(L(p))->kind == c_empty)
                converted = C(L(p));
            else
                converted = function_conversion (C(L(p)), expected);

            /*
             * There's implicit semantics for some functions.
             */
            
            /* Implicitly call fn:round() and cast the result to integer
               for the second and third argument of fn:subsequence(). */
            if (! PFqname_eq (fun->qname, PFqname (PFns_fn, "subsequence")) &&
                param_num > 0) {
                converted = APPLY (function (PFqname (PFns_fn, "round")),
                                   converted);
		converted = cast (seqtype (PFty_xs_integer ()),
                                  converted);
            }
            
            /* Implicitly call fn:boolean() before fn:not(). */
            if (! PFqname_eq (fun->qname, PFqname (PFns_fn, "not")))
                implicit = function (PFqname (PFns_fn, "boolean"));

            /* Implicitly call fn:string() before fn:concat(). */
            if (! PFqname_eq (fun->qname, PFqname (PFns_fn, "concat")))
                implicit = function (PFqname (PFns_fn, "string"));

            /* replace the dummy placeholder in the list of function
               arguments by the real input */
            if (implicit)
                *dummy = *(APPLY (implicit, converted));
            else
                *dummy = *(converted); 

            /* propagate the complete function call */
            C(p) = C(R(p));
        } break;

        /* OrderedExpr:            ordered (Expr) */
        case 188:
            /* FIXME: Better treatment of `ordered' information?
             *         - Make this rule TOPDOWN
             *         - Keep a state variable during compilation
             */
            C(p) = ordered (C(L(p)));
            break;

        /* UnorderedExpr:          unordered (Expr) */
        case 189:
            C(p) = unordered (C(L(p)));
            break;

        /* Constructor:            ElementConstructor */
        case 190:
            break;

        /* Constructor:            AttributeConstructor */
        case 191:
            break;

        /* Constructor:            TextConstructor */
        case 192:
            break;

        /* Constructor:            CommentConstructor */
        case 193:
            break;

        /* Constructor:            PIConstructor */
        case 194:
            break;

        /* Constructor:            DocumentConstructor */
        case 195:
            break;

        /* ElementConstructor:     elem (TagName, OptContentExpr) */
        case 196:
        {
            PFfun_t *pf_matn
                = function (PFqname (PFns_pf, "merge-adjacent-text-nodes"));

            C(p) = constr_elem (C(L(p)), APPLY (pf_matn, C(R(p))));

        } break;

        /* AttributeConstructor:   attr (TagName, AttributeValue) */
        case 197:
            /*
             * The translation rules rooted at AttributeValue compile
             * the attribute value into a single string.
             */
            C(p) = constr_attr (C(L(p)), C(R(p)));
            break;

        /* TextConstructor:        text (Expr) */
        case 198:
        {
            /*
             * FIXME:
             *   Since June 2005 the W3C specs describe different variants
             *   for fs:item-sequence-to-untypedAtomic() for each node
             *   kind. fs:item-sequence-to-untypedAtomic-text() handles
             *   the empty sequence differently. For now, we leave everything
             *   as it is and wait until the W3C guys know what they want.
             */
            PFfun_t *is2uA =
	      function (PFqname (PFns_pf, "item-sequence-to-untypedAtomic"));

	    /* 
	     * Text node construction:
	     *
	     * text { e }
	     *  ==
	     * text { pf:item-sequence-to-untypedAtomic (fn:data (e))
                      cast as xs:string? }
	     */
	    C(p) = constr (p->kind, 
                           cast (seqtype (PFty_opt (PFty_xs_string ())),
                                 APPLY (is2uA, fn_data (C(L(p))))));
        } break;

        /* CommentConstructor:     comment (Expr) */
        case 199:
        {
            PFfun_t *is2uA =
                function (PFqname (PFns_pf, "item-sequence-to-untypedAtomic"));

            C(p) = constr (p->kind, APPLY (is2uA, C(L(p))));

        } break;

        /* PIConstructor:          pi (OptExpr, OptContentExpr) */
        case 200:
        {
            PFfun_t *is2uA =
                function (PFqname (PFns_pf, "item-sequence-to-untypedAtomic"));

            C(p) = constr_pi (C(L(p)), APPLY (is2uA, C(R(p))));
        } break;

        /* DocumentConstructor:    doc (Expr) */
        case 201:
        {
            PFfun_t *is2ns =
                function (PFqname (PFns_pf, "item-sequence-to-node-sequence"));

            C(p) = constr (p->kind, APPLY (is2ns, C(L(p))));
        } break;

        /* TagName:                tag */
        case 202:
            C(p) = constr_tag (p->sem.qname);
            break;

        /* TagName:                Expr */
        case 203:
        {
            /*
             * FIXME:
             *   The specs are a bit odd in this case.  The conversion
             *   of a string argument to a computed element constructor
             *   is not formally described.
             *
             *   fn:resolve-QName is not actually a solution:
             *
             *    (1) fn:resolve-QName() requires a second argument that
             *        describes the "in-scope namespaces" to resolve the
             *        prefix in the string argument.  Unfortunately we
             *        cannot provide one here.
             *
             *    (2) "in-scope namespaces" are not actually what we need
             *        here anyway.  Tag names for computed element
             *        constructors are based on "statically-known
             *        namespaces".
             *
             *   Hence, we introduce our own fn:resolve-QName (xs:string?)
             *   here and let the back-end try to give its best.  In
             *   milprint_summer.c, fn:resolve-QName will simply be the
             *   identity, as element construction will fiddle with strings
             *   there anyway.
             */
            PFfun_t *resolve_qname
                = function (PFqname (PFns_fn, "resolve-QName"));

            C(p) = APPLY (resolve_qname,
                          fs_convert_op_by_type (fn_data (C(p)),
                                                 PFty_xs_string ()));
        } break;

        /* OptContentExpr:         OptExpr */
        case 204:
        {
            PFfun_t *is2ns =
                function (PFqname (PFns_pf, "item-sequence-to-node-sequence"));

            C(p) = APPLY (is2ns, C(p));

        } break;

        /* OptContentExpr:         empty_seq */
        case 205:
            C(p) = empty ();
            break;

        /* OptContentExpr:         contseq (OptContentExpr, OptContentExpr) */
        case 206:
            C(p) = seq (C(L(p)), C(R(p)));
            break;

        /* AttributeValue:         Expr */
        case 207:
        {
            /*
             * Convert enclosed expression in attribute values as described in
             * [XQ, Section 3.7.1.1]:
             *
             *  a. Apply atomization (fn:data) to the enclosed expression.
             *  b. Cast each atomic value in the resulting sequence to a string
             *  c. Merge all the resulting strings with a space character
             *     inbetween.
             */
            PFvar_t *v = new_var (NULL);
            PFfun_t *string = function (PFqname (PFns_fn, "string"));
            PFfun_t *string_join = function (PFqname (PFns_fn, "string-join"));

            C(p) = APPLY (string_join,
                          flwr (for_ (forbind (forvars (var (v),
                                                        nil ()),
                                               C(p)),
                                      nil()),
                                APPLY (string, var (v))),
                          str (" "));
        } break;

        /* AttributeValue:         empty_seq */
        case 208:
            /*
             * An empty attribute value shall result in an empty string.
             */
            C(p) = str ("");
            break;

        /* AttributeValue:         contseq (AttributeValue, AttributeValue) */
        case 209:
        {
            /*
             * Both arguments have type xs:string after they had been
             * converted in the rule AttrEnclosedExpr. All that is left
             * to do is to concatenate them.
             *
             * fn:concat ([[ e1 ]], [[ e2 ]])
             */
            PFfun_t *concat = function (PFqname (PFns_fn, "concat"));

            C(p) = APPLY (concat, C(L(p)), C(R(p)));
        } break;

        /* ExtensionExpr:          ext_expr (Pragmas, nil) */
        case 210:
            /*
             * If an extension expression has an empty body we shall
             * raise a static error (W3C XQuery 3.14) if we don't
             * understand the pragma.
             */
            PFoops (OOPS_NOTSUPPORTED,
                    "err:XQST0079: pragma(s) not understood and pragma "
                    "expression body is absent");
            break;

        /* ExtensionExpr:          ext_expr (Pragmas, Expr) */
        case 211:
            /* We only understand caching pragmas */
            if (C(L(p))) {
                PFfun_t *cache = function (PFqname (PFns_pf, "query-cache"));
                C(p) = APPLY (cache, C(L(p)), C(R(p)));
            }
            /* Just ignore the pragma.  We don't understand it anyway. */
            else
                C(p) = C(R(p));
            break;

        /* Pragmas:                nil */
        case 212:
            /* We don't understand pragmas anyway */
            C(p) = NULL;
            break;

        /* Pragmas:                pragmas (pragma, Pragmas) */
        case 213:
            if (PFqname_eq (L(p)->sem.pragma.qn.qname,
                            PFqname (PFns_lib, "cache")) == 0)
                C(p) = str (PFstrUtils_trim_spaces (
                                PFstrdup (L(p)->sem.pragma.content)));
            /* transform session pragma into global session options */
            else if (PFqname_eq (L(p)->sem.pragma.qn.qname,
                                 PFqname (PFns_lib, "session")) == 0 ||
                     PFqname_eq (L(p)->sem.pragma.qn.qname,
                                 PFqname (PFns_lib, "session-use")) == 0) {
                char *content = PFstrUtils_trim_spaces (
                                    PFstrdup (L(p)->sem.pragma.content));
                PFenv_bind (PFoptions,
                            PFqname (PFns_lib, "session-id"),
                            PFstrUtils_dupUntil (content, ':'));
                PFenv_bind (PFoptions,
                            PFqname (PFns_lib, "session-timeout"),
                            PFstrUtils_skipUntilSubString (content, ":"));

                /* mark as ready-only caching session */
                if (PFqname_eq (L(p)->sem.pragma.qn.qname,
                                PFqname (PFns_lib, "session-use")) == 0)
                    PFenv_bind (PFoptions,
                                PFqname (PFns_lib, "session-nocache"),
                                PFstrdup ("true"));
            }
            else
                /* We don't understand pragmas anyway */
                C(p) = C(R(p));
            break;

        /* DeclImport:             RevalidationDecl */
        case 214:
            break;

        /* RevalidationDecl:       revalid */
        case 215:
            break;

        /* ExprSingle:             InsertExpr */
        case 216:
            assert (C(p));
            break;

        /* ExprSingle:             DeleteExpr */
        case 217:
            assert (C(p));
            break;

        /* ExprSingle:             RenameExpr */
        case 218:
            assert (C(p));
            break;

        /* ExprSingle:             ReplaceExpr */
        case 219:
            assert (C(p));
            break;

        /* ExprSingle:             TransformExpr */
        case 220:
            assert (C(p));
            break;

        /* InsertExpr:             insert (SourceExpr, TargetExpr) */
        case 221:
        {
            PFvar_t *target = new_var (NULL);
            PFvar_t *v1     = new_var (NULL);
            PFvar_t *v2     = new_var (NULL);
            PFvar_t *v3     = new_var (NULL);
            PFvar_t *v4     = new_var (NULL);
            PFvar_t *ins    = new_var (NULL);

            PFfun_t *exactly_one
                = function (PFqname (PFns_fn, "exactly-one"));
            PFfun_t *is2ns
                = function (PFqname (PFns_pf, "item-sequence-to-node-sequence"));
            PFfun_t *matn
                = function (PFqname (PFns_pf, "merge-adjacent-text-nodes"));
            PFfun_t *insert = NULL;

            switch (p->sem.insert) {

                case p_first_into:
                    if (!insert)
                        insert
                            = function (PFqname (PFns_upd,"insertIntoAsFirst"));
                    /* fall through */

                case p_last_into:
                case p_into:
                    /*
                     * With "into", the position is implementation defined.
                     * we use it as a synonym for "into as last"
                     */
                    if (!insert)
                        insert
                            = function (PFqname (PFns_upd, "insertIntoAsLast"));
                    /* fall through */

                case p_after:
                    if (!insert)
                        insert
                            = function (PFqname (PFns_upd, "insertAfter"));
                    /* fall through */

                case p_before:
                    if (!insert)
                        insert
                            = function (PFqname (PFns_upd, "insertBefore"));

                    /*
                     *  1. Compile <SourceExpr> as if it were the content
                     *     of an element constructor.
                     *
                     *  2. If the insertion expression contains a document
                     *     node, replace it in the insertion sequence by
                     *     its children.
                     *
                     *  3. Invoke the respective update operation.
                     *
                     * NOTE: We do not guarantee here that (in case of
                     *   "before" or "after") that parent property of the
                     *   target expression is not empty.
                     *
                     * let $target := [[ <TargetExpr> ]]
                     * let $v1 := [[ <SourceExpr> ]]
                     * let $v2 := pf:item-sequence-to-node-sequence ($v1)
                     * let $v3 := pf:merge-adjacent-text-nodes ($v2)
                     * let $ins := for $v4 in $v3 return
                     *               typeswitch ($v4)
                     *                 case document-node() return $v4/node()
                     *                 default return $v4
                     * return
                     *   upd:insertXXX ($target, $ins)
                     */
                    C(p)
                        = flwr (
                            let (letbind (var (target), APPLY(exactly_one, C(R(p)))),
                            let (letbind (var (v1), C(L(p))),
                            let (letbind (var (v2), APPLY (is2ns, var (v1))),
                            let (letbind (var (v3), APPLY (matn, var (v2))),
                            let (
                                letbind (
                                    var (ins),
                                    flwr (
                                        for_ (
                                            forbind (
                                                forvars (var (v4), nil ()),
                                                var (v3)),
                                            nil ()),
                                        typeswitch (
                                            var (v4),
                                            cases (
                                                case_ (
                                                    seqtype (
                                                        PFty_doc (
                                                            PFty_xs_anyType())),
                                                    locsteps (
                                                        step (
                                                            p_child,
                                                            seqtype (
                                                                PFty_node())),
                                                        var (v4))),
                                                default_ (var (v4)))))),
                                nil ()))))),
                            APPLY (insert, var (target), var (ins)));
                    break;
            }

        } break;

        /* DeleteExpr:             delete (SourceExpr) */
        case 222:
        {   /*
             * [[ do delete <TargetExpr> ]]
             *  ==
             * let $v1 := [[ <TargetExpr> ]] return
             *   for $v2 in $v1 return
             *     upd:delete ($v2)
             */
            PFvar_t *v1 = new_var (NULL);
            PFvar_t *v2 = new_var (NULL);
            PFfun_t *delete = function (PFqname (PFns_upd, "delete"));

            C(p) = flwr (let (letbind (var (v1), C(L(p))), nil ()),
                         flwr (for_ (forbind (forvars (var (v2), nil ()),
                                              var (v1)),
                                     nil ()),
                               APPLY (delete, var (v2))));
        } break;


        /* ReplaceExpr:            replace (TargetExpr, ExprSingle) */
        case 223:
            if (p->sem.tru) {  
                /*
                 * The semantics of the expression
                 *
                 *   do replace value of <TargetExpr> with <ExprSingle>
                 *
                 * is described in the XQuery Update Facility:
                 *
                 * 1. <ExprSingle> is evaluated as if it were the content
                 *    expression of a text node constructor.
                 *
                 * 2. The <TargetExpr> must evaluate to a single node.
                 *
                 * 3. If <TargetExpr> evaluates to an element node,
                 *    generate the update primitive upd:relaceElementContent()
                 *
                 * 4. If <TargetExpr> evaluates to an attribute, text, or
                 *    processing-instruction node, replace their content
                 *    with the string obtained in Step 1.
                 *
                 * 5. If <TargetExpr> evaluates to a document node, raise
                 *    an error.
                 *
                 * [[ do replace value of <TargetExpr> with <ExprSingle> ]]
                 *  ==
                 * let $expr := [[ <ExprSingle> ]]
                 * let $string := fn:data ($expr)
                 * let $step1 := pf:item-sequence-to-untypedAtomic($string)
                 * let $target := [[ <TargetExpr> ]] return
                 *   typeswitch ($target)
                 *     case $e as xs:anyElement return
                 *       let $textnode := text { $step1 cast as xs:string? }
                 *       return upd:replaceElementContent ($e, $textnode)
                 *     case $a as xs:anyAttribute return
                 *       upd:replaceValue ($a, $step1)
                 *     case $t as text() return
                 *       upd:replaceValue ($t, $step1)
                 *     case $p as processing-instruction() return
                 *       upd:replaceValue ($p, $step1)
                 *     case $c as comment() return
                 *       upd:replaceValue ($c, $step1)
                 *     default return
                 *       error ("illegal expression in `do replace' clause")
                 */
                PFvar_t *expr     = new_var (NULL);
                PFvar_t *string   = new_var (NULL);
                PFvar_t *step1    = new_var (NULL);
                PFvar_t *target   = new_var (NULL);
                PFvar_t *e        = new_var (NULL);
                PFvar_t *textnode = new_var (NULL);
                PFvar_t *a        = new_var (NULL);
                PFvar_t *t        = new_var (NULL);
                PFvar_t *pi       = new_var (NULL);
                PFvar_t *c        = new_var (NULL);

                PFfun_t *exactly_one
                    = function (PFqname (PFns_fn, "exactly-one"));
                PFfun_t *is2uA
                    = function (PFqname (PFns_pf,
                                         "item-sequence-to-untypedAtomic"));
                PFfun_t *replCont
                    = function (PFqname (PFns_upd, "replaceElementContent"));
                PFfun_t *replace
                    = function (PFqname (PFns_upd, "replaceValue"));
                PFfun_t *error   = function (PFqname (PFns_fn, "error"));

                C(p)
                  = flwr (
                        let (letbind (var (expr), C(R(p))),
                            let (letbind (var (string), fn_data (var (expr))),
                                let (letbind (var (step1),
                                              APPLY (is2uA, var (string))),
                                    let (letbind (var (target),
                                                  APPLY(exactly_one, C(L(p)))),
                                         nil())))),
                        typeswitch (
                            var (target),
                            cases (
                                case_ (
                                    seqtype (PFty_xs_anyElement ()),
                                    flwr (
                                        let (letbind (var (e),
                                                seqcast (
                                                    seqtype (
                                                        PFty_xs_anyElement ()),
                                                    var (target))),
                                        let (letbind (var (textnode),
                                                constr (p_text,
                                                    cast (seqtype (
                                                        PFty_opt (
                                                            PFty_xs_string ())),
                                                        var (step1)))),
                                        nil ())),
                                        APPLY (replCont, var (e),
                                               var (textnode)))),
                            default_ (typeswitch (var (target), cases (
                                case_ (
                                    seqtype (PFty_xs_anyAttribute ()),
                                    flwr (
                                        let (letbind (var (a),
                                            seqcast (
                                                seqtype (
                                                    PFty_xs_anyAttribute ()),
                                                var (target))), nil ()),
                                        APPLY (replace, var (a), var (step1)))),
                            default_ (typeswitch (var (target), cases (
                                case_ (
                                    seqtype (PFty_text ()),
                                    flwr (
                                        let (letbind (var (t),
                                                      seqcast (
                                                          seqtype (
                                                              PFty_text ()),
                                                          var (target))),
                                             nil ()),
                                        APPLY (replace, var (t), var (step1)))),
                            default_ (typeswitch (var (target), cases (
                                case_ (
                                    seqtype (PFty_pi (NULL)),
                                    flwr (
                                        let (letbind (var (pi),
                                                      seqcast (
                                                          seqtype (
                                                              PFty_pi (NULL)),
                                                          var (target))),
                                             nil ()),
                                        APPLY (replace,
                                               var (pi), var (step1)))),
                            default_ (typeswitch (var (target), cases (
                                case_ (
                                    seqtype (PFty_comm ()),
                                    flwr (
                                        let (letbind (var (c),
                                                      seqcast (
                                                          seqtype (
                                                              PFty_comm ()),
                                                          var (target))),
                                             nil ()),
                                        APPLY (replace, var (c), var (step1)))),
                            default_ (
                                APPLY (error,
                                      str ("illegal expression in `do replace' "
                                           "clause"))
                                ))))))))))))))));

            }
            else {
                /*
                 * Evaluate the clause
                 *
                 *   do replace <TargetExpr> with <ExprSingle>
                 *
                 * as follows:
                 *
                 *  1. evaluate <ExprSingle> as if it were enclosed in
                 *     an element constructor, yielding the "replacement
                 *     sequence".
                 *
                 *  2. If the replacement sequence contains a document
                 *     node, replace it by its children.
                 *
                 * (3. Raise an error if the replacement sequence contains
                 *     an attribute node that does follow a node that is
                 *     not an attribute.)
                 *
                 *  4. <TargetExpr> must not have an empty parent property.
                 *
                 *  5. If <TargetExpr> is an element, text, comment, or
                 *     processing instruction node, generate the update
                 *     operation
                 *
                 *       upd:replaceNode ($target, $clist)
                 *
                 *  6. If <TargetExpr> is an attribute node, generate the
                 *     operations
                 *
                 *       upd:insertIntoAsLast ($parent, $clist)
                 *       upd:delete ($target).
                 *
                 * [[ do replace <TargetExpr> with <ExprSingle> ]]
                 *  ==
                 * let $target := [[ <TargetExpr> ]]
                 * let $v1 := [[ <ExprSingle> ]]
                 * let $v2 := fs:item-sequence-to-node-sequence ($v1)
                 * let $v3 := fs:merge-adjacent-text-nodes ($v2)
                 * let $repl := for $v4 in $v3 return
                 *                typeswitch ($v4)
                 *                  case document-node() return $v4/node()
                 *                  default return $v4
                 * let $parent := $target/parent::node()
                 * return
                 *   typeswitch ($parent)
                 *     case empty() return error ("parent of `replace' empty")
                 *     default return
                 *       typeswitch ($target)
                 *         case xs:anyAttribute return
                 *           let $v4 := upd:insertIntoAsLast ($parent, $repl)
                 *           let $v5 := upd:delete ($target)
                 *           return
                 *             ($v4, $v5)
                 *         case node() return
                 *           let $v6 := upd:insertBefore ($target, $repl)
                 *           let $v7 := upd:delete ($target)
                 *           return
                 *             ($v6, $v7)
                 *         default
                 *           error ("illegal expression in `do replace' expr")
                 */
                PFvar_t *target = new_var (NULL);
                PFvar_t *v1     = new_var (NULL);
                PFvar_t *v2     = new_var (NULL);
                PFvar_t *v3     = new_var (NULL);
                PFvar_t *v4     = new_var (NULL);
                PFvar_t *v5     = new_var (NULL);
                PFvar_t *v6     = new_var (NULL);
                PFvar_t *repl   = new_var (NULL);
                PFvar_t *parent = new_var (NULL);

                PFfun_t *exactly_one
                    = function (PFqname (PFns_fn, "exactly-one"));
                PFfun_t *is2ns
                    = function (PFqname (PFns_pf,
                                         "item-sequence-to-node-sequence"));
                PFfun_t *matn
                    = function (PFqname (PFns_pf,
                                         "merge-adjacent-text-nodes"));
                PFfun_t *insertlast
                    = function (PFqname (PFns_upd, "insertIntoAsLast"));
                PFfun_t *delete  = function (PFqname (PFns_upd, "delete"));
                PFfun_t *replaceNode
                    = function (PFqname (PFns_upd, "replaceNode"));
                PFfun_t *error   = function (PFqname (PFns_fn, "error"));

                C(p)
                  = flwr (
                      let (letbind (var (target), APPLY(exactly_one, C(L(p)))),
                      let (letbind (var (v1), C(R(p))),
                      let (letbind (var (v2), APPLY (is2ns, var (v1))),
                      let (letbind (var (v3), APPLY (matn, var (v2))),
                      let (letbind (
                          var (repl),
                          flwr (
                              for_ (
                                  forbind (
                                      forvars (var (v4), nil ()),
                                      var (v3)),
                                  nil ()),
                              typeswitch (
                                  var (v4),
                                  cases (
                                      case_ (
                                          seqtype (
                                              PFty_doc(PFty_plus(PFty_node()))),
                                          locsteps (step (p_child,
                                                          seqtype(PFty_node())),
                                                    var (v4))),
                                      default_ (var (v4)))))),
                      let (letbind (var (parent),
                                    locsteps (step (p_parent,
                                                    seqtype (PFty_node ())),
                                              var (target))),
                      nil ())))))),
                      typeswitch (
                          var (parent),
                          cases (
                            case_ (seqtype (PFty_empty ()),
                                   APPLY (error,
                                          str("parent of `do replace' empty"))),
                            default_ (
                                typeswitch (
                                    var (target),
                                    cases (
                                        case_ (
                                            seqtype (PFty_xs_anyAttribute ()),
                                            flwr (
                                                let (
                                                    letbind (
                                                        var (v4),
                                                        APPLY (insertlast,
                                                               var (parent),
                                                               var (repl))),
                                                let (
                                                    letbind (
                                                        var (v5),
                                                        APPLY (delete,
                                                               var (target))),
                                                nil ())),
                                                seq (var (v4), var (v5)))),
                                        default_ (
                                            typeswitch (var (target), cases (
                                        case_ (
                                            seqtype (PFty_node ()),
                                            flwr (
                                                let (
                                                    letbind (
                                                        var (v6),
                                                        APPLY (replaceNode,
                                                               var (target),
                                                               var (repl))),
                                                nil ()),
                                                var (v6))),
                                        default_ (
                                            APPLY (
                                                error,
                                                str ("illegal expression in "
                                                     "`do replace expression"))
                                            ))))))))));
            }
            
            break;

        /* RenameExpr:             rename (SourceExpr, NewNameExpr) */
        case 224:
        {
            /*
             * [[ do rename <TargetExpr> into <NewNameExpr> ]]
             *  ==
             * let $target := [[ <TargetExpr> ]] return
             *   let $newName := [[ <NewNameExpr ]] return
             *     upd:rename ($target, $newName)
             */
            PFvar_t *target  = new_var (NULL);
            PFvar_t *newName = new_var (NULL);
            PFfun_t *rename  = function (PFqname (PFns_upd, "rename"));
            PFfun_t *exactly_one = function (PFqname (PFns_fn, "exactly-one"));

            C(p) = flwr (let (letbind (var (target), APPLY(exactly_one, C(L(p)))),
                              nil ()),
                         flwr (let (letbind (var (newName), C(R(p))),
                                    nil ()),
                               APPLY (rename, var (target), var (newName))));
        } break;

        /* TransformExpr:          transform (TransformBindings,
                                             modify (ExprSingle, ExprSingle)) */
        case 225:
            PFoops (OOPS_NOTSUPPORTED,
                    "`transform' expressions are not supported in Pathfinder");
            break;


        /* TransformBindings:      nil */
        case 226:
            /* we do not support transform expressions anyway */
            break;

        /* TransformBindings:      transbinds (TransformBinding,
                                    TransformBindings) */
        case 227:
            /* we do not support transform expressions anyway */
            break;


        /* TransformBinding:       let (var_type (Var, nil), OptExpr) */
        case 228:
            /* we do not support transform expressions anyway */
            break;


        /* SourceExpr:             ExprSingle */
        case 229:
            assert (C(p));
            break;

        /* TargetExpr:             ExprSingle */
        case 230:
            assert (C(p));
            break;

        /* NewNameExpr:            ExprSingle */
        case 231:
        {
            /*
             * XQuery Update Facility (May 8, 2006, Sect. 2.3.4):
             * ``NewNameExpr'' is evaluated as though it were the name
             * expression of a computed element constructor.''
             *
             * (see also Rule TagName: OptExpr)
             */
            PFvar_t *v1 = new_var (NULL);
            PFvar_t *v2 = new_var (NULL);
            PFfun_t *resolve_qname
                = function (PFqname (PFns_fn, "resolve-QName"));

            C(p) = flwr (let (letbind (var (v1), fn_data (C(p))), nil ()),
                         flwr (let (letbind (
                                     var (v2),
                                     fs_convert_op_by_type (var (v1),
                                                            PFty_xs_string ())),
                                    nil ()),
                               APPLY (resolve_qname, var (v2))));
        } break;

        /* SequenceType:           seq_ty (stmt_ty (nil)) */
        case 232:
            /*
             * statement type for updating user-defined functions.
             */
            C(p) = seqtype (PFty_star (PFty_stmt ()));
            break;

        /* ExprSingle:             RecursiveExpr */
        case 233:
            break;

        /* RecursiveExpr:          recursion (
                                       var_type (Var, OptTypeDeclForRec),
                                       seed (ExprSingle, ExprSingle)) */
        case 234:
        {
            PFvar_t *v1 = new_var (NULL);
            PFvar_t *v2 = new_var (NULL);

            /*
             * with $var [as type1] seeded by expr1 recurse expr2
             *
             * Both expressions must evaluate to a subtype of type1
             * (we allow empty sequences for expr2, though).  If
             * absent, type1 defaults to `node*'.  The static type
             * of the overall expression is type1.
             */
            C(p) = flwr (let (letbind (var (v1), C(RL(p))),
                         let (letbind (var (v2),
                                       proof (subty (var (v1), C(LR(p))),
                                              seqcast (C(LR(p)), var (v1)))),
                              nil ())),
                         recursion (C(LL(p)),
                                    seed (var (v2),
                                          proof (
                                              subty (
                                                  C(RR(p)),
                                                  seqtype (
                                                      PFty_opt (
                                                          C(LR(p))->sem.type))),
                                                 C(RR(p))))));
        } break;

        /* OptTypeDeclForRec:      nil */
        case 235:
            /*
             * If no type is given in an
             *
             *   with $x [as type] seeded by e1 return e2
             *
             * expression, we default to node*
             */
            C(p) = seqtype (PFty_star (PFty_node ()));
            break;

        /* OptTypeDeclForRec:      SequenceType */
        case 236:
            break;

        /* XRPCCall:               xrpc (Expr, FunctionCall) */
        case 237:
        {
            PFvar_t *v1 = new_var (NULL);
            PFvar_t *v2 = new_var (NULL);
            PFvar_t *v3 = new_var (NULL);

            C(p) = flwr (let (letbind (var (v1), C(L(p))),
                         let (letbind (var (v2), fn_data (var (v1))),
                         let (letbind (var (v3),
                                       cast (seqtype (PFty_xs_string ()),
                                             var (v2))),
                              nil ()))),
                         xrpc (var (v3), C(R(p))));
        } break;

        /* SequenceType:           seq_ty (docmgmt_ty (nil)) */
        case 238:
            /*
             * statement type for document managemtn user-defined functions.
             */
            C(p) = seqtype (PFty_star (PFty_docmgmt ()));
            break;

        default:
            PFoops_loc (OOPS_FATAL, p->loc, "untranslated expression");
            break;
    }
}

static void
comparison_worker (PFpnode_t *p, PFfun_t *fun)
{
    PFvar_t *v1 = new_var (NULL);
    PFvar_t *v2 = new_var (NULL);

    C(p) = some (
               var (v1), fn_data (C(L(p))),
               some (
                   var (v2), fn_data (C(R(p))),
                   APPLY (
                       fun,
                       fs_convert_op_by_type (
                           fs_convert_op_by_expr (var (v1), var (v2)),
                           PFty_xs_string ()),
                       fs_convert_op_by_type (
                           fs_convert_op_by_expr (var (v2), var (v1)),
                           PFty_xs_string ()))));
}

static void
val_comparison_worker (PFpnode_t *p, PFfun_t *fun)
{
    C(p) = APPLY (fun, 
                  fs_convert_op_by_type (fn_data (C(L(p))),
                                         PFty_xs_string ()),
                  fs_convert_op_by_type (fn_data (C(R(p))),
                                         PFty_xs_string ()));
}

static void
calculation_worker (PFpnode_t *p, PFfun_t *fun)
{
    C(p) = APPLY (fun,
                  fs_convert_op_by_type (fn_data (C(L(p))),
                                         PFty_xs_double ()),
                  fs_convert_op_by_type (fn_data (C(R(p))),
                                         PFty_xs_double ()));
}

static void
two_arg_fun_worker (PFpnode_t *p, PFfun_t *fun)
{
    C(p) = APPLY (fun, C(L(p)), C(R(p)));
}

/**
 * Extract sequence types from function parameter list and
 * put them into a single array, suitable for insertion into
 * PFfun_t.
 *                fun
 *               /   \
 *         fun_sig    Body
 *        /       \
 *    params      ReturnType
 *   /     \
 *  ty      params
 *         /      \
 *        ty      ...
 *                  \
 *                  nil
 *
 * Call with the topmost `param' node.
 */
static PFty_t *
parameter_types (PFpnode_t *paramlist)
{
    /* we assume functions to have less than 15 parameters */
    PFarray_t *par_ty = PFarray (sizeof (PFty_t), 15);
    PFpnode_t *n = paramlist;

    assert (paramlist);

    while (n->kind != p_nil) {
        assert (n->kind == p_params);
        assert (L(n) && L(n)->kind == p_param);
        assert (LL(n) && C(LL(n)) && C(LL(n))->kind == c_seqtype);

        *((PFty_t *) PFarray_add (par_ty)) = C(LL(n))->sem.type;

        n = R(n);
        assert (n);
    }

    return par_ty->base;
}

/**
 * Convert function argument or return value according to
 * to the function conversion rules in W3C XQuery 3.1.5.
 *
 * -- If the expected type is an atomic type,
 * 
 *    (1) Apply atomization (i.e., fn:data)
 *    (2) Cast each xdt:untypedAtomic item to the expected
 *        atomic type. For built-in functions with expected
 *        type `numeric', cast them to xs:double.
 *    (3) Apply numeric type promotion, if possible.
 *
 * -- Leave other arguments as they are. (FIXME: correct?)
 *
 * We use fn:data() for Step 1, fs:convert-operand() for Step 2.
 * FIXME: Step 3?
 *
 * In XQuery Core:
 *
 * -- The expected type t is an atomic type:
 *
 *    let $v1 := arg return
 *      for $v2 in $v1 return
 *        let $v3 := fn:data ($v2) return
 *          convert-operand ($v3, prime(expected))
 *
 */
static PFcnode_t *
function_conversion (const PFcnode_t *e, PFty_t expected)
{
    PFty_t exp_prime;

    PFcnode_t *ret;
    PFvar_t   *v = new_var (NULL);

    assert (e);

    exp_prime = PFty_prime (PFty_defn (expected));

    if (PFty_subtype (exp_prime, PFty_atomic ())) {

        /*
         * Built-in functions allow the expected type to be `numeric'.
         * In that case, we shall convert to xs:double.
         */
        if (PFty_eq (exp_prime, PFty_numeric ()))
            exp_prime = PFty_xs_double ();

        /* if type exp_prime is atomic untypedAtomic 
           does not need to be casted to atomic 
           -- we therefore can skip the convert-op */
        if (PFty_eq (exp_prime, PFty_atomic ()))
            ret = fn_data (e);
        else 
            /* Steps 1 and 2:
             *
             * for $v in fn:data (e) return
             *   convert-op ($v, exp_prime)
             */
            ret = flwr (for_ (forbind (forvars (var (v), nil ()),
                                       fn_data (e)), nil ()),
                        fs_convert_op_by_type (var (v), exp_prime));

    }
    else
        /* do nothing if expected is not atomic */
        ret = (PFcnode_t *) e;

    return ret;
}


/* initialize global values */
void 
PFfs_init (void)
{
    PFquery = NULL;

    /* initially, the context item is undefined */
    fs_dot = NULL;
    fs_position = NULL;
    fs_last = NULL;

    /* initially, there is no function on the current function stack */
    funs = PFarray (sizeof (PFfun_t *), 20);

    /* initially, there is no argument list on the current arguments stack */
    args = PFarray (sizeof (PFcnode_t *), 20);

    /* initially, there nothing on the current argument number stack */
    arg_num = PFarray (sizeof (short), 20);

    /* We collect function signatures in the query prolog here. */
    fun_sigs = nil ();

    /* We collect variable declarations in the query prolog
     * in a root/hole pair. */
    var_decls.root = var_decls.hole = NULL;

    any = PFty_xs_anyType ();
}

/**
 * Compile parse tree into XQuery Core tree.
 *
 * @param r root of the abstract syntax tree
 * @return the Core equivalent of @a r
 */
PFcnode_t *
PFfs (PFpnode_t *r, PFquery_t *query)
{
    assert (r);

    PFfs_init ();
    PFquery = query;

    /* label the parse tree bottom up */
    PFfs_label (r);

    /* invoke compilation */
    process_fun_sigs (r, 1);    /* compile function signatures first */
    reduce (r, 1);              /* then all the rest */

    /* sanity: current function/argument list stacks need to be empty */
    assert (PFarray_empty (funs));
    assert (PFarray_empty (args));

    return r->core;
}


/* vim:set shiftwidth=4 expandtab filetype=c: */
